/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Cookies Storage
 *
 * ************************************************************************
 */

"use strict";

var gCookieStorage = {
  init: function() {
    if (typeof(cookieManagers) == "object" &&
        cookieManagers.__proto__.constructor.name == "Array" &&
        cookieManagers.indexOf(gCookieStorage) == -1) {
      cookieManagers.push(gCookieStorage);
    }
  },

  uninit: function() {
    if (typeof(cookieManagers) == "object" && cookieManagers.__proto__.constructor.name == "Array") {
      let index = cookieManagers.indexOf(gCookieStorage);
      if (index > -1) {
        cookieManagers.splice(index, 1);
      }
    }
  },

  makeProtectedObject: function(aHost, aName, aPath) {
    function cookieEquals(aItem) {
      return aItem      !=  null &&
        (aItem instanceof Ci.nsICookie2 || aItem.getType() === COOKIE_TYPE_STANDARD) &&
        this.host       === aItem.host &&
        this.name       === aItem.name &&
        this.path       === aItem.path;
    }
    function getHost() {
      return this.host;
    }
    function getRawHost() {
      return gCommon.makeStrippedHost(this.host);
    }
    function getName() {
      return this.name;
    }
    function getPath() {
      return this.path;
    }
    function getType() {
      return COOKIE_TYPE_STANDARD;
    }

    let pc = {
      // Properties
      host       : aHost,
      name       : aName,
      path       : aPath,
      // Helper functions
      equals     : cookieEquals,
      getRawHost : getRawHost,
      getHost    : getHost,
      getName    : getName,
      getPath    : getPath,
      getType    : getType
    };

    return pc;
  },

  makeCookieObject: function (aCookie) {
    function cookieEquals(aItem) {
      return (typeof(aItem) == "object") &&
        aItem.getType() == COOKIE_TYPE_STANDARD &&
        aItem.host      == this.host &&
        aItem.name      == this.name &&
        aItem.path      == this.path;
    }
    function getHost() {
      return this.host;
    }
    function getName() {
      return this.name;
    }
    function getType() {
      return COOKIE_TYPE_STANDARD;
    }

    /**
     * Update a cookie with a transform mozilla cookie info.
     * This is used by the observer, when a cookie content is changed.
     */
    function updateFrom(aCookie) {
      if (aCookie.getType() == COOKIE_TYPE_STANDARD) {
        this.value        = aCookie.value;
        this.isDomain     = aCookie.isDomain;
        this.isSecure     = aCookie.isSecure;
        this.expires      = aCookie.expires;
        this.creationTime = aCookie.creationTime;
        this.expiry       = aCookie.expiry;
        this.isHttpOnly   = aCookie.isHttpOnly;
        this.isSession    = aCookie.isSession;
        this.lastAccessed = aCookie.lastAccessed;
        if (typeof(aCookie.originAttributes) == "object") {
          this.originAttributes = aCookie.originAttributes;
        }
      }
    }

    // Some expiry value have I think the int64 max val: 0x7FFFFFFFFFFFFFFF
    // If so, set it to DATE_MAX instead
    var strippedHost = gCommon.makeStrippedHost(aCookie.host);
    var c = {
      // Properties
      name        : aCookie.name,
      value       : aCookie.value,
      isDomain    : aCookie.isDomain,
      host        : aCookie.host,
      rawHost     : strippedHost,
      path        : aCookie.path,
      isSecure    : aCookie.isSecure,
      expires     : aCookie.expires*1000,
      creationTime: aCookie.creationTime/1000,
      expiry      : aCookie.expiry > DATE_MAX ? DATE_MAX : aCookie.expiry*1000,
      isHttpOnly  : aCookie.isHttpOnly,
      isSession   : aCookie.isSession,
      lastAccessed: aCookie.lastAccessed/1000,
      isProtected : false,
      level       : 1,
      parentIndex : -1,
      // Helper functions
      equals      : cookieEquals,
      updateFrom  : updateFrom,
      getHost     : getHost,
      getName     : getName,
      getType     : getType
    };

    c.isProtected = CookieKeeper.isProtectedItem(c);

    if (typeof(aCookie.originAttributes) == "object") {
      c.originAttributes = aCookie.originAttributes;
    }

    if (typeof(aCookie.appId) != "undefined") {
      log.d("Found appId:", aCookie.appId);
    }
    return c;
  },

  /**
   * @brief Fetch the Cookies Storage datas (async load)
   * @param aCallbackFunction a function which accept one arg: an array of the loaded items.
   */
  fetchItems: function(aCallbackFunction) {
    setTimeout(function() {
      var cookies     = [];
      var cookiesEnum = Services.cookies.enumerator;

      // Load cookies
      while (cookiesEnum.hasMoreElements()) {
        let cookie = cookiesEnum.getNext();

        if (cookie instanceof Ci.nsICookie2) {
          if (cookie.originAttributes && cookie.originAttributes.appId != 0) {
            // We can't change this cookie, usually a ".google.com" / NID cookie
            // created when Firefox fetch the "safe browsing" list.
            continue;
          }
          // Search if it is a protected cookie

          let cookieObj = gCookieStorage.makeCookieObject(cookie);
          cookies.push(cookieObj);
        }
      }

      aCallbackFunction(cookies);
    }, 0);
  },

  _performDeletion: function (deleteItems, doBlock, blockFutureCookies) {
    let sc = Services.cookies;

    for (let i=0; i<deleteItems.length; i++) {
      let item = deleteItems[i];
      if (item.getType() == COOKIE_TYPE_STANDARD) {
        // remove() function signature changed with https://bugzilla.mozilla.org/show_bug.cgi?id=1245184
        if (typeof(item.originAttributes) == "object") {
          sc.remove(item.host, item.name, item.path, item.originAttributes, blockFutureCookies);
        } else {
          sc.remove(item.host, item.name, item.path, blockFutureCookies);
        }

        if (item.isProtected) {
          CookieKeeper.removeProtectedItem(item);
        }
      }
    }
  },

  performDeletion: function (deleteItems, forceBlock) {
    if (deleteItems.length > 0) {
      this._performDeletion(deleteItems, true, forceBlock);
    }
  },

  performDeletionAll: function(deleteItems) {
    if (deleteItems.length > 0) {
      this._performDeletion(deleteItems, false, false);
    }
  },

  /// @brief Remove all unwanted cookies.
  /// @note This function is used at program exit,
  /// or via the main menus (tools menu/toolbar menu)
  clearUnwanted: function(protList) {
    function isProtectedCookie(aCookie) {
      var isProtected = false;
      for (let i=0; i<protList.length && !isProtected; i++) {
        let pi = protList[i];
        isProtected = pi.equals(aCookie);
      }

      return isProtected;
    }

    var clearList = [];
    var sCookies  = Services.cookies;
    var cEnum     = sCookies.enumerator;

    // Ususally it is not good to delete item on the same list we use an enumerator.
    // Since I don't know the implementation, I prefer to use a safe method:
    // Get all items we need to delete, then delete them.
    // This way, I'm sure to not messed up the iterator.
    while (cEnum.hasMoreElements()) {
      let cookie = cEnum.getNext();
      if (cookie instanceof Ci.nsICookie2 && !isProtectedCookie(cookie)) {
        clearList.push(cookie);
      }
    }

    for (let i=0; i<clearList.length; i++) {
      let cookie = clearList[i];

      if (typeof(cookie.originAttributes) == "object") {
        sCookies.remove(cookie.host, cookie.name, cookie.path, cookie.originAttributes, false);
      } else {
        sCookies.remove(cookie.host, cookie.name, cookie.path, false);
      }

    }
  }

};
