/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Dom HTML5 Storage
 *
 * Using sqlite module, since there is no interface for it.
 *
 * ************************************************************************
 */

"use strict";

var gDomStorage = {

  init: function() {
    if (typeof(cookieManagers) == "object" &&
        cookieManagers.__proto__.constructor.name == "Array" &&
        cookieManagers.indexOf(gDomStorage) == -1) {
      cookieManagers.push(gDomStorage);
    }
  },

  uninit: function() {
    if (typeof(cookieManagers) == "object" && cookieManagers.__proto__.constructor.name == "Array") {
      let index = cookieManagers.indexOf(gDomStorage);
      if (index > -1) {
        cookieManagers.splice(index, 1);
      }
    }
  },

  makeProtectedObject: function(aScope, aKey) {
    function storageEquals(aItem) {
      // aItem can be a protected storage, or a dom storage
      return  aItem !=  null &&
              COOKIE_TYPE_STORAGE == aItem.getType() &&
              this.scope === aItem.scope &&
              this.key === aItem.key;
    }
    function getHost() {
      let aScope = this.scope.split(":");
      if (aScope.length == 3) {
        return gCommon.reverseStr(aScope[0]);
      }
      return this.scope;
    }
    function getProtocol() {
      let aScope = this.scope.split(":");
      if (aScope.length == 3) {
        return aScope[1];
      }
      return this.scope;
    }
    function getPort() {
      let aScope = this.scope.split(":");
      if (aScope.length == 3) {
        return aScope[2];
      }
      return this.scope;
    }
    function getRawHost() {
      return gCommon.makeStrippedHost(this.getHost());
    }
    function getName() {
      return this.key;
    }
    function getPath() {
      return "";
    }
    function getType() {
      return COOKIE_TYPE_STORAGE;
    }

    var ps = {
      // Properties
      scope       : aScope,
      key         : aKey,
      // Helper functions
      equals      : storageEquals,
      getHost     : getHost,
      getProtocol : getProtocol,
      getPort     : getPort,
      getRawHost  : getRawHost,
      getName     : getName,
      getPath     : getPath,
      getType     : getType
    };

    return ps;
  },

  makeDomStorageObject: function(aScope, aKey, aValue) {
    /// Inner DOM Storage item function
    function itemEquals(aItem) {
      return (typeof(aItem) == "object") &&
        aItem.getType() == COOKIE_TYPE_STORAGE &&
        aItem.scope     == this.scope &&
        aItem.key       == this.key;
    } // function itemEquals()
    function getHost() {
      let aScope = this.scope.split(":");
      if (aScope.length == 3) {
        return gCommon.reverseStr(aScope[0]);
      }
      return this.scope;
    }
    function getProtocol() {
      let aScope = this.scope.split(":");
      if (aScope.length == 3) {
        return aScope[1];
      }
      return this.scope;
    }
    function getPort() {
      let aScope = this.scope.split(":");
      if (aScope.length == 3) {
        return aScope[2];
      }
      return this.scope;
    }
    function getRawHost() {
      return gCommon.makeStrippedHost(this.getHost());
    }
    function getName() {
      return this.key;
    }
    function getType() {
      return COOKIE_TYPE_STORAGE;
    }

    /// Inner DOM Storage item function
    function isJSON() {
      try {
        JSON.parse(this.value);
        return true;
      } catch (ex) {
        return false;
      }
    } // function isJSON()

    function updateFrom(aItem) {
      if (aItem.getType() == COOKIE_TYPE_STORAGE) {
        this.value = aItem.value;
      }
    }

    let scopeArray = aScope.split(":");
    let host  = "";
    let proto = "";
    let port  = "";
    switch (scopeArray.length) {
    case 3:
      port  = scopeArray[2];
      // fall-through
    case 2:
      proto = scopeArray[1];
      // fall-through
    case 1:
      host  = gCommon.reverseStr(scopeArray[0]);
    }

    let item = {
      // Properties
      scope       : aScope,
      host        : host,
      rawHost     : gCommon.makeStrippedHost(host),
      proto       : proto,
      port        : port,
      key         : aKey,
      value       : aValue,
      path        : "",
      isDomain    : false,
      isSecure    : (proto == "https"),
      expires     : DATE_MAX,
      creationTime: 1000,
      expiry      : DATE_MAX,
      isHttpOnly  : (proto == "http" || proto == "https"),
      isSession   : false,
      lastAccessed: 1000,
      isProtected : false,
      level       : 1,
      parentIndex : -1,
      // Helper functions
      equals      : itemEquals,
      updateFrom  : updateFrom,
      isJSON      : isJSON,
      getHost     : getHost,
      getProtocol : getProtocol,
      getPort     : getPort,
      getRawHost  : getRawHost,
      getName     : getName,
      getType     : getType
    };
    item.isProtected = CookieKeeper.isProtectedItem(item);

    return item;
  },

  /**
   * @brief Fetch the DOM Storage datas (async load).
   * @param aCallbackFunction a function which accept one arg: an array of the loaded items.
   */
  fetchItems: function(aCallbackFunction) {
    var storages = [];
    var connect  = null;

    function onRow(aRow) {
      if (aRow == null) {
        log.e("Fetch DOM storage (create item) row is null");
        return;
      }

      try {
        let scope  = aRow.getString(0);
        let key    = aRow.getString(1);
        let value  = (aRow.getTypeOfIndex(2) == aRow.VALUE_TYPE_TEXT    ? aRow.getString(2) : null);
        if (scope == null || scope.length == 0 || key == null || key.length == 0) {
          return;
        }

        let item = gDomStorage.makeDomStorageObject(scope, key, value);
        storages.push(item);
      } catch (ex) {
        log.e("Fetch DOM storage (create item) exception: " + ex.message, ex.stack);
      }
    } // function onRow()

    function onSelectComplete(result) {
      try {
        let typeofResult = typeof(result);
        if (typeofResult != "boolean" && (typeofResult == "undefined" || result != null)) {
          log.e("Fetch DOM storage (SELECT) exception: ", result);
        }

        aCallbackFunction(storages);

      } catch (ex) {
        log.e("Fetch DOM storage (callback): " + ex.message, ex.stack);
      } finally {
        connect.close();
      }
    } // function onSelectComplete()

    function doSelect(conn) {

      connect = conn;
      try {
        // ignore chrome:// schemes
        let res = connect.execute("SELECT scope, key, value FROM webappsstore2 where scope LIKE :doms", {doms: "%:%:%"}, onRow);
        res.then(onSelectComplete, onSelectComplete);
      } catch (ex) {
        log.e("Fetch DOM storage (connection) exception:" + ex.message, ex.stack);
        aCallbackFunction([]);
      }
    } // function doSelect()

    let connection = Sqlite.openConnection({path: OS.Path.join(OS.Constants.Path.profileDir, DOM_DB_NAME)});
    connection.then(doSelect, Cu.reportError);
  },

  /**
   * Delete dom storage items
   * @param aItems: An array of DOM Storage items to delete.
   * @return true if the delete is launch (aItem is not null and contains data)
   */
  performDeletion: function (aItems, doBloc) {
    var db    = null;
    var error = null;

    if (typeof(doBloc) != "boolean") {
      doBloc = false;
    }

    function doTransaction() {
      let nItems = aItems.length;
      for (let i=0; i<nItems; i++) {
        let item = aItems[i];
        if (item.getType() == COOKIE_TYPE_STORAGE) { // safety check
          yield db.execute("DELETE FROM webappsstore2 WHERE scope=? AND key=?", [item.scope, item.key], null);
        }
      }
    }

    function doDelete() {
      try {
        db = yield Sqlite.openConnection({path: OS.Path.join(OS.Constants.Path.profileDir, DOM_DB_NAME)});
        yield db.executeTransaction(doTransaction);
      } catch (ex) {
        log.e(ex, ex.stack);
        error = ex.message;
      } finally {
        if (db != null) {
          db.close();
          db = null;
        }
        // Item are deleted, notify observers
        let subject = {
          error  : error,
          deleted: error == null ? aItems : []
        }
        Services.obs.notifyObservers({ wrappedJSObject: subject }, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_DOM_DELETE);

        // Block the domains if needed
        if (doBloc) {
          let nItems = aItems.length;
          for (let i=0; i<nItems; i++) {
            let item = aItems[i];
            if (item.getType() == COOKIE_TYPE_STORAGE) {
              let aHost = item.getHost();
              aHost = aHost.charAt(0) == "." ? aHost.substring(1, aHost.length) : aHost;
              let uri = Services.io.newURI(item.getProtocol() + "://" + aHost, null, null);
              CKPermissionsManager.prototype.add(uri, MOZ_COOKIE_PERM_DENY);
            }
          }
        }
      }
    }

    if (aItems == null || aItems.length == 0) {
      return false;
    }
    Task.spawn(doDelete);
    return true;
  },

  /**
   * @brief Insert items in the webappsstore.sqlite database.
   * If an item already exists, it will be updated.
   * @param aItems An array of dom storage items to insert.
   * @return true if the insert is launch (aItem is not null and contains data)
   * @note
   * Each item must be in this form:
   * let item = {scope:"ascope", key:"akey", value:"avalue"};
   * Note that the scope must be in db format: the domain name must be reversed
   * (eg: gro.allizom.snodda.:https:443 for addons.mozilla.org)
   */
  performInsertion: function (aItems) {
    var db    = null;
    var error = null;
    var inserted = [];

    function doTransaction() {

      let nItems = aItems.length;
      for (let i=0; i<nItems; i++) {
        let item = aItems[i];
        yield db.execute("INSERT OR REPLACE INTO webappsstore2 (scope,key,value) VALUES (?,?,?)", [item.scope, item.key, item.value]);
      }
    }
    function doInsert() {
      try {
        db = yield Sqlite.openConnection({path: OS.Path.join(OS.Constants.Path.profileDir, DOM_DB_NAME)});
        yield db.executeTransaction(doTransaction);

        inserted = aItems;
      } catch (ex) {
        log.e("Insert DOM Storage:" + ex.message, ex.stack);
        error = ex;
      } finally {
        if (db != null) {
          db.close();
          db = null;
        }
        // Notify the update is done
        let subject = {
          error : error,
          insert: inserted
        }
        Services.obs.notifyObservers({ wrappedJSObject: subject }, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_DOM_INSERT);
      }
    }

    if (aItems == null || aItems.length == 0) {
      return false;
    }
    Task.spawn(doInsert);
    return true;
  },

  /// @brief Remove all unwanted DOM Storage.
  /// @note This function is used at program exit,
  /// or via the main menus (tools menu/toolbar menu)
  clearUnwanted: function(protList) {
    const scopeAndKey = "(scope=? AND key=?)";
    var delQuery      = "DELETE FROM webappsstore2 WHERE NOT ((scope NOT LIKE \"%:%:%\")";
    var protInfo      = [];
    var connection    = null;

    // Get protected DOM storage infos in a single list
    for (let i=0; i<protList.length; i++) {
      let pi = protList[i];
      protInfo.push(pi.scope);
      protInfo.push(pi.key);
    }
    // Build the query

    try {
      // Perform the clean sync, not using Sqlite.jsm, because
      // it will not be executed on shutdown.
      // This module will be unregistered, before the AsyncShutdown task is launch.

      let path   = OS.Path.join(OS.Constants.Path.profileDir, DOM_DB_NAME);
      let file   = FileUtils.File(path);
      connection = Services.storage.openDatabase(file);

      if (protInfo.length > 0) {
        for (let i=0; i<protInfo.length; i+=2) {
          delQuery = delQuery + " OR " + scopeAndKey;
        }
        delQuery = delQuery + ")";

        if (connection.connectionReady) {
          let stmt   = connection.createStatement(delQuery);

          for (let i=0; i<protInfo.length; i++) {
            stmt.bindByIndex(i, protInfo[i]);
          }
          stmt.execute();
          stmt.finalize();
        }
      } else {
        let path   = OS.Path.join(OS.Constants.Path.profileDir, DOM_DB_NAME);
        let file   = FileUtils.File(path);
        connection = Services.storage.openDatabase(file);
        if (connection.connectionReady) {
          connection.executeSimpleSQL("DELETE FROM webappsstore2 WHERE (scope LIKE \"%:%:%\")");
        }
      }
    } catch (ex) {
      log.e("Error cleaning dom storage:", ex);
    } finally {
      if (connection != null) {
        connection.close();
        sObserver.notifyObservers(null, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_DOM_CLEAN);
      }
    }
  }

};
