/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Transform a normal HTML table.
 * - Allow to resize columns.
 * - Content may be scrollable, with the header stay in place.
 * - Allow to select rows.
 *
 * ************************************************************************
 */
"use strict";

const IMG_SORT_INDICATOR_UP   = "url('chrome://cookiekeeper/skin/arrow_up.svg')";
const IMG_SORT_INDICATOR_DOWN = "url('chrome://cookiekeeper/skin/arrow_down.svg')";

/**
 * @brief DynamicTable: Class used to transform & manage an HTML table like a native table.
 * @par
 * Expect having something like:
 * <table>
 *  <thead>
 *    <tr>
 *      <th>head 1</th><th>head 2</th>
 *    </tr>
 *  </thead>
 *  <tbody>
 *    <tr>
 *      <td>cell 1</td><td>cell 2</td>
 *    </tr>
 *  </tbody>
 * </table>
 */
function DynTable(hTable, sPrefixId) {
  // Constants properties:
  Object.defineProperty(this, "CSS_MBP_PROPS", {
    writable: false,
    value: ["margin-left", "margin-right", "border-left-width", "border-right-width", "padding-left", "padding-right"]
  });

  if (typeof(hTable) != "object" || typeof(hTable.nodeName) != "string" || hTable.nodeName != "TABLE") {
    throw "Invalid parameter hTable";
  }
  if (typeof(sPrefixId) != "string") {
    throw "Invalid parameter sPrefixId";
  }

  // Properties:
  this.hTable           = hTable;     // The table to "tranform" into a dynamic table -> like a native application table.
  this.sPrefixId        = sPrefixId;  // Prefix of "id" attribute
  this.nColumns         = 0;          // Number of columns of the table
  this.nMBPHeader       = 0;          // width of margin + border + padding for a header cell element
  this.nMBPBody         = Number.NaN; // width of margin + border + padding for a body cell element
  this.hDivGrips        = null;       // HTML Element/Node used to resize the columns
  this.nVScrollbarWidth = 0;          // Vertical scrollbar width
  this.aColsSizes       = [];         // Size (width) of the columns
  this.aMinColsSize     = [];         // Minimup columns sizes
  this.aSelected        = [];         // Selected rows
  this.nIdxLastSelected = -1;         // Last item index selected
  this.cbItemSelection  = null;       // Callback method called when the selection change
  this.aSortColumns     = null;       // Array of sort columns
  this.fSortComparator  = null;       // Function used to sort two row elements

  // Constructor like code:
  this.hTable.dynTable = this; // Needed for the selection
  let headerHeight = 0;

  // Get nColumns, nMBPHeader, default columns size values, and others values
  let hTHead = this.getTableHeader();
  if (hTHead == null) {
    throw "<thead> element not found";
  } else while (true) {
    let hTrHeader = hTHead.firstElementChild;
    if (hTrHeader == null) break;
    this.nColumns = hTrHeader.children.length;
    if (this.nColumns == 0) break;
    headerHeight  = hTrHeader.clientHeight; // use to set the height value of a grip

    // Get the margin, border, padding of an header cell:
    let cs = window.getComputedStyle(hTrHeader.firstElementChild, null);

    // Get nMBPHeader value
    // The width style property value is added to this value, to get the total element width
    for (let j=0; j<this.CSS_MBP_PROPS.length; j++) {
      let cssProp = cs.getPropertyValue(this.CSS_MBP_PROPS[j]);
      // Values return by getPropertyValue() will be in the form "1px". We need to transform them to an integer
      this.nMBPHeader += Number.parseInt(cssProp.substring(0, cssProp.length - 2), 10);
    }

    // Get the columns default and minimum sizes
    // TODO get saved columns sizes if any
    for (let j=0; j<this.nColumns; j++) {
      let cw, cmw, nw, nmw;
      let hTdElem = hTrHeader.children[j];
      cs  = window.getComputedStyle(hTdElem, null);
      cw  = cs.getPropertyValue("width");
      cmw = cs.getPropertyValue("min-width");

      nw  = Number.parseInt(cw.substring(0, cw.length - 2), 10);
      nmw = Number.parseInt(cmw.substring(0, cmw.length - 2), 10);

      // Be sure to respect the minimum sizes
      if (nw < nmw) {
        nw = nmw;
      }

      this.aColsSizes.push(nw);
      this.aMinColsSize.push(nmw);
    }

    break;
  }

  if (this.nColumns == 0) {
    throw "The table do not seems to have any columns";
  }
  if (this.nColumns != this.aColsSizes.length) {
    throw "Number of column(s) mismatch";
  }

  // Create the columns resizers:
  this.hDivGrips = document.getElementById(sPrefixId + "_table_scroll_grips");
  if (this.hDivGrips == null) {
    /*
    Create the resizer div wich looks like:
    <div id="[sPrefixId]_table_scroll_grips" class="table_scroll_grips">
      <div class="table_scroll_grip"><div id="[sPrefixId]_resizer_0" class="table_scroll_resizer"></div></div>
      <div class="table_scroll_grip"><div id="[sPrefixId]_resizer_1" class="table_scroll_resizer"></div></div>
      <div class="table_scroll_grip"><div id="[sPrefixId]_resizer_2" class="table_scroll_resizer"></div></div>
      <div class="table_scroll_grip"><div id="[sPrefixId]_resizer_3" class="table_scroll_resizer"></div></div>
    </div>
    */
    this.hDivGrips = document.createElement("div");
    this.hDivGrips.setAttribute("id", sPrefixId + "_table_scroll_grips");
    this.hDivGrips.style.height = "0px";
    this.hDivGrips.style.position = "relative";

    let sGripHeight = headerHeight.toString(10) + "px";

    for (let i=0; i<(this.nColumns - 1); i++) {
      let hDivGrip = document.createElement("div");
      hDivGrip.leftPosition   = 0;
      hDivGrip.style.position = "absolute";
      hDivGrip.style.margin   = "0px";
      hDivGrip.style.padding  = "0px";
      hDivGrip.style.height   = sGripHeight;

      // The div resizer will listen for a mouse down event, to act like a capture mouse,
      // we can then resize the columns by knowing the movementX of the mouse.
      let hDivResizer = document.createElement("div");
      hDivResizer.setAttribute("id", sPrefixId + "_resizer_" + i.toString(10));
      hDivResizer.style.width    = "0.4rem";
      hDivResizer.style.height   = "100%";
      hDivResizer.style.position = "absolute";
      hDivResizer.style.cursor   = "col-resize";
      // hDivResizer.style.backgroundColor = "blue"; // DEBUG
      hDivResizer.nColIndex      = i;
      hDivResizer.addEventListener("mousedown", DynTable.prototype.onGripMouseDown, false);
      hDivGrip.appendChild(hDivResizer);

      this.hDivGrips.appendChild(hDivGrip);
    }

    hTable.parentNode.insertBefore(this.hDivGrips, hTable);
  }
  this.hDivGrips.dynTable = this;

  // Now that the grips are attached to the document,
  // we can adjust their positions (style.left)

  // We need first to get the resizer grip width in pixel.
  // Actually we need the half width grip size, to centered it between 2 header cols.
  let nResizerHalfWidth = 0
  {
    let cs = window.getComputedStyle(this.hDivGrips.firstElementChild.firstElementChild, null);
    let cw = cs.getPropertyValue("width");
    let nw = Number.parseInt(cw.substring(0, cw.length - 2), 10);
    nResizerHalfWidth = Math.trunc(nw / 2);

    // To prevent any problem when doing the resize, set the minimum column width
    // to the resizer width
    for (let i=0; i<this.nColumns; i++) {
      if (this.aMinColsSize[i] < nw) {
        this.aMinColsSize[i] = nw;
      }
    }
  }

  let leftPosition = 0;
  for (let i=0; i<(this.nColumns - 1); i++) {
    let hDivGrip = this.hDivGrips.children[i];

    leftPosition         += this.aColsSizes[i] + this.nMBPHeader;
    hDivGrip.leftPosition = leftPosition - nResizerHalfWidth;
    hDivGrip.style.left   = hDivGrip.leftPosition.toString(10) + "px";
  }

  // Finally adjust the <tbody> rows columns size
  this.resizeBodyColumns();

  this.hTable.addEventListener("keypress", DynTable.prototype.tableKeyListener, false);
}

// DynTable Functions:

/**
 * @brief Fake destructor, to help a little the garbage collector
 */
DynTable.prototype.destructor = function() {
  // Be sure to remove document listener, normally they shouldn't be there.
  this.removeMouseEvents();
  // Do other cleaning
  if (this.hDivGrips != null) {
    while (this.hDivGrips.firstElementChild != null) {
      let hDivGrip    = this.hDivGrips.firstElementChild;
      let hDivResizer = hDivGrip.firstElementChild;
      hDivResizer.removeEventListener("mousedown", DynTable.prototype.onGripMouseDown, false);
      hDivGrip.removeChild(hDivResizer);
      this.hDivGrips.removeChild(hDivGrip);
    }
    if (this.hDivGrips.parentNode != null) {
      this.hDivGrips.parentNode.removeChild(this.hDivGrips);
    }
  }
  if (this.aSortColumns != null) {
    for (let i=0; i<this.nColumns; i++) {
      let hThElem = this.getHeader(i);
      // Remove sort indicator
      hThElem.style.backgroundImage = "";
      // Stop listening the "click" event
      hThElem.removeEventListener("click", DynTable.prototype.onSort, false);
    }
    this.fSortComparator = null;
    this.aSortColumns    = null;
  }

  let hTBody = this.getTableBody();
  hTBody.removeEventListener("click", DynTable.prototype.onRowClick, false);

  delete this.aSelected;
  delete this.cbItemSelection;
  delete this.hTable.dynTable;
  delete this.hDivGrips.dynTable;
  delete this.hDivGrips;
  delete this.hTable;
  delete this.aColsSizes;
  delete this.sPrefixId;
}

/**
 * @return the HTML Element of the table header: <thead>
 */
DynTable.prototype.getTableHeader = function() {
  const nChildren = this.hTable.children.length;
  for (let i=0; i<nChildren; i++) {
    let hElem = this.hTable.children[i];
    if (hElem.nodeName == "THEAD") {
      return hElem;
    }
  }

  return null;
}

/**
 * @return the HTML Element of the table body: <tbody>
 */
DynTable.prototype.getTableBody = function() {
  const nChildren = this.hTable.children.length;
  for (let i=0; i<nChildren; i++) {
    let hElem = this.hTable.children[i];
    if (hElem.nodeName == "TBODY") {
      return hElem;
    }
  }

  return null;
}

DynTable.prototype.getHeader = function(index) {
  let hTHead = this.getTableHeader();
  if (hTHead != null && index >= 0 && index < this.nColumns) {
    let hTrElem = hTHead.firstElementChild;
    return hTrElem.children[index];
  }
  return null;
}

/**
 * @return the width in pixel (number) margin+border+padding, of a table body cell.
 */
DynTable.prototype.getMBPBody = function() {
  let nbpBody = 0;
  let hTBody = this.getTableBody();

  if (hTBody == null) {
    throw "Table element <tbody> not found";
  }
  if (hTBody.firstElementChild != null) {
    let hTrElem = hTBody.firstElementChild;
    if (hTrElem.firstElementChild == null) {
      throw "First row od the body do not have any <td> (cell) element";
    }

    let hTdElem   = hTrElem.firstElementChild;
    let cs        = window.getComputedStyle(hTdElem, null);

    for (let j=0; j<this.CSS_MBP_PROPS.length; j++) {
      let cssProp = cs.getPropertyValue(this.CSS_MBP_PROPS[j]);
      // Values return by getPropertyValue() will be in the form "1px". We need to transform them to an integer
      nbpBody += Number.parseInt(cssProp.substring(0, cssProp.length - 2), 10);
    }
  }

  return nbpBody;
}

/*
 TODO
DynTable.prototype.resizeHeaderColumns = function() {
}
*/

DynTable.prototype.resizeBodyColumns = function() {
  const hTBody = this.getTableBody();

  if (hTBody == null) {
    throw "Table element <tbody> not found";
  }

  if (hTBody.children.length == 0) {
    // nothing to do
    return;
  }

  const nCols = this.nColumns;          // Number of columns
  const nLCol = nCols - 1;              // Last column id
  const nRows = hTBody.children.length; // Number of rows

  if (Number.isNaN(this.nMBPBody)) {
    this.nMBPBody = this.getMBPBody();
  }

  // Set the size to the minimum size first.
  // Older version of Gecko engine allow the table to be bigger than expected.
  // As a result, we have some line wrapped into another line.
  for (let r=0; r<nRows; r++) {
    let hTrElem = hTBody.children[r];

    if (hTrElem.children.length != nCols) {
      let nMax = Math.min(hTrElem.children.length, nCols);

      log.e("Table body row " + r.toString(10) + " do not have the same number of columns (" +
            hTrElem.children.length.toString(10) + ") than the header (" + nCols.toString(10) + ")");

      for (let c=0; c<nMax; c++) {
        let hTdElem         = hTrElem.children[c];
        hTdElem.style.width = this.aMinColsSize[c].toString(10) + "px";
      }

    } else for (let c=0; c<nCols; c++) {
      let hTdElem         = hTrElem.children[c];
      hTdElem.style.width = this.aMinColsSize[c].toString(10) + "px";
    }
  }

  const nTrWidth = hTBody.firstElementChild.clientWidth;

  // We can deduce the vertical scrollbar width now too
  this.nVScrollbarWidth = hTBody.offsetWidth - hTBody.clientWidth;
  // So update the minimum column size of the last column
  if (this.nVScrollbarWidth > 0) {
    // This will prevent when resizing the last 2 columns size,
    // to wrap to another line
    this.aMinColsSize[nLCol] = this.aMinColsSize[nLCol] + this.nVScrollbarWidth;
  }

  // create the array of width style strings elements
  let aBodyColsSize = new Array(nCols);
  let totalWidth    = 0;

  for (let c=0; c<nCols; c++) {
    let width = 0;
    if (c == nLCol) {
      // For the last column, we may have the scrollbar, so it's width is reduce
      width = nTrWidth - totalWidth - this.nMBPBody;
    } else {
      width = this.aColsSizes[c] + this.nMBPHeader - this.nMBPBody;
      totalWidth += width; // this.aColsSizes[c] + this.nMBPHeader;
    }

    // Be sure the column width is not less than the minimum col size
    if (width < this.aMinColsSize[c]) {
      width = this.aMinColsSize[c];
    }

    aBodyColsSize[c] = width.toString(10) + "px";
  }

  // Apply the width style to all <td> elems inside the <tbody>
  for (let r=0; r<nRows; r++) {
    let hTrElem = hTBody.children[r];

    if (hTrElem.children.length != nCols) {
      let nMax = Math.min(hTrElem.children.length, nCols);
      for (let c=0; c<nMax; c++) {
        let hTdElem         = hTrElem.children[c];
        hTdElem.style.width = aBodyColsSize[c];
      }
    } else for (let c=0; c<nCols; c++) {
      let hTdElem         = hTrElem.children[c];
      hTdElem.style.width = aBodyColsSize[c];
    }
  }
}

DynTable.prototype.doGripMove = function(evtMouseMove) {
  let movementX = 0;
  let direction = 1;

  if (typeof(evtMouseMove.movementX) == "number") {
    movementX = evtMouseMove.movementX;
  } else {
    // Older gecko version (FFx ESR 38.0):
    movementX = evtMouseMove.mozMovementX;
  }

  if (movementX == 0) {
    // nothing to do.
    return;
  }

  if (typeof(this.nCurrentColIndex) != "number") {
    log.e("this.nCurrentColIndex is unset");
    return;
  }

  // Calc the header new row sizes.
  // One header width is increase, the other decrease

  let nFirstIdx  = this.nCurrentColIndex;
  let nSecondIdx = nFirstIdx + 1;

  // We want the second column width decrease, first increase.
  if (movementX < 0) {
    // First column decrease, Second increase, invert the values
    direction  = -1;
    nSecondIdx = this.nCurrentColIndex;
    nFirstIdx  = nSecondIdx + 1;
    movementX  = -movementX;
  }

  let hFirstHeader  = this.getHeader(nFirstIdx);
  let hSecondHeader = this.getHeader(nSecondIdx);

  if (hFirstHeader == null || hSecondHeader == null) {
    log.e("DynTable.doGripMove: index error: " + this.nCurrentColIndex.toString(10), evtMouseMove);
    return;
  }

  let cs = window.getComputedStyle(hSecondHeader, null);
  let cw = cs.getPropertyValue("width");
  let nw = Number.parseInt(cw.substring(0, cw.length - 2), 10);
  let mw = this.aMinColsSize[nSecondIdx];

  // Check the minimum size
  if (nw - movementX < mw) {
    if (nw > mw) {
      // We may reduce the size little bit more
      movementX = nw - mw;
    } else {
      // don't allow the resize further, we have reach the minimum size.
      // log.w("minimum column size reached", movementX, nw, mw);
      return;
    }
  } // No problem

  hSecondHeader.style.width = (nw - movementX).toString(10) + "px";
  cs = window.getComputedStyle(hFirstHeader, null);
  cw = cs.getPropertyValue("width");
  nw = Number.parseInt(cw.substring(0, cw.length - 2), 10);
  hFirstHeader.style.width = (nw + movementX).toString(10) + "px";

  // Update the grip left style value
  let hDivGrip = this.hDivGrips.children[this.nCurrentColIndex];
  if (direction == -1) {
    hDivGrip.leftPosition = hDivGrip.leftPosition - movementX;
  } else {
    hDivGrip.leftPosition = hDivGrip.leftPosition + movementX;
  }
  hDivGrip.style.left = hDivGrip.leftPosition.toString(10) + "px";

  // Change the width for the body cells
  let hTBody  = this.getTableBody();
  let sFWidth = null;
  let sSWidth = null;
  const nRows = hTBody.children.length;

  for (let i=0; i<nRows; i++) {
    let hTrElem = hTBody.children[i];
    if (hTrElem.children.length < this.nCurrentColIndex + 1) {
      //  not enough columns
      continue;
    }

    let hFTd = hTrElem.children[nFirstIdx];
    let hSTd = hTrElem.children[nSecondIdx];

    if (sFWidth == null) {
      cs = window.getComputedStyle(hFTd, null);
      cw = cs.getPropertyValue("width");
      nw = Number.parseInt(cw.substring(0, cw.length - 2), 10);
      sFWidth = (nw + movementX).toString(10) + "px";

      cs = window.getComputedStyle(hSTd, null);
      cw = cs.getPropertyValue("width");
      nw = Number.parseInt(cw.substring(0, cw.length - 2), 10);
      sSWidth = (nw - movementX).toString(10) + "px";
    }

    // Doing the decrease first, seems more appropriate
    hSTd.style.width = sSWidth;
    hFTd.style.width = sFWidth;
  }
}

DynTable.prototype.onGripMouseMove = function(evtMouseMove) {

  if (document.dynTable != null) {
    document.dynTable.doGripMove(evtMouseMove);
    evtMouseMove.stopImmediatePropagation();
    evtMouseMove.stopPropagation();
    evtMouseMove.preventDefault();
    return false;
  } else {
    log.e("DynTable.onGripMouseMove: document.dynTable is null");
  }
  return true;
}
DynTable.prototype.onGripMouseUp = function(evtMouseUp) {

  if (evtMouseUp.button == 0) {
    if (document.dynTable == null) {
      log.e("DynTable.onGripMouseUp: document.dynTable is null");
    } else {
      document.dynTable.removeMouseEvents();
      evtMouseUp.stopImmediatePropagation();
      evtMouseUp.stopPropagation();
      evtMouseUp.preventDefault();
      return false;
    }
  }
  return true;
}
DynTable.prototype.onGripMouseDown = function(evtMouseDown) {

  if (evtMouseDown.button == 0) {
    let hElem = this.parentNode;
    while (hElem != null && typeof(hElem.dynTable) != "object") {
      hElem = hElem.parentNode;
    }
    if (hElem == null) {
      log.e("DynTable.onGripMouseDown: our dynTable instance is not found");
      return true;
    }

    document.addEventListener("mousemove", DynTable.prototype.onGripMouseMove, false);
    document.addEventListener("mouseup",   DynTable.prototype.onGripMouseUp,   false);
    document.dynTable                  = hElem.dynTable;
    document.dynTable.nCurrentColIndex = this.nColIndex;

    // Set the <body> cursor to "col-resize"
    // sometimes there is a glitch: cursor change form, so apply the same cursor
    // everywhere to prevent it.
    while (hElem != null && hElem.nodeName != "BODY") {
      hElem = hElem.parentNode;
    }
    if (hElem != null) {
      hElem.style.cursor = "col-resize";
    }

    // Prevent text selection, or any other things
    evtMouseDown.stopImmediatePropagation();
    evtMouseDown.stopPropagation();
    evtMouseDown.preventDefault();
    return false;
  }
  return true;
}

DynTable.prototype.removeMouseEvents = function() {
  document.removeEventListener("mousemove", DynTable.prototype.onGripMouseMove, false);
  document.removeEventListener("mouseup", DynTable.prototype.onGripMouseUp, false);
  // Reset the <body> cursor to it's default value
  let hElem = this.hTable.parentNode;
  while (hElem.nodeName != "BODY") {
    hElem = hElem.parentNode;
  }
  hElem.style.cursor = "";

  delete document.dynTable;
  delete this.nCurrentColIndex;
}

DynTable.prototype.setSelectListener = function(aCallback) {
  this.addSelectEventListener();
  this.cbItemSelection = aCallback;
}

DynTable.prototype.addSelectEventListener = function() {
  const hTBody = this.getTableBody();

  if (hTBody == null) {
    throw "Table element <tbody> not found";
  }

  hTBody.addEventListener("click", DynTable.prototype.onRowClick, false);
}

DynTable.prototype.onRowClick = function(evtMouseClick) {
  let hTable = null;
  let index  = -1;

  if (this.nodeName == "TBODY" && evtMouseClick.target != null) {
    // Find the row
    let hTrElem = null;
    if (evtMouseClick.target.nodeName == "TD") {
      hTrElem = evtMouseClick.target.parentNode;
    } else if (evtMouseClick.target.nodeName == "TR") {
      hTrElem = evtMouseClick.target;
    }

    if (hTrElem != null) {
      index  = Array.prototype.indexOf.call(hTrElem.parentNode.children, hTrElem);
      hTable = hTrElem.parentNode;
      while (hTable.nodeName != "TABLE") {
        hTable = hTable.parentNode;
      }
    }
  }

  if (index != -1 && hTable.dynTable.doSelection(evtMouseClick, index)) {
    evtMouseClick.stopImmediatePropagation();
    evtMouseClick.stopPropagation();
    evtMouseClick.preventDefault();
    return false;
  }

  return true;
}

DynTable.prototype.setSelection = function(aIndexes) {
  const hTBody      = this.getTableBody();
  let nChildren     = hTBody.children.length;
  let hTRSelect     = null;
  let aAddedItems   = []
  let aRemovedItems = null;

  // Check parameters
  if (typeof(aIndexes) == "number") {
    aIndexes = [aIndexes];
  } else if (!Array.isArray(aIndexes)) {
    return false;
  } else for (let i=0; i<aIndexes.length; i++) {
    let elem = aIndexes[i];
    if (typeof(elem) != "number") {
      return false;
    } else if (elem < 0 || elem >= nChildren) {
      return false;
    }
  }

  // Remove all selection indicator, to be sure
  for (let i=0; i<nChildren; i++) {
    let hTrElem = hTBody.children[i];
    hTrElem.classList.remove("table_scroll_contentrow_focus");
    hTrElem.classList.remove("table_scroll_contentrow_selected");
  }

  let nItems = aIndexes.length;
  for (let i=0; i<nItems; i++) {
    let idx = aIndexes[i];
    let hTrElem = hTBody.children[idx];
    hTrElem.classList.add("table_scroll_contentrow_selected");

    let selIdx = this.aSelected.indexOf(idx);
    if (selIdx == -1) {
      aAddedItems.push(idx);
      this.aSelected.splice(selIdx, 1);
    }

    if (i == nItems - 1) {
      this.nIdxLastSelected = idx;
      hTrElem.classList.add("table_scroll_contentrow_focus");
      hTrElem.focus();
    }
  }

  aRemovedItems  = this.aSelected;
  this.aSelected = new Array(nItems);
  let that = this;
  aIndexes.forEach(function(element, idx, array) {
    that.aSelected[idx] = element;
  });

  // Notify the listener, using the callback method
  if (typeof(this.cbItemSelection) == "function") {
    this.cbItemSelection(aAddedItems, aRemovedItems, aIndexes);
  }

  return true;
}

DynTable.prototype.doSelection = function(evt, index) {
  const hTBody      = this.getTableBody();
  let hTRSelect     = null;
  let aAddedItems   = []
  let aRemovedItems = [];
  let aCurrentItems = null;

  if (typeof(index) != "number" || index < 0 || index >= hTBody.children.length) {
    // log.e("DynTable.doSelection: invalid index: " + index.toString(10), evt);
    return false;
  }

  this.clearLastSelectedBorder(hTBody);

  hTRSelect = hTBody.children[index];
  if (evt.ctrlKey) {
    // Control key pressed: add the item to the current selection,
    // if already added, remove it.
    if (hTRSelect) {
      let isAlreadySelected = this.aSelected.indexOf(index);
      if (isAlreadySelected == -1) {
        hTRSelect.classList.add("table_scroll_contentrow_selected");
        this.aSelected.push(index);
        aAddedItems.push(index);
      } else {
        hTRSelect.classList.remove("table_scroll_contentrow_selected");
        this.aSelected.splice(isAlreadySelected, 1);

        aRemovedItems.push(index);
      }
    }
  } else if (evt.shiftKey) {
    // Add a list of items (from the previously selected item)
    // to the current one
    if (this.nIdxLastSelected == -1) {
      this.nIdxLastSelected = 0;
    }
    let start = this.nIdxLastSelected;
    let end   = index;
    if (index < this.nIdxLastSelected) {
      start = index;
      end   = this.nIdxLastSelected;
    }

    for (let i=start; i<=end; i++) {
      let hTrElem = hTBody.children[i];
      hTrElem.classList.add("table_scroll_contentrow_selected");
      if (this.aSelected.indexOf(i) == -1) {
        this.aSelected.push(i);
        aAddedItems.push(i);
      }
    }
  } else {
    // Change the current selection.
    for (let i=0; i < this.aSelected.length; i++) {
      let hTrElem = hTBody.children[this.aSelected[i]];
      hTrElem.classList.remove("table_scroll_contentrow_selected");
    }
    this.aSelected = [ index ];
    aAddedItems    = [ index ];
    hTRSelect.classList.add("table_scroll_contentrow_selected");
  }

  this.nIdxLastSelected = index;

  // Show the border to have a visual indication on the last selected item
  hTRSelect.classList.add("table_scroll_contentrow_focus");

  // Sort the selection, to have them always ordered
  this.aSelected.sort(function(a, b) { return a - b; });

  // Copy the this.aSelected array
  aCurrentItems = new Array(this.aSelected.length);
  this.aSelected.forEach(function(element, idx, array) {
    aCurrentItems[idx] = element;
  });

  // Notify the listener, using the callback method
  if (typeof(this.cbItemSelection) == "function") {
    this.cbItemSelection(aAddedItems, aRemovedItems, aCurrentItems);
  }

  hTRSelect.focus();

  return true;
}

DynTable.prototype.clearLastSelectedBorder = function(hTBody) {
  if (this.nIdxLastSelected >= 0 && this.nIdxLastSelected < hTBody.children.length) {
    // Remove the borders of the previously selected row
    let hTRSelect = hTBody.children[this.nIdxLastSelected];
    hTRSelect.classList.remove("table_scroll_contentrow_focus");
  }
}

DynTable.prototype.clearSelection = function() {
  const hTBody = this.getTableBody();
  let aAddedItems   = []
  let aRemovedItems = null;
  let aCurrentItems = [];

  this.clearLastSelectedBorder(hTBody);

  for (let i=0; i < this.aSelected.length; i++) {
    let hTrElem = hTBody.children[this.aSelected[i]];
    hTrElem.classList.remove("table_scroll_contentrow_selected");
  }

  aRemovedItems = this.aSelected;

  this.nIdxLastSelected = -1;
  this.aSelected        = [];

  if (this.cbItemSelection != null) {
    this.cbItemSelection(aAddedItems, aRemovedItems, aCurrentItems);
  }
}

DynTable.prototype.resetSelection = function() {
  this.nIdxLastSelected = -1;
  this.aSelected = [];
}

/**
 * When items are removed, update the selection index array,
 * using the class properties. Seems to be the more simple
 * way to update the selection list.
 */
DynTable.prototype.updateSelection = function() {
  let hTBody    = this.getTableBody();
  let nHtrElems = hTBody.children.length;

  this.resetSelection();

  for (let i=0; i<nHtrElems; i++) {
    let hTrElem = hTBody.children[i];
    if (hTrElem.classList.contains("table_scroll_contentrow_selected")) {
      this.aSelected.push(i);
    }
    if (hTrElem.classList.contains("table_scroll_contentrow_focus")) {
      this.nIdxLastSelected = i;
    }
  }

  if (this.nIdxLastSelected == -1 && this.aSelected.length > 0) {
    this.nIdxLastSelected = this.aSelected[this.aSelected.length - 1];
  }
}

DynTable.prototype.tableKeyListener = function(evtKeyPressed) {
  let ok = true;
  if (evtKeyPressed.keyCode == 40 || evtKeyPressed.key == "ArrowDown") {
    if (this.dynTable.nIdxLastSelected < 0) {
      ok = !this.dynTable.doSelection(evtKeyPressed, 0);
    } else {
      ok = !this.dynTable.doSelection(evtKeyPressed, (this.dynTable.nIdxLastSelected + 1));
    }
  } else if (evtKeyPressed.keyCode == 38 || evtKeyPressed.key == "ArrowUp") {
    if (this.dynTable.nIdxLastSelected > 0) {
      ok = !this.dynTable.doSelection(evtKeyPressed, (this.dynTable.nIdxLastSelected - 1));
    } else {
      ok = !this.dynTable.doSelection(evtKeyPressed, 0);
    }
  } else if (evtKeyPressed.ctrlKey && (evtKeyPressed.key == "a" || evtKeyPressed.key == "A")) {
    // Select all
    let hTBody    = this.dynTable.getTableBody();
    let nChildren = hTBody.children.length;
    let aAllIdx   = new Array(nChildren);
    for (let i=0; i<nChildren; i++) {
      aAllIdx[i] = i;
    }
    this.dynTable.setSelection(aAllIdx);
  }
  return ok;
}

DynTable.prototype.enableSortColumn = function(fctSortComparator) {
  if (typeof(fctSortComparator) != "function") {
    throw "Invalid parameter fctSortComparator";
  }
  if (this.aSortColumns != null) {
    return;
  }

  this.fSortComparator = fctSortComparator;
  this.aSortColumns    = new Array(this.nColumns);

  for (let i=0; i<this.nColumns; i++) {
    let hThElem = this.getHeader(i);
    hThElem.addEventListener("click", DynTable.prototype.onSort, false);
    this.aSortColumns[i] = i+1;
  }
}

DynTable.prototype.onSort = function(evtClick) {
  try {
    let dynTable = this.parentNode.parentNode.parentNode.dynTable;
    return dynTable.doSort(evtClick)
  } catch (ex) {
    log.e(ex);
  }
  return true;
}

DynTable.prototype.doSort = function(evtClick) {
  if (this.aSortColumns == null) {
    throw "Sorting is not enabled";
  }

  let hTBody     = this.getTableBody();
  let hTHead     = this.getTableHeader();
  let hTrElem    = hTHead.firstElementChild;
  let hThElem    = (evtClick == null) ? hTrElem.firstElementChild : evtClick.target;
  let nItems     = hTBody.children.length;
  let sortColumn = 1;

  if (nItems < 2) {
    // no item, or only one. No need to do a sort.
    return false;
  }

  if (evtClick != null) {
    sortColumn = Array.prototype.indexOf.call(hTrElem.children, hThElem) + 1;
  }
  if (Math.abs(this.aSortColumns[0]) == sortColumn) {
    // Invert selection only if we had a click on the header
    if (evtClick != null) this.aSortColumns[0] = -(this.aSortColumns[0]);
  } else {
    // Change sort column order
    let newSortColumns = new Array(this.nColumns);
    newSortColumns[0] = sortColumn;
    let j=1;
    for (let i=0; i<this.nColumns; i++) {
      let nColId = Math.abs(this.aSortColumns[i]);
      if (nColId != sortColumn) {
        newSortColumns[j++] = nColId;
      }
      // Remove previous sort indicator
      hTrElem.children[i].style.backgroundImage = "";
    }
    this.aSortColumns = newSortColumns;
  }

  // Get HTML row elements, and remove them from the body:
  let aTrElems = new Array(nItems);
  for (let i=0; i<nItems; i++) {
    let hTrBodyElem = hTBody.firstElementChild;
    aTrElems[i]     = hTrBodyElem;
    hTBody.removeChild(hTrBodyElem);
  }

  // Use the Array.sort() function, using the specified
  // comparator set by DynTable.enableSortColumn
  Array.prototype.sort.call(aTrElems, this.fSortComparator);

  // Add childrens again and update selected item indexes
  this.aSelected = [];
  for (let i=0; i<nItems; i++) {
    let hTrBodyElem = aTrElems[i];
    if (hTrBodyElem.classList.contains("table_scroll_contentrow_selected")) {
      this.aSelected.push(i);
    }
    if (hTrBodyElem.classList.contains("table_scroll_contentrow_focus")) {
      this.nIdxLastSelected = i;
    }
    if ((i % 2) == 0) {
      hTrBodyElem.classList.remove("table_scroll_contentrow_odd");
      hTrBodyElem.classList.add("table_scroll_contentrow_even");
    } else {
      hTrBodyElem.classList.remove("table_scroll_contentrow_even");
      hTrBodyElem.classList.add("table_scroll_contentrow_odd");
    }
    hTBody.appendChild(hTrBodyElem);
  }

  if (this.aSortColumns[0] > 0) {
    hThElem.style.backgroundImage = IMG_SORT_INDICATOR_UP;
  } else {
    hThElem.style.backgroundImage = IMG_SORT_INDICATOR_DOWN;
  }

  return false;
}
