/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Import/Export cookies.
 *
 * ************************************************************************
 */

"use strict";

const TRUE                = "true";
const FALSE               = "false";
const TYPE_COOKIE         = "Cookie";
const TYPE_STORAGE        = "Dom Storage";
const TYPE_LSO            = "LSO";
const LSO_PREFIX_SO       = "sharedobject:";
const LSO_PREFIX_SETTINGS = "settings:";
const EXPORT_LIST         = 0; // Protected list only
const EXPORT_ALL          = 1; // Full items content
const EXPORT_PROT         = 2;
const EXPORT_SEL          = 3;
const EXPORT_PERM         = 4;
const CK_CSV_HEADER       = "host,path,name,value,creationTime,lastAccessed,expires,isDomain,isSecure,isHttpOnly,isSession,isProtected";
const CK_CSV_HEADER_PLIST = "type,host,name,path";
const CK_CSV_HEADER_PERM  = "host,type,capability,capabilityString";
const FILE_FORMATS        = [
                              { ext: "csv",  bundle: "csv_file",  filter: "*.csv"  },
                              { ext: "json", bundle: "json_file", filter: "*.json" }
                            ]; // Supported import/export files formats
const FILE_PREFIX_COOKIES = "cookies-";
const FILE_PREFIX_LIST    = "protection-";
const FILE_PREFIX_PERM    = "permissions-";

var gImportObserver = {
  onImportDone: function() {
    gCookieKeeper.reload(true);
  },

  init: function() {
    Services.obs.addObserver(this, CK_OBS_PCOOKIE_TOPIC, false);
  },

  observe: function(subject, topic, action) {
    if (action == CK_OBS_PCOOKIE_DOM_INSERT) {
      try {
        this.onImportDone();
      } finally {
        Services.obs.removeObserver(this, CK_OBS_PCOOKIE_TOPIC);
      }
    }
  },
};

var gImportExport = {

  onImport: function() {
    var defaultFormat = 0; // CSV
    let bundle        = document.getElementById("bundleCookieKeeper");
    let filePicker    = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);

    function fpCallback(aResult) {
      if (aResult != Ci.nsIFilePicker.returnCancel && filePicker.fileURL) {
        gImportExport._import(filePicker.file.path);
      }
    }

    if (Services.prefs.getPrefType(PREF_EXPORT_FORMAT) == Services.prefs.PREF_STRING) {
      let defaultExtension = Services.prefs.getCharPref(PREF_EXPORT_FORMAT);
      for (let i=0; i<FILE_FORMATS.length; i++) {
        if (defaultExtension == FILE_FORMATS[i].ext) {
          defaultFormat = i;
          i = FILE_FORMATS.length;
        }
      }
    }

    filePicker.appendFilter(bundle.getString(FILE_FORMATS[defaultFormat].bundle), FILE_FORMATS[defaultFormat].filter);
    for (let i=0; i<FILE_FORMATS.length; i++) {
      if (i != defaultFormat) {
        filePicker.appendFilter(bundle.getString(FILE_FORMATS[i].bundle), FILE_FORMATS[i].filter);
      }
    }

    filePicker.init(window, bundle.getString("select_file"), Ci.nsIFilePicker.modeOpen);
    filePicker.open(fpCallback);
    gCookieKeeper._tree.focus();
  },

  onExportProtectionList: function() {
    this._doExport(EXPORT_LIST);
    gCookieKeeper._tree.focus();
  },

  onExportPermissions: function() {
    this._doExport(EXPORT_PERM);
    gCookieKeeper._tree.focus();
  },

  onExportAll: function() {
    this._doExport(EXPORT_ALL);
    gCookieKeeper._tree.focus();
  },

  onExportProtected: function() {
    this._doExport(EXPORT_PROT);
    gCookieKeeper._tree.focus();
  },

  onExportSelected: function() {
    this._doExport(EXPORT_SEL);
    gCookieKeeper._tree.focus();
  },

  _showInfoNotification: function(bundleStr, nImport, nTotal) {
    var bundle          = document.getElementById("bundleCookieKeeper");
    var str             = bundle.getString(bundleStr);
    var notification    = document.createElementNS(XUL_NS, "notification");
    var notificationBox = document.getElementById("notificationBox");

    str = str.replace("%1", nImport.toString(10));
    str = str.replace("%2", nTotal.toString(10));

    notification.setAttribute("type", "info");
    notification.setAttribute("label", str);
    notificationBox.appendChild(notification);
    log.i(str);
  },

  _showErrorNotification: function(bundleStr) {
    var bundle          = document.getElementById("bundleCookieKeeper");
    var str             = bundle.getString(bundleStr);
    var notification    = document.createElementNS(XUL_NS, "notification");
    var notificationBox = document.getElementById("notificationBox");

    notification.setAttribute("type", "critical");
    notification.setAttribute("label", str);
    notificationBox.appendChild(notification);
    log.e(str);
  },

  _import: function(aPath) {
    log.i("import " + aPath);

    try {
      let importFile = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
      importFile.initWithPath(aPath);
      if (!importFile.isFile()) {
        return;
      }

      let istream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
      istream.init(importFile, -1, -1, false);

      let fileStr = NetUtil.readInputStreamToString(istream, istream.available(), {charset: "UTF-8"});
      istream.close();
      log.d("File read size: " + fileStr.length.toString(10));

      if (!this._importJSON(fileStr)) {
        if (!this._importCSV(fileStr)) {
          this._showErrorNotification("notification_import.error.file_unknown");
        }
      }
    } catch (ex) {
      let bundle        = document.getElementById("bundleCookieKeeper");
      let msg           = "";
      let msgError      = "";
      let title         = bundle.getString("import_error");

      switch (ex.result) {
      case Components.results.NS_ERROR_FILE_ACCESS_DENIED:
        msgError = bundle.getString("file_access_denied");
        break;
      case Components.results.NS_ERROR_FILE_IS_LOCKED:
        msgError = bundle.getString("file_is_locked");
        break;
      default:
        msgError = ex.message;
        break;
      }

      msg = bundle.getFormattedString("read_file_failed", [aPath, msgError]);
      log.e(msg);
      Services.prompt.alert(window, title, msg);

      gImportObserver.onImportDone();
      // Be sure to unblock events if something failed.
      gCookiesObserver.blockCookiesEvents     = false;
      gCookiesObserver.blockProtectionsEvents = false;
    }
  },

  _splitCSVLine: function(aLine, aLineNumber) {
    // Use a regex to parse the line?
    var idx    = 0;
    var datas  = [];
    var lineLen = aLine.length;

    let expectColSep = false;

    while (idx < lineLen) {
      if (expectColSep || aLine[idx] == ',') {
        if (expectColSep && aLine[idx] != ',') {
          log.e("Line " + aLineNumber + ":" + (idx+1) + " expected comma, found \'" + aLine[idx] + "\'");
          idx = aLine.length+1;
          datas = [];
        } else if (!expectColSep) {
          // Empty column
          datas.push(null);
          // log.d("Line " + aLineNumber + ":" + (idx+1) + " empty column " + datas.length);
        }
        expectColSep = false;
        idx++;
      } else if (aLine[idx] == '\"') {
        // it is a quote protected column
        // search for the end of this protected string

        let endIdx  = idx + 1;
        let nDQuote = 1;
        let isPcdq  = false; // Is previous char a double quote
        while (endIdx < lineLen && nDQuote > 0) {
          let ch = aLine[endIdx];
          // log.d("Line " + aLineNumber + ":" + (endIdx+1) + ": ch="+ch);
          if (ch == "\"") {
            nDQuote++;
            if (isPcdq) {
              nDQuote -= 2;
              isPcdq   = false;
            } else {
              isPcdq = true;
            }
          } else if (isPcdq) {
            // If it is a valid string normally we have reach the end here
            // The current char must be the comma.
            // log.d("Line " + aLineNumber + ":" + (endIdx+1) + ": isPcdq=true nDQuote="+nDQuote);
            nDQuote -= 2;
            isPcdq   = false;
          }
          endIdx++;
        }

        if ((nDQuote == 0) || (nDQuote == 2 && endIdx == lineLen)) { // && aLine[endIdx-1] == ','
          // Valid string
          endIdx = nDQuote == 0 ? endIdx-1 : endIdx;
          let data = aLine.substring(idx+1, endIdx-1);
          data = data.replace("\\r", "\r", "g");
          data = data.replace("\\n", "\n", "g");
          data = data.replace("\\\\", "\\", "g");
          data = data.replace("\"\"", "\"", "g");
          datas.push(data);
          idx = endIdx;
          expectColSep = true;
          // log.d("Line " + aLineNumber + ": endIdx=" + endIdx + " found quote protected string=\'" + data + "\'");
        } else {
          log.e("Line " + aLineNumber + ": Invalid quote string at pos " + (idx+1));
          // log.d("nDQuote="+nDQuote + " isPcdq="+isPcdq+" endIdx="+endIdx);
          idx = aLine.length+1;
          datas = [];
        }
      } else {
        // not a quote protected column
        let nextComma = aLine.indexOf(',', idx);
        let endIdx    = nextComma > -1 ? nextComma : aLine.length;
        let data      = aLine.substring(idx, endIdx);
        datas.push(data);
        // log.d("Line " + aLineNumber + ":" + (idx+1) + ": found normal column " + datas.length + " \'" + data + "\'");
        idx = endIdx;
        expectColSep = true;
      }
    }
    if (idx == aLine.length && aLine[lineLen-1] == ',') {
      // Last char was a comma, so empty column
      datas.push(null);
    }

    return datas;
  },

  _importCSVProtectedList: function(lines) {
    var nImport = 0;
    var nTotal  = 0;

    for (let cLine=0; cLine<lines.length; cLine++) {
      let lineNumber = (cLine + 1).toString(10);
      let line = lines[cLine];
      if (line.length == 0 || (line.length > 1 && line[0] == '#') || CK_CSV_HEADER_PLIST == line) {
        // skip this line
        continue;
      }
      nTotal++;

      let datas = this._splitCSVLine(line, lineNumber);
      if (datas.length != 4) {
        log.e("Line " + lineNumber + ": Invalid number of columns: expected 4 having " + datas.length, datas);
        continue;
      }
      let added = false;
      if (datas[0] == TYPE_COOKIE) {
        let item = gCookieStorage.makeProtectedObject(datas[1], datas[2], datas[3]);
        added = CookieKeeper.addProtectedItem(item);
      } else if (datas[0] == TYPE_STORAGE) {
        let item = gDomStorage.makeProtectedObject(datas[1], datas[2]);
        added = CookieKeeper.addProtectedItem(item);
      } else if (datas[0] == TYPE_LSO) {
        let file = datas[3];
        let isSettings = false;
        if (file.indexOf(LSO_PREFIX_SETTINGS) == 0) {
          file = file.substr(LSO_PREFIX_SETTINGS.length);
          isSettings = true;
        } else if (file.indexOf(LSO_PREFIX_SO) == 0) {
          file = file.substr(LSO_PREFIX_SO.length);
        } else {
          log.w("Line " + lineNumber + ": Error in LSO file: " + datas[3]);
          continue;
        }
        let item = gLocalSharedObject.makeProtectedObject(datas[1], datas[2], file, isSettings);
        added = CookieKeeper.addProtectedItem(item);
      } else {
        log.w("Line " + lineNumber + ": Unknown cookie type: " + datas[0]);
      }
      if (added) {
        nImport++;
      }
    }

    // Show a user notification, to display the result
    this._showInfoNotification("notification_import.info.plist", nImport, nTotal);
    return true;
  },

  _importCSVCookieList: function(lines) {
    var dStorage    = [];
    var cm          = Services.cookies; // Cookie Manager
    var nImport     = 0;
    var nTotal      = 0;
    var regexNumber = new RegExp("^[0-9]+$");
    var regexBool   = new RegExp("^(true|false)$");

    gCookiesObserver.blockCookiesEvents     = true;
    gCookiesObserver.blockProtectionsEvents = true;

    for (let cLine=0; cLine<lines.length; cLine++) {
      let lineNumber = (cLine + 1).toString(10);
      let line = lines[cLine];
      if (line.length == 0 || (line.length > 1 && line[0] == '#') || CK_CSV_HEADER == line) {
        // skip this line
        log.i("Line " + lineNumber + ": empty, comment or header");
        continue;
      }
      nTotal++;
      let datas = this._splitCSVLine(line, lineNumber);
      if (datas.length != 12) {
        log.e("Line " + lineNumber + ": Invalid number of columns: expected 12 having " + datas.length, datas);
        continue;
      }

      let path        = datas[1];
      let isProtected = (TRUE == datas[11]);

      if (path != null && path.length > 0) {
        // It is a cookie
        let host         = datas[0];
        let name         = datas[2];
        let value        = datas[3];
        let sCreation    = datas[4];
        let sAccessed    = datas[5];
        let sDate        = datas[6]; // Expire date (string)
        let sIsDomain    = datas[7];
        let sIsSecure    = datas[8];
        let sIsHttpOnly  = datas[9];
        let sIsSession   = datas[10];
        let sIsProtected = datas[11];

        // Check datas:
        if (host == null || host.length == 0) {
          log.e("Line " + lineNumber + ": Invalid cookie host: empty");
          continue;
        }
        if (name == null || name.length == 0) {
          log.e("Line " + lineNumber + ": Invalid cookie name: empty");
          continue;
        }
        if (!regexNumber.test(sCreation)) {
          log.e("Line " + lineNumber + ": Invalid cookie creation date: not a number");
          continue;
        }
        if (!regexNumber.test(sAccessed)) {
          log.e("Line " + lineNumber + ": Invalid cookie last accessed date: not a number");
          continue;
        }
        if (!regexNumber.test(sDate)) {
          log.e("Line " + lineNumber + ": Invalid cookie expires date: not a number");
          continue;
        }
        if (!regexBool.test(sIsDomain)) {
          log.e("Line " + lineNumber + ": Invalid isDomain value: \'" + sIsDomain + "\', expected 'true' or 'false'");
          continue;
        }
        if (!regexBool.test(sIsSecure)) {
          log.e("Line " + lineNumber + ": Invalid isSecure value: \'" + sIsSecure + "\', expected 'true' or 'false'");
          continue;
        }
        if (!regexBool.test(sIsHttpOnly)) {
          log.e("Line " + lineNumber + ": Invalid isHttpOnly value: \'" + sIsHttpOnly + "\', expected 'true' or 'false'");
          continue;
        }
        if (!regexBool.test(sIsSession)) {
          log.e("Line " + lineNumber + ": Invalid isSession value: \'" + sIsSession + "\', expected 'true' or 'false'");
          continue;
        }
        if (!regexBool.test(sIsProtected)) {
          log.e("Line " + lineNumber + ": Invalid isProtected value: \'" + sIsProtected + "\', expected 'true' or 'false'");
          continue;
        }

        let isSecure   = (TRUE == sIsSecure);
        let isHttpOnly = (TRUE == sIsHttpOnly);
        let isSession  = (TRUE == sIsSession);
        // Convert expires date to a number
        let iDate = parseInt(sDate, 10);
        if (sDate.endsWith("000") && iDate != DATE_MAX_SECONDS) {
          // Workaround a previous bug #1297486
          iDate = (iDate/1000);
        } else if (iDate == 0 || isSession) {
          // Some session cookies have 0 for their expire date (previous export bug)
          iDate = DATE_MAX_SECONDS;
        }
        cm.add(host, path, name, value, isSecure, isHttpOnly, isSession, iDate);

        if (isProtected) {
          let cookie = gCookieStorage.makeProtectedObject(host, name, path);
          if (!CookieKeeper.isProtectedItem(cookie)) {
            // Add only if it is not already protected
            CookieKeeper.addProtectedItem(cookie);
          }
        }
      } else {
        // It is a dom storage
        let tmp   = datas[0];
        let key   = datas[2];
        let value = datas[3];
        let sarr  = tmp.split(":"); // scope array
        let scope = "";
        for (let i=0; i<sarr.length; i++) {
          if (i==0) {
            scope = gCommon.reverseStr(sarr[0]);
          } else {
            scope = scope + ":" + sarr[i];
          }
        }
        if (scope.length == 0) {
          log.e("Line " + lineNumber + ": host (scope) value is empty");
          continue;
        }
        if (key == null || key.length == 0) {
          log.e("Line " + lineNumber + ": name (key) is empty");
          continue;
        }
        let storage = gDomStorage.makeDomStorageObject(scope, key, value, null, null);
        dStorage.push(storage);

        if (isProtected && !CookieKeeper.isProtectedItem(storage)) {
          // Add only if it is not already protected
          CookieKeeper.addProtectedItem(storage);
        }
      }

      nImport++;
    }

    if (dStorage.length > 0) {
      gImportObserver.init();
      gDomStorage.performInsertion(dStorage);
    } else {
      gImportObserver.onImportDone();
    }

    // Show a user notification, to display the result
    this._showInfoNotification("notification_import.info.cookies", nImport, nTotal);
    return true;
  },

  _importCSVPermissionsList: function(lines) {
    var nImport     = 0;
    var nTotal      = 0;
    var regexNumber = new RegExp("^[0-9]+$");
    var ioService   = Services.io;
    var permSrv     = Services.perms;

    for (let cLine=0; cLine<lines.length; cLine++) {
      let line = lines[cLine];
      let lineNumber = (cLine + 1).toString(10);
      if (line.length == 0 || (line.length > 1 && line[0] == '#') || line === CK_CSV_HEADER_PERM) {
        // skip this line
        continue;
      }
      nTotal++;
      let datas = this._splitCSVLine(line, lineNumber);
      if (datas.length != 4) {
        log.e("Line " + lineNumber + ": Invalid number of columns: expected 4 having " + datas.length, datas);
        continue;
      }
      let host   = datas[0];
      let type   = datas[1];
      let capStr = datas[2];
      if (!regexNumber.test(capStr)) {
        log.e("Line " + lineNumber + ": Capability is not a number");
        continue;
      }
      let capability = parseInt(capStr, 10);

      let uri = ioService.newURI("http://" + host, null, null);
      permSrv.add(uri, type, capability);
      nImport++;
    }

    this._showInfoNotification("notification_import.info.permissions", nImport, nTotal);
    return true;
  },

  _importCSV: function(fileStr) {
    var lines    = fileStr.split("\n");
    var fileType = -1;

    for (let cLine=0; cLine<lines.length && fileType == -1; cLine++) {
      let line = lines[cLine];
      if (line.length == 0 || (line.length > 1 && line[0] == '#')) {
        // skip this line
        continue;
      }

      if (CK_CSV_HEADER == line) {
        fileType = EXPORT_ALL;
      } else if (CK_CSV_HEADER_PLIST == line) {
        fileType = EXPORT_LIST;
      } else if (CK_CSV_HEADER_PERM == line) {
        fileType = EXPORT_PERM;
      } else {
        log.e("Unknown CSV file type");
        fileType = -2;
      }
    }

    switch (fileType) {
    case EXPORT_ALL:
      return this._importCSVCookieList(lines);
    case EXPORT_LIST:
      return this._importCSVProtectedList(lines);
    case EXPORT_PERM:
      return this._importCSVPermissionsList(lines);
    }
    log.w("Import: This CSV file is not valid, header not found");
    return false;
  },

  _importJSON: function(fileStr) {
    var nImport = 0;
    var nTotal  = 0;

    try {
      let json = JSON.parse(fileStr);
      let typeOfArray = typeof([]); // "object" not "array" :-(
      let fileKnown   = false;

      if (typeof(json.method) == typeof(EXPORT_LIST)) {
        if (json.method == EXPORT_PERM && typeof(json.perms) == typeOfArray && typeof(json.perms.length) == typeof(0)) {
          fileKnown = true;
        } else if (json.method >= EXPORT_LIST      && json.method <= EXPORT_SEL &&
          typeof(json.cookies) == typeOfArray      && typeof(json.storage) == typeOfArray &&
          typeof(json.cookies.length) == typeof(0) && typeof(json.storage.length) == typeof(0)) {
          fileKnown = true;
        }
      }
      if (!fileKnown) {
        log.d("_importJSON fileKnown=false");
        this._showErrorNotification("notification_import.error.file_unknown");
        return true;
      }

      if (json.method == EXPORT_LIST) {
        // Protection list
        if (json.cookies) {
          nTotal = json.cookies.length;

          for (let i=0; i<json.cookies.length; i++) {
            let item = json.cookies[i];
            item = gCookieStorage.makeProtectedObject(item.host, item.name, item.path);
            if (CookieKeeper.addProtectedItem(item)) {
              nImport++;
            }
          }
        }

        if (json.storage) {
          nTotal += json.storage.length;

          for (let i=0; i<json.storage.length; i++) {
            let item = json.storage[i];
            item = gDomStorage.makeProtectedObject(item.scope, item.key);
            if (CookieKeeper.addProtectedItem(item)) {
              nImport++;
            }
          }
        }

        if (json.lso) {
          nTotal += json.lso.length;

          for (let i=0; i<json.lso.length; i++) {
            let item = json.lso[i];
            item = gLocalSharedObject.makeProtectedObject(item.host, item.name, item.file, item.isSettings);
            if (CookieKeeper.addProtectedItem(item)) {
              nImport++;
            }
          }
        }

        this._showInfoNotification("notification_import.info.plist", nImport, nTotal);
      } else if (json.method == EXPORT_PERM) {
        // Permissions list
        let ioService = Services.io;
        let permSrv   = Services.perms;
        nTotal = json.perms.length;

        for (let i=0; i<nTotal; i++) {
          let item = json.perms[i];
          let uri = ioService.newURI("http://" + item.host, null, null);
          if (typeof(item.capability) != "number") {
            log.e("Item " + i.toString(10) + ": Capability is not a number", item);
            continue;
          }
          permSrv.add(uri, item.type, item.capability);
          nImport++;
        }

        this._showInfoNotification("notification_import.info.permissions", nImport, nTotal);
      } else {
        // Cookies & DOM Storage
        gCookiesObserver.blockCookiesEvents     = true;
        gCookiesObserver.blockProtectionsEvents = true;

        nTotal = json.cookies.length + json.storage.length;
        let cm = Services.cookies; // Cookie Manager
        for (let i=0; i<json.cookies.length; i++) {
          let item = json.cookies[i];
          cm.add(item.host, item.path, item.name, item.value, item.isSecure, item.isHttpOnly, item.isSession, item.expires);
          if (item.isProtected) {
            CookieKeeper.addProtectedItem(item);
          }
          nImport++;
        }
        for (let i=0; i<json.storage.length; i++) {
          let item = json.storage[i];
          if (item.isProtected) {
            CookieKeeper.addProtectedItem(item);
          }
          nImport++;
        }

        if (json.storage.length == 0) {
          gImportObserver.onImportDone();
        } else {
          gImportObserver.init();
          gDomStorage.performInsertion(json.storage);
        }

        this._showInfoNotification("notification_import.info.cookies", nImport, nTotal);
      }

      return true;
    } catch (ex) {
      log.i("Import: Not a JSON file", ex);
      return false;
    }
  },

  int2str: function(aNumber, aPrecision) {
    let s = aNumber.toString(10);
    while (s.length < aPrecision) {
      s = "0" + s;
    }

    return s;
  },

  _toCSVString: function(aString) {
    let value = aString.toString().replace("\\", "\\\\", "g");
    value = value.replace("\n", "\\n",  "g"); // Some JSON contents may have a \n char
    value = value.replace("\r", "\\r",  "g");
    value = value.replace("\"", "\"\"", "g");
    value = "\"" + value + "\"";

    return value;
  },

  _doExport: function(aMethod) {
    var bundle        = document.getElementById("bundleCookieKeeper");
    var defaultFormat = 0; // CSV
    var filtersIdx    = [];

    function askChangeExtension(_filePath, extension) {
      let filePath    = _filePath;
      let doNotAsk    = false; // Default to false
      let doChangeExt = true;  // Default to true

      if (Services.prefs.getPrefType(PREF_NO_ASK_AGAIN_CHEXT) == Services.prefs.PREF_BOOL) {
        doNotAsk = Services.prefs.getBoolPref(PREF_NO_ASK_AGAIN_CHEXT);
      }
      if (Services.prefs.getPrefType(PREF_EXPORT_CHANGE_EXT) == Services.prefs.PREF_BOOL) {
        doChangeExt = Services.prefs.getBoolPref(PREF_EXPORT_CHANGE_EXT);
      }

      if (!filePath.endsWith(extension) && (doChangeExt || !doNotAsk)) {
        let lastSlash_p1 = (filePath.lastIndexOf("/")+1)|0;
        let filename     = filePath.substring(lastSlash_p1);
        let pathname     = filePath.substring(0, lastSlash_p1);
        let baseFilename = filename;

        if (filename.lastIndexOf(".") >= 1) {
          let lastDot_m1 = (filename.lastIndexOf("."))|0;
          baseFilename   = filename.substring(0, lastDot_m1);
        }
        let proposedFilename = baseFilename + extension;

        let result = true;
        if (!doNotAsk) {
          let prompt    = Services.prompt;
          let title      = bundle.getString("question_export_title");
          let question   = bundle.getString("question_change_extension");
          let noAskAgain = bundle.getString("no_ask_again");
          let check      = {value: false}; // default the checkbox to false
          let btnsYesNo  = (prompt.BUTTON_POS_0 * prompt.BUTTON_TITLE_YES) +
                           (prompt.BUTTON_POS_1 * prompt.BUTTON_TITLE_NO)  +
                           prompt.BUTTON_POS_0_DEFAULT;

          question = question.replace("%1", extension);
          question = question.replace("%2", filename);
          question = question.replace("%3", proposedFilename);
          result   = prompt.confirmEx(window, title, question, btnsYesNo, "", "", "", noAskAgain, check) == 0;
          if (check.value) {
            // Save pref
            Services.prefs.setBoolPref(PREF_NO_ASK_AGAIN_CHEXT, true);
            Services.prefs.setBoolPref(PREF_EXPORT_CHANGE_EXT, result);
          }
        }

        if (result) {
          filePath = pathname + proposedFilename;
        }
      }
      return filePath;
    }

    function fpCallback(aResult) {
      if (aResult != Ci.nsIFilePicker.returnCancel && filePicker.fileURL) {
        let filePath = filePicker.file.path;
        let format   = filtersIdx[filePicker.filterIndex];

        filePath = askChangeExtension(filePath, "." + FILE_FORMATS[format].ext);
        gImportExport._export(filePath, aMethod, FILE_FORMATS[format]);
        Services.prefs.setCharPref(PREF_EXPORT_FORMAT, FILE_FORMATS[format].ext);
      }
    }

    if (Services.prefs.getPrefType(PREF_EXPORT_FORMAT) == Services.prefs.PREF_STRING) {
      let defaultExtension = Services.prefs.getCharPref(PREF_EXPORT_FORMAT);
      for (let i=0; i<FILE_FORMATS.length; i++) {
        if (defaultExtension == FILE_FORMATS[i].ext) {
          defaultFormat = i;
          i = FILE_FORMATS.length;
        }
      }
    }

    var filePicker = Cc["@mozilla.org/filepicker;1"].createInstance(Ci.nsIFilePicker);

    var date = new Date();
    var date_str =
      this.int2str(date.getFullYear(),    4) +
      this.int2str((date.getMonth() + 1), 2) +
      this.int2str(date.getDate(),        2) +
      "_"                                    +
      this.int2str(date.getHours(),       2) +
      this.int2str(date.getMinutes(),     2) +
      this.int2str(date.getSeconds(),     2);

    filePicker.appendFilter(bundle.getString(FILE_FORMATS[defaultFormat].bundle), FILE_FORMATS[defaultFormat].filter);
    filtersIdx.push(defaultFormat);
    for (let i=0; i<FILE_FORMATS.length; i++) {
      if (i != defaultFormat) {
        filePicker.appendFilter(bundle.getString(FILE_FORMATS[i].bundle), FILE_FORMATS[i].filter);
        filtersIdx.push(i);
      }
    }
    if (aMethod == EXPORT_LIST) {
      filePicker.defaultString  = FILE_PREFIX_LIST    + date_str + "." + FILE_FORMATS[defaultFormat].ext;
    } else if (aMethod == EXPORT_ALL || aMethod == EXPORT_SEL || aMethod == EXPORT_PROT) {
      filePicker.defaultString  = FILE_PREFIX_COOKIES + date_str + "." + FILE_FORMATS[defaultFormat].ext;
    } else if (aMethod == EXPORT_PERM) {
      filePicker.defaultString  = FILE_PREFIX_PERM + date_str + "." + FILE_FORMATS[defaultFormat].ext;
    }
    filePicker.defaultExtension = FILE_FORMATS[defaultFormat].ext;
    filePicker.filterIndex      = 0;
    filePicker.init(window, bundle.getString("select_file"), Ci.nsIFilePicker.modeSave);
    filePicker.open(fpCallback);
  },

  _export: function(aPath, aMethod, aFormat) {
    let converter = null;
    log.i("Export to " + aPath);

    try {
      let exportFile = Cc["@mozilla.org/file/local;1"].createInstance(Ci.nsIFile);
      exportFile.initWithPath(aPath);
      if (exportFile.exists()) {
        exportFile.remove(false);
      }
      exportFile.create(Ci.nsIFile.NORMAL_FILE_TYPE, parseInt("600", 8)); // Unix file perm: -rw-------

      let ostream = FileUtils.openFileOutputStream(exportFile, FileUtils.MODE_WRONLY | FileUtils.MODE_CREATE);
      converter = Cc["@mozilla.org/intl/converter-output-stream;1"].createInstance(Ci.nsIConverterOutputStream);
      converter.init(ostream, "UTF-8", 0, 0);

      if (aFormat.ext == "csv") {
        this._exportCSV(aMethod, converter);
      } else {
        this._exportJSON(aMethod, converter);
      }
    } catch (ex) {
      let bundle        = document.getElementById("bundleCookieKeeper");
      let msgError      = "";
      let msg           = "";
      let title         = bundle.getString("export_error");

      switch (ex.result) {
      case Components.results.NS_ERROR_FILE_ACCESS_DENIED:
        msgError = bundle.getString("file_access_denied");
        break;
      case Components.results.NS_ERROR_FILE_IS_LOCKED:
        msgError = bundle.getString("file_is_locked");
        break;
      case Components.results.NS_ERROR_FILE_NO_DEVICE_SPACE:
      case Components.results.NS_ERROR_FILE_DISK_FULL:
        msgError = bundle.getString("write_file_no_device_space");
        break;
      default:
        msgError = ex.message;
        break;
      }

      msg = bundle.getFormattedString("write_file_failed", [aPath, msgError]);
      log.e(msg);
      Services.prompt.alert(window, title, msg);
    } finally {
      if (converter != null) {
        converter.close();
      }
    }
  },

  _getItems: function(aMethod) {
    var items = [];
    if (aMethod == EXPORT_SEL) {
      // Export the current selection
      let seln = gCookiesTreeView.selection;

      if (gCookiesTreeView.viewMode == VIEW_MODE.TREE) {
        let ci   = seln.currentIndex;
        let item = ci == -1 ? null : gCookiesTreeView.getItemAtIndex(ci);
        if (item != null && item.getType() == COOKIE_TYPE_CONTAINER) {
          let nItems = item.items.length;
          for (let cItem=0; cItem<nItems; cItem++) {
            items.push(item.items[cItem]);
          }
        } else if (item != null) {
          items.push(item);
        }
      } else { // gCookiesTreeView.viewMode == VIEW_MODE.FLAT
        let rangeCount = seln.getRangeCount();
        for (let i=0; i<rangeCount; i++) {
          let min = {};
          let max = {};
          seln.getRangeAt(i, min, max);
          for (let j = min.value; j <= max.value; j++) {
            let item = gCookiesTreeView.getItemAtIndex(j);
            if (item == null) {
              continue;
            }
            if (item.getType() == COOKIE_TYPE_CONTAINER) {
              let nItems = item.items.length;
              for (let cItem=0; cItem<nItems; cItem++) {
                items.push(item.items[cItem]);
              }
            } else {
              items.push(item);
            }
          }
        }
      }

    } else {
      // Export all/Export protected
      for (let key in gCookiesTreeView._hosts) {
        let host = gCookiesTreeView._hosts[key];
        let nItems = host.items.length;
        for (let cItem=0; cItem<nItems; cItem++) {
          let item = host.items[cItem];
          if (aMethod == EXPORT_ALL || item.isProtected) {
            items.push(item);
          }
        }
      }
    }

    return items;
  },

  _getPermissions: function() {
    var pm    = Cc["@mozilla.org/permissionmanager;1"].getService(Ci.nsIPermissionManager);
    var pEnum = pm.enumerator;
    var list  = [];

    while (pEnum.hasMoreElements()) {
      let mozPerm = pEnum.getNext().QueryInterface(Ci.nsIPermission);
      let cString = "";

      switch (mozPerm.capability) {
      case MOZ_COOKIE_PERM_ALLOW:
        cString = "allow";
        break;
      case MOZ_COOKIE_PERM_DENY:
        cString = "deny";
        break;
      case MOZ_COOKIE_PERM_FIRST:
        cString = "canAccessFirstParty";
        break;
      case MOZ_COOKIE_PERM_SESSION:
        cString = "session";
        break;
      }

      if (mozPerm.type == "cookie") {
        let perm = {
          host            : mozPerm.host,
          type            : mozPerm.type,
          capability      : mozPerm.capability,
          capabilityString: cString
        };
        list.push(perm);
      }
    }

    return list;
  },

  _getExportCookieLine: function(aCookie) {
    if (aCookie == null) {
      return null;
    }
    // deal with [,\n\r] chars in the string
    let value = this._toCSVString(aCookie.value);

    // creationTime and lastAccessed are in microseconds since midnight (00:00:00), January 1, 1970 UTC.
    // expires and expiry are in seconds since midnight (00:00:00), January 1, 1970 UTC...
    // Export everything in seconds.
    let creationDate = Math.floor(aCookie.creationTime / 1000).toString(10);
    let accessedDate = Math.floor(aCookie.lastAccessed / 1000).toString(10);
    let expires      = (aCookie.isSession ? aCookie.expiry : aCookie.expires);
    let expiresDate  = Math.floor(expires/1000).toString(10);

    let str = aCookie.host   +
          "," + aCookie.path +
          "," + aCookie.name +
          "," + value        +
          "," + creationDate +
          "," + accessedDate +
          "," + expiresDate  +
          "," + (aCookie.isDomain    ? TRUE : FALSE) +
          "," + (aCookie.isSecure    ? TRUE : FALSE) +
          "," + (aCookie.isHttpOnly  ? TRUE : FALSE) +
          "," + (aCookie.isSession   ? TRUE : FALSE) +
          "," + (aCookie.isProtected ? TRUE : FALSE) +
          "\n";
    return str;
  },

  _getExportStorageLine: function(aItem) {
    if (aItem == null) {
      return null;
    }
    let key   = this._toCSVString(aItem.key);
    let value = this._toCSVString(aItem.value);
    let scope = aItem.host;

    if (aItem.proto.length > 0) {
      scope = scope + ":" + aItem.proto;
    }
    if (aItem.port.length > 0) {
      scope = scope + ":" + aItem.port;
    }
    scope = this._toCSVString(scope);

    let str = scope +
          "," + // Path
          "," + key   +
          "," + value +
          "," + // Creation date
          "," + // Accessed date
          "," + // Expire date
          "," + FALSE +
          "," + (aItem.isSecure     ? TRUE : FALSE) +
          "," + (aItem.isHttpOnly   ? TRUE : FALSE) +
          "," + FALSE +
          "," + (aItem.isProtected  ? TRUE : FALSE) +
          "\n";
    return str;
  },

  _exportCSV: function(aMethod, aConverter) {
    let nWritten = 0;
    let nExport  = 0;
    let nTotal   = 0;

    if (aMethod == EXPORT_LIST) {
      // Export only the protection list
      let pItems = CookieKeeper.getProtectedItems();

      nWritten += aConverter.writeString(CK_CSV_HEADER_PLIST);
      nWritten += aConverter.writeString("\n");

      if (pItems.cookies) {
        let items = pItems.cookies;
        let il    = items.length;

        nTotal += il;

        for (let i=0; i<il; i++) {
          let cookie = items[i];
          nWritten += aConverter.writeString(TYPE_COOKIE + "," + cookie.getHost() + "," + cookie.getName() + "," + cookie.getPath() + "\n");
          nExport++;
        }
      }

      if (pItems.storage) {
        let items = pItems.storage;
        let il    = items.length;

        nTotal += il;

        for (let i=0; i<il; i++) {
          let dom   = items[i];
          let scope = this._toCSVString(dom.scope);
          let key   = this._toCSVString(dom.key);
          nWritten += aConverter.writeString(TYPE_STORAGE + "," + scope + "," + key + ",\n");
          nExport++;
        }
      }

      if (pItems.lso) {
        let items = pItems.lso;
        let il    = items.length;

        nTotal += il;

        for (let i=0; i<il; i++) {
          let lso = items[i];
          let path;
          if (lso.isSettings) {
            path = LSO_PREFIX_SETTINGS + lso.getPath();
          } else {
            path = LSO_PREFIX_SO + lso.getPath();
          }
          nWritten += aConverter.writeString(TYPE_LSO + "," + lso.getHost() + "," + lso.getName() + "," + path + "\n");
          nExport++;
        }
      }

      this._showInfoNotification("notification_export.info.plist", nExport, nTotal);
    } else if (aMethod == EXPORT_ALL || aMethod == EXPORT_SEL || aMethod == EXPORT_PROT) {
      // Export full items
      let items   = this._getItems(aMethod);
      let il      = items.length;
      nTotal += il;

      nWritten += aConverter.writeString(CK_CSV_HEADER);
      nWritten += aConverter.writeString("\n");

      for (let i=0; i<il; i++) {
        let item = items[i];
        let str = null;

        switch (item.getType()) {
        case COOKIE_TYPE_STANDARD:
          str = this._getExportCookieLine(item);
          break;
        case COOKIE_TYPE_STORAGE:
          str = this._getExportStorageLine(item);
          break;
        case COOKIE_TYPE_LSO:
          break; // ignore
        default:
          log.e("Unknown cookie type " + item.getType());
          break;
        }

        if (str != null) {
          nWritten += aConverter.writeString(str);
          nExport++;
        } else {
          log.i("Export: Ignore item ", item);
        }
      }
      this._showInfoNotification("notification_export.info.cookies", nExport, nTotal);
    } else if (aMethod == EXPORT_PERM) {
      // Export permissions
      let perms = this._getPermissions();

      nTotal    = perms.length;
      nWritten += aConverter.writeString(CK_CSV_HEADER_PERM);
      nWritten += aConverter.writeString("\n");

      for (let i=0; i<nTotal; i++) {
        let perm = perms[i];
        nWritten += aConverter.writeString(perm.host + "," + perm.type + "," + perm.capability + "," + perm.capabilityString + "\n");
        nExport++;
      }

      this._showInfoNotification("notification_export.info.permissions", nExport, nTotal);
    }

    log.i(nWritten.toString(10) + " lines written");
  },

  _exportJSON: function(aMethod, aConverter) {
    let items     = null;
    let itemsJSON = null;
    let nExport  = 0;
    let nTotal   = 0;

    if (aMethod == EXPORT_LIST) {
      // Export only the protection list
      items = CookieKeeper.getProtectedItems();
      items.method = aMethod;
      items.date   = Date.now();

      nTotal = items.cookies.length + items.storage.length + items.lso.length;
      nExport = nTotal;
    } else if (aMethod == EXPORT_ALL || aMethod == EXPORT_SEL || aMethod == EXPORT_PROT) {
      // Export full items
      let tmpItems  = this._getItems(aMethod);
      let nTmpItems = tmpItems.length|0;
      items         = {method: aMethod, date: Date.now(), cookies:[], storage: []};
      for (let i=0; i<nTmpItems; i++) {
        let item = tmpItems[i];
        nTotal++;
        switch (item.getType()) {
        case COOKIE_TYPE_STANDARD: {
            // Export expires date in seconds
            let expires      = (item.isSession ? item.expiry : item.expires);
            let expiresDate  = Math.floor(expires/1000);
            let cookie = {
              host        : item.host,
              path        : item.path,
              name        : item.name,
              value       : item.value,
              isSecure    : item.isSecure,
              isHttpOnly  : item.isHttpOnly,
              isSession   : item.isSession,
              expires     : expiresDate,
              isProtected : item.isProtected
            };
            items.cookies.push(cookie);
          }
          nExport++;
          break;
        case COOKIE_TYPE_STORAGE: {
            let dom = {
              scope       : item.scope,
              key         : item.key,
              value       : item.value,
              isProtected : item.isProtected
            };
            items.storage.push(dom);
          }
          nExport++;
          break;
        case COOKIE_TYPE_LSO:
          // ignore
          break;
        default:
          log.e("Unknown cookie type " + item.getType());
          break;
        }
      }
    } else if (aMethod == EXPORT_PERM) {
      // Export permissions
      let perms = this._getPermissions();
      nTotal    = perms.length;
      nExport   = nTotal;
      items = {method: aMethod, date: Date.now(), perms: perms};
    }
    itemsJSON = JSON.stringify(items);
    aConverter.writeString(itemsJSON);

    if (aMethod == EXPORT_LIST) {
      this._showInfoNotification("notification_export.info.plist", nExport, nTotal);
    } else if (aMethod == EXPORT_PERM) {
      this._showInfoNotification("notification_export.info.permissions", nExport, nTotal);
    } else {
      this._showInfoNotification("notification_export.info.cookies", nExport, nTotal);
    }

    items     = null;
    itemsJSON = null;
  }
};
