/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2015 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Local shared object (LSOs) / Flash cookies.
 *
 * Thanks to:
 * - The Mozilla add-on "BetterPrivacy" for the code examples!
 * - GNU Gnash (gnash/cygnal/libamf/sol.cpp)
 * - AMF specifications: https://www.adobe.com/content/dam/Adobe/en/devnet/amf/pdf/amf-file-format-spec.pdf
 *
 * ************************************************************************
 */

"use strict";

const LSO_PRERAND_DIR       = "#SharedObjects";
const LSO_REGEX_SOLEXT      = /.*\.sol$/;
const LSO_REGEX_SPLITPATH   = /.*#$/;
const LSO_REGEX_WINSEP      = /\\/g;
const LSO_KEY_MAIN_SETTINGS = "<flash.main_setting>"; // special host, see gCommon.isSpecialHost() && cookiekeeper.properties

var gSolReader = {
  /*
  Sol file format:
  - Magic number: 2 bytes (0x00BF)
  - Length      : 4 bytes (the length of the file including the Marker bytes)
  - Marker      : 10 bytes (always "TCSO0x000400000000")
  - Object Name : variable (the name of the object as an AMF encoded string)
  - Padding     : 4 bytes (the padding is not always 0x00000000 for AMF0, but 0x00000003 for AMF3?)
                  It seems to also be 0x00000000 for settings files, and 0x00000003 for others files.
  - After this is a series of AMF objects
  */
  solToJson: function(aSolFile) {
    const SOL_MAGIC = 0x00BF;
    const SOL_TCSO = 0x5443534F; // "TCSO"
    const SOL_BLOCK_MARK = 0x0004;
    var binIStream = null;
    var objectRet = {};

    if (typeof(aSolFile) != "object") {
      log.e("solToJson: invalid parameter ", aSolFile);
      return "Error: solToJson: invalid parameter";
    }

    try {
      if (aSolFile.exists() && aSolFile.isFile()) {
        let fistream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
        fistream.init(aSolFile, -1, -1, false);
        binIStream = Cc["@mozilla.org/binaryinputstream;1"].createInstance(Ci.nsIBinaryInputStream);
        binIStream.setInputStream(fistream);

        objectRet.fileInfos = {
          filename         : aSolFile.path,
          fileSize         : aSolFile.fileSize,
          lastModifiedTime : gCommon.formatDateString(aSolFile.lastModifiedTime, "unknown")
        };

        let magic = binIStream.read16();
        if (magic != SOL_MAGIC) {
          throw new Error("Not a SOL file, expected magic " + SOL_MAGIC + " having:" + magic);
        }
        let length = binIStream.read32();
        if (length + 6 != aSolFile.fileSize) {
          throw new Error("SOL file may be corrupted, expected file size " + (length + 6) + " having " + aSolFile.fileSize);
        }
        let tcso = binIStream.read32();
        if (tcso != SOL_TCSO) {
          throw new Error("SOL file may be corrupted, expected " + SOL_TCSO + " (string \"TCSO\") having: " + tcso);
        }
        let blockMark = binIStream.read16();
        if (blockMark != SOL_BLOCK_MARK) {
          throw new Error("SOL file may be corrupted, expected block mark " + SOL_BLOCK_MARK + " having:" + blockMark);
        }
        let endBlockMark = binIStream.read32();
        if (endBlockMark != 0) {
          log.w("endBlockMark != 0");
        }
        let objectNameLength = binIStream.read16();
        objectRet.name = binIStream.readBytes(objectNameLength);

        let fileFormat = binIStream.read32();

        // Read the AMF part:
        if (fileFormat == 0) {
          objectRet.format = "AMF0";
          objectRet.content = this.readAMF0(binIStream);
        } else if (fileFormat == 3) {
          objectRet.format = "AMF3";
          objectRet.content = this.readAMF3(binIStream);
        } else {
          objectRet.content = "Unknown format";
        }
      } else {
        log.e("File " + aSolFile.path + " is not readable");
        objectRet.readError = "File " + aSolFile.path + " is not readable";
      }

    } catch (ex) {
      objectRet.readError = ex.message;
      objectRet.readErrorStack = ex.stack;
      log.e(ex);
    }

    if (binIStream != null) {
      binIStream.close();
    }

    return JSON.stringify(objectRet);
  },

  readAMF0: function(binIStream) {
    const NUMBER_MARKER         = 0x00;
    const BOOLEAN_MARKER        = 0x01;
    const STRING_MARKER         = 0x02;
    const OBJECT_MARKER         = 0x03;
    const MOVIECLIP_MARKER      = 0x04; // reserved, not supported
    const NULL_MARKER           = 0x05;
    const UNDEFINED_MARKER      = 0x06;
    const REFERENCE_MARKER      = 0x07;
    const ECMA_ARRAY_MARKER     = 0x08;
    const OBJECT_END_MARKER     = 0x09;
    const STRICT_ARRAY_MARKER   = 0x0A;
    const DATE_MARKER           = 0x0B;
    const LONG_STRING_MARKER    = 0x0C;
    const UNSUPPORTED_MARKER    = 0x0D;
    const RECORDSET_MARKER      = 0x0E; // reserved, not supported
    const XML_DOCUMENT_MARKER   = 0x0F;
    const TYPE_OBJECT_MARKER    = 0x10;
    const AVMPLUS_OBJECT_MARKER = 0x11;

    var amf0Content = {};

    function readStringAMF0() {
      let str = "";
      let strLen = binIStream.read16();
      if (strLen > 0) {
        str = binIStream.readBytes(strLen);
      }

      return str;
    }

    function readEcmaArrayAMF0() {
      let count = binIStream.read32();
      let ecma = {};
      for (let i=0; i<count; i++) {
        let name = readStringAMF0();
        Object.defineProperty(ecma, name, readValueTypeAMF0());
      }
      let endName = readStringAMF0();
      let objectEndMarker = binIStream.read8();
      if (endName.length != 0 || objectEndMarker != OBJECT_END_MARKER) {
        throw new Error("readEcmaArrayAMF0: End marker is not found");
      }
      return ecma;
    }

    function readObjectTypeAMF0() {
      let retObj = {};
      let endObject = false;
      while (!endObject) {
        let name = readStringAMF0();
        if (name.length == 0) {
          let objectEndMarker = binIStream.read8();
          if (objectEndMarker != OBJECT_END_MARKER) {
            throw new Error("readObjectTypeAMF0: Expected " + OBJECT_END_MARKER + " having " + objectEndMarker);
          }
          endObject = true;
        } else {
          Object.defineProperty(retObj, name, readValueTypeAMF0());
        }
      }
      return retObj;
    }

    function readStrictArrayAMF0() {
      let count = binIStream.read32();
      let array = [];
      for (let i=0; i<count; i++) {
        let val = readValueTypeAMF0();
        if (val.__proto__ != null) {
          // not an undefined value
          array.push(val);
        } else {
          array.push("<undefined>"); // FIXME?
        }
      }
      return array;
    }

    function readDateAMF0() {
      let millisec = binIStream.readDouble();
      let timeZone = binIStream.read16(); // unsuported, always 0x0000
      let date = new Date(millisec);
      return date;
    }

    function readValueTypeAMF0() {
      if (binIStream.available() == 0) {
        throw new Error("Unexpected end of file");
      }
      let marker = binIStream.read8();
      let retVal = null;

      switch (marker) {
      case NUMBER_MARKER:
        retVal = {
          value: binIStream.readDouble(),
          enumerable: true
        };
        break;
      case BOOLEAN_MARKER:
        retVal = {
          value: binIStream.readBoolean(),
          enumerable: true
        };
        break;
      case STRING_MARKER:
        retVal = {
          value: readStringAMF0(),
          enumerable: true
        };
        break;
      case OBJECT_MARKER:
        retVal = {
          value: readObjectTypeAMF0(),
          enumerable: true
        };
        break;
      case MOVIECLIP_MARKER:
        throw new Error("MOVIECLIP_MARKER: Movieclip type is not supported");
      case NULL_MARKER:
        retVal = {
          value: null,
          enumerable: true
        };
      case UNDEFINED_MARKER:
        retVal = {
          __proto__: null,
          enumerable: true
        };
        break;
      case REFERENCE_MARKER:
        retVal = {
          value: "<Reference to object #" + binIStream.read16() + ">",
          enumerable: true
        };
        break;
      case ECMA_ARRAY_MARKER:
        retVal = {
          value: readEcmaArrayAMF0(),
          enumerable: true
        }
        break;
      case OBJECT_END_MARKER:
        throw new Error("OBJECT_END_MARKER: Unexpected marker");
      case STRICT_ARRAY_MARKER:
        retVal = {
          value: readStrictArrayAMF0(),
          enumerable: true
        };
        break;
      case DATE_MARKER:
        retVal = {
          value: readDateAMF0(),
          enumerable: true
        };
        break;
      case LONG_STRING_MARKER:
        throw new Error("LONG_STRING_MARKER: TODO");
      case UNSUPPORTED_MARKER:
        retVal = {
          value: "UNSUPPORTED_MARKER",
          enumerable: true
        };
        break;
      case RECORDSET_MARKER:
        throw new Error("RECORDSET_MARKER: RecordSet type is not supported");
      case XML_DOCUMENT_MARKER:
        throw new Error("XML_DOCUMENT_MARKER: TODO");
      case TYPE_OBJECT_MARKER:
        throw new Error("TYPE_OBJECT_MARKER: TODO");
      case AVMPLUS_OBJECT_MARKER:
        throw new Error("AVMPLUS_OBJECT_MARKER: TODO");
      default:
        throw new Error("Unknown marker type:" + marker);
      }

      return retVal;
    }

    try {
      while (binIStream.available() > 0) {
        let key = readStringAMF0();
        if (key.length > 0) {
          let valObj = readValueTypeAMF0();
          Object.defineProperty(amf0Content, key, valObj);
        } else {
          throw new Error("Key is an empty string");
        }
        let nullChar = binIStream.read8();
        if (nullChar != 0) {
          throw new Error("End of object != 0: " + nullChar);
        }
      }
    } catch (ex) {
      Object.defineProperty(amf0Content, "readError", {value: ex.message, enumerable: true});
      Object.defineProperty(amf0Content, "readErrorStack", {value: ex.stack, enumerable: true});
      log.e(ex);
    }

    return amf0Content;
  },

  readAMF3: function(binIStream) {

    const UNDEFINED_MARKER     = 0x00;
    const NULL_MARKER          = 0x01;
    const FALSE_MARKER         = 0x02;
    const TRUE_MARKER          = 0x03;
    const INTEGER_MARKER       = 0x04;
    const DOUBLE_MARKER        = 0x05;
    const STRING_MARKER        = 0x06;
    const XML_DOC_MARKER       = 0x07;
    const DATE_MARKER          = 0x08;
    const ARRAY_MARKER         = 0x09;
    const OBJECT_MARKER        = 0x0A;
    const XML_MARKER           = 0x0B;
    const BYTE_ARRAY_MARKER    = 0x0C;
    const VECTOR_INT_MARKER    = 0x0D;
    const VECTOR_UINT_MARKER   = 0x0E;
    const VECTOR_DOUBLE_MARKER = 0x0F;
    const VECTOR_OBJECT_MARKER = 0x10;
    const DICTIONARY_MARKER    = 0x11;

    var amf3Content = {};

    var objectsRefTable = {
      STRING:        [],
      XML_DOC:       [],
      DATE:          [],
      ARRAY:         [],
      OBJECT:        [],
      OBJECT_TRAITS: [],
      XML:           [],
      BYTE_ARRAY:    [],
      VECTOR_INT:    [],
      VECTOR_UINT:   [],
      VECTOR_DOUBLE: [],
      VECTOR_OBJECT: [],
      DICTIONARY:    []
    };

    function readIntegerAMF3() {
      // Variable Length Unsigned 29-bit Integer Encoding

      let u8 = binIStream.read8();
      let val = u8 & 0x7F; // remove the encoding bit

      for (let i=0; i<3 && (u8&0x80) == 0x80; i++) {
        val = val<<7;
        u8 = binIStream.read8();
        val += u8 & 0x7F;
      }

      return val;
    }

    function readStringAMF3() {
      let str = null;
      let strType = readIntegerAMF3();

      if (strType & 0x1) {
        // U29s-value
        let strLen = strType >> 1;
        if (strLen == 0) {
          str = ""; // empty string
        } else if (binIStream.available() >= strLen) {
          str = binIStream.readBytes(strLen);
          objectsRefTable.STRING.push(str);
        } else {
          throw new Error("Unexpected end of file while reading a string");
        }
      } else if (strType > 0) {
        let idx = strType >> 1;
        if (idx < objectsRefTable.STRING.length) {
          str = objectsRefTable.STRING[idx];
        } else {
          str = "<Error: String reference index out of bound: " +
                idx +
                " >= " +
                objectsRefTable.STRING.length +
                ">";
        }
      } // else : null string

      return str;
    }

    function readArrayAMF3() {
      // array-type = array-marker (U29O-ref | (U29A-value (UTF-8-empty | *(assoc-value) UTF-8-empty) *(value-type)))
      let arrayType = readIntegerAMF3();
      if ((arrayType & 0x01) == 0x01) {
        // U29A-value
        let denseCount = (arrayType >> 1);

        let arrayObj = [];
        let assocValues = {};
        let haveAssocValues = false;
        let assocValueKey = readStringAMF3();
        while (assocValueKey.length > 0) {
          haveAssocValues = true;
          let assocValue = readValueTypeAMF3();
          Object.defineProperty(assocValues, assocValueKey, assocValue);
          assocValueKey = readStringAMF3();
        }
        if (haveAssocValues) {
          // TODO do the associations
          arrayObj.push(assocValues);
        }

        for (let i=0; i<denseCount; i++) {
          let data = readValueTypeAMF3();
          let value;
          if (typeof(data.value) != "undefined") {
            value = data.value;
          } else {
            value = "<undefined>";
          }
          arrayObj.push(value);
        }

        objectsRefTable.ARRAY.push(arrayObj);

        return arrayObj;
      } else {
        // U29O-ref
        let refId = arrayType >> 1;
        return {array_object_ref: refId};
      }
    }

    /*
    object-marker
      (
      U29O-ref |
      (U29O-traits-ext class-name *(U8)) |
      U29O-traits-ref |
      (U29O-traits class-name *(UTF-8-vr))) *(value-type) *(dynamic-member))
      )
    */
    function readObjectAMF3() {
      let object = {};
      let objType = readIntegerAMF3();

      if ((objType & 0x01) == 0x00) {
        // U29O-ref : xxxxxxx0
        let idx = objType >> 1;
        if (DEBUG) {
          object.type = "U29O-ref";
          object.refIndex = idx;
        }
        if (idx < objectsRefTable.OBJECT.length) {
          object = objectsRefTable.OBJECT[idx];
        } else {
          object.decodeError = "<Error: Object reference index out of bound: " +
                                 idx +
                                 " >= " +
                                 objectsRefTable.OBJECT.length.toString(10) +">";
        }

      } else if ((objType & 0x02) == 0x00) {
        // U29O-traits-ref : xxxxxx01
        let idx = objType >> 2;
        let baseObj = null;
        if (idx >= objectsRefTable.OBJECT_TRAITS.length) {
          baseObj = {};
          baseObj.dynamicMembers = {};
          if (DEBUG) {
            object.decodeError = "<Error: Object traits reference index out of bound: " +
                                 idx +
                                 " >= " +
                                 objectsRefTable.OBJECT_TRAITS.length.toString(10) +">";
          }
        } else {
          baseObj = objectsRefTable.OBJECT_TRAITS[idx];
        }

        if (DEBUG) {
          object.type = "U29O-traits-ref";
          object.traitsRefIndex = idx;
        }

        if (typeof(baseObj.dynamicMembers) != "undefined") {
          object.dynamicMembers = {};
          // Get the specific dynamicMembers for this one
          let key = readKeyValueAMF3(object.dynamicMembers);
          while (key.length > 0) {
            key = readKeyValueAMF3(object.dynamicMembers);
          }
          // Set to null the missing properties
          for (let baseObjKey in baseObj.dynamicMembers) {
            if (baseObj.dynamicMembers.hasOwnProperty(baseObjKey) && !object.dynamicMembers.hasOwnProperty(baseObjKey)) {
              object.dynamicMembers[baseObjKey] = null;
            }
          }
        } else {
          object.TODO = "<An object of type U29O-traits-ref with no dynamic members>";
        }

        if (Object.keys(object).length == 1) {
          // Most of traits objects have only the dynamics members, so remove the unusefull infos.
          let key = Object.keys(object)[0];
          object = object[key];
        }

      } else if ((objType & 0x04) == 0x00) {
        // U29O-traits : xxxxx011
        // Something wrong in the spec, the parentheses are not correctly closed
        // U29O-traits class-name *(UTF-8-vr))) *(value-type)*(dynamic-member)

        let typeDynamic = (objType & 0x08) == 0x08;
        let numSealed   = objType>>4;
        let className   = readStringAMF3();
        let isAnonymous = className == null || className.length == 0;

        if (DEBUG) {
          object.type = "U29O-traits";
          object.traitIndex  = objectsRefTable.OBJECT_TRAITS.length;
        }

        if (!isAnonymous) {
          object.className = className;
        }

        if (numSealed > 0) {
          object.sealedMembers = {};
          let membersName = [];

          // read the sealed traits members names
          for (let i=0; i<numSealed; i++) {
            membersName.push(readStringAMF3());
          }
          // read the sealed traits members values
          for (let i=0; i<numSealed; i++) {
            let sealedValue = readValueTypeAMF3();
            if (typeof(sealedValue.value) != "undefined") {
              object.sealedMembers[membersName[i]] = sealedValue;
            } else {
              object.sealedMembers[membersName[i]] = "<undefined type>";
            }
          }
        }

        if (typeDynamic) {
          object.dynamicMembers = {};
          let key = readKeyValueAMF3(object.dynamicMembers);
          while (key.length > 0) {
            key = readKeyValueAMF3(object.dynamicMembers);
          }
        }

        objectsRefTable.OBJECT_TRAITS.push(object);
        if (Object.keys(object).length == 1) {
          // Most of traits objects have only the dynamics members, so remove the unusefull infos.
          let key = Object.keys(object)[0];
          object = object[key];
        }

      } else {
        // U29O-traits-ext : xxxxx111
        // U29O-traits-ext class-name *(U8)
        // The number of bytes following is unknown...
        // read until the end :(
        object.type = "U29O-traits-ext";
        object.info = "Opaque binary type: Can't decode more";
        object.opaque = [];
        while (binIStream.available() > 0) {
          object.opaque.push(binIStream.read8());
        }
      }

      return object;
    }

    // Read int/uint/double vectors
    function readVectorAMF3(vectorType) {
      let vectorInfos = {};
      switch (vectorType) {
      case VECTOR_INT_MARKER:
        vectorInfos.refTable = objectsRefTable.VECTOR_INT;
        vectorInfos.readFunc = "read32"; // FIXME, read return an uint
        break;
      case VECTOR_UINT_MARKER:
        vectorInfos.refTable = objectsRefTable.VECTOR_UINT;
        vectorInfos.readFunc = "read32";
        break;
      case VECTOR_DOUBLE_MARKER:
        vectorInfos.refTable = objectsRefTable.VECTOR_DOUBLE;
        vectorInfos.readFunc = "readDouble";
        break;
      default:
        throw new Error("Unknown vector type " + vectorType);
      }

      let u29 = readIntegerAMF3();
      if (u29 & 0x01 == 0x00) {
        // U29O-ref
        let idx = u29 >> 1;
        if (idx < vectorInfos.refTable.length) {
          return vectorInfos.refTable[idx];
        }
        throw new Error("Invalid vector ref index " + idx.toString(10));
      } else {
        let count = u29 >> 1;
        let retObj = {};
        retObj.isFixedVector = binIStream.readBoolean();
        retObj.vector = [];

        for (let i=0; i<count; i++) {
          retObj.vector.push(binIStream[vectorInfos.readFunc]());
        }

        vectorInfos.refTable.push(retObj);

        return retObj;
      }
    }

    function readDateAMF3() {
      let u29d = readIntegerAMF3();
      if (u29d & 0x01) {
        let timeSinceEpochMilli = binIStream.readDouble();
        let date = new Date(timeSinceEpochMilli);
        objectsRefTable.DATE.push(date);
        return date;
      } else {
        let dateIdx = u29d >> 1;
        if (objectsRefTable.DATE.length > dateIdx) {
          return objectsRefTable.DATE[dateIdx];
        } else {
          throw new Error("Invalid date index: " + dateIdx.toString(10));
        }
      }
    }

    function readKeyValueAMF3(contentTarget) {
      if (binIStream.available() > 0) {
        let key = readStringAMF3();
        if (key == null || key.length == 0) {
          return key;
        }

        let debugKey = key;
        let tmpKey = key;
        for (let i=1; contentTarget.hasOwnProperty(tmpKey); i++) {
          tmpKey = debugKey + " (" + i.toString(10) + ")";
        }
        if (tmpKey.length != key.length) {
          key = tmpKey;
        }

        let data = readValueTypeAMF3();
        Object.defineProperty(contentTarget, key, data);

        return key;
      }
      return null;
    }

    function readValueTypeAMF3() {
      let retVal = null;
      if (binIStream.available() > 0) {

        let dataType = binIStream.read8();
        switch (dataType) {
        case UNDEFINED_MARKER:
          retVal = {
            __proto__: null,
            enumerable: true
          };
          break;
        case NULL_MARKER:
          retVal = {
            value: null,
            enumerable: true
          };
          break;
        case FALSE_MARKER:
          retVal = {
            value: false,
            enumerable: true
          };
          break;
        case TRUE_MARKER:
          retVal = {
            value: true,
            enumerable: true
          };
          break;
        case INTEGER_MARKER:
          retVal = {
            value: readIntegerAMF3(),
            enumerable: true
          };
          break;
        case DOUBLE_MARKER:
          retVal = {
            value: binIStream.readDouble(),
            enumerable: true
          };
          break;
        case STRING_MARKER:
          retVal = {
            value: readStringAMF3(),
            enumerable: true
          };
          break;
        case DATE_MARKER:
          retVal = {
            value: readDateAMF3(),
            enumerable: true
          };
          break;
        case ARRAY_MARKER:
          retVal = {
            value: readArrayAMF3(),
            enumerable: true
          };
          break;
        case OBJECT_MARKER:
          retVal = {
            value: readObjectAMF3(),
            enumerable: true
          };
          break;
        case XML_MARKER: {
          // FIXME: Never see an xml-marker!
          let u29 = readIntegerAMF3();
          if (u29 & 0x01 == 0x00) {
            // U29O-ref
            let xmlRef = objectsRefTable.XML[u29>>1];
            retVal = {
              value: xmlRef,
              enumerable: true
            };
          } else {
            let xmlLen = u29>>1;
            let xmlStr = binIStream.readBytes(xmlLen);
            retVal = {
              value: xmlStr,
              enumerable: true
            };
          }
          break;
        }
        case BYTE_ARRAY_MARKER: {
          // FIXME: Never seen an byte-array-marker!
          let u29 = readIntegerAMF3();
          if (u29 & 0x01 == 0x00) {
            // U29O-ref
            let byteArrayRef = objectsRefTable.BYTE_ARRAY[u29>>1];
            retVal = {
              value: byteArrayRef,
              enumerable: true
            };
          } else {
            let byteArrayLen = u29 >> 1;
            let array = [];
            if (binIStream.available() < byteArrayLen) {
              throw new Error("binIStream.available() < byteArrayLen");
            }
            for (let i=0; i<byteArrayLen; i++) {
              array.push(binIStream.read8());
            }
            retVal = {
              value: array,
              enumerable: true
            };
            objectsRefTable.BYTE_ARRAY.push(array);
          }
          break;
        }
        case VECTOR_INT_MARKER:
        case VECTOR_UINT_MARKER:
        case VECTOR_DOUBLE_MARKER:
          retVal = {
            value: readVectorAMF3(dataType),
            enumerable: true
          };
          break;
        case VECTOR_OBJECT_MARKER:
          throw new Error("TODO: VECTOR_OBJECT_MARKER");
        case DICTIONARY_MARKER:
          throw new Error("TODO: DICTIONARY_MARKER");
        default:
          throw new Error("Unknown marker: " + dataType.toString(10));
        }
      } else {
        throw new Error("Unexpected EOF");
      }

      return retVal;
    }

    try {
      let key = null;
      let undefinedCount = 0;
      while (binIStream.available() > 0) {
        key = readKeyValueAMF3(amf3Content);

        if (key == null || key.length == 0) {
          throw new Error("Invalid object");
        }

        let endEntry = binIStream.read8();
        if (endEntry != 0) {
          throw new Error("End entry != 0");
        }
      }
    } catch (ex) {
      Object.defineProperty(amf3Content, "readError", {value: ex.message, enumerable: true});
      Object.defineProperty(amf3Content, "readErrorStack", {value: ex.stack, enumerable: true});
      log.e(ex);
    }

    return amf3Content;
  }
};

var gLocalSharedObject = {
  _macromediaDirectory : null, // ex: /home/<user>/.macromedia
  _settingsDirectory   : null, // ex: /home/<user>/.macromedia/Flash_Player/macromedia.com/support/flashplayer/sys/
  _cookiesDirectory    : null, // ex: /home/<user>/.macromedia/Flash_Player/#SharedObjects/2QMDL227/

  init: function() {
    // Register this cookie manager
    if (typeof(cookieManagers) == "object" &&
        cookieManagers.__proto__.constructor.name == "Array" &&
        cookieManagers.indexOf(gLocalSharedObject) == -1) {
      cookieManagers.push(gLocalSharedObject);
    }

    try {

      if (this._macromediaDirectory == null) {
        var OSName = Services.appinfo.OS;
        var macromediaDir = null;

        if (OSName == "WINNT") {
          let dataDir = Services.dirsvc.get("AppData", Ci.nsIFile);

          if (dataDir && dataDir.exists() && dataDir.isDirectory()) {
            let dataDirVista = dataDir.clone();
            // Win Vista & sup
            dataDirVista.append("Roaming");
            dataDirVista.append("Macromedia");
            if (dataDirVista.exists() && dataDirVista.isDirectory()) {
              macromediaDir = dataDirVista;
            } else {
              dataDir.append("Macromedia");
              macromediaDir = dataDir.exists() && dataDir.isDirectory() ? dataDir : null;
            }
          }

        } else if (OSName == "Darwin") {
          // MacOSX
          let usrPrfs = Services.dirsvc.get("UsrPrfs", Ci.nsIFile);
          let uLibDir = Services.dirsvc.get("ULibDir", Ci.nsIFile);

          if (usrPrfs && usrPrfs.exists() && usrPrfs.isDirectory()) {
            usrPrfs.append("Macromedia");
            macromediaDir = usrPrfs.exists() && usrPrfs.isDirectory() ? usrPrfs : null;
          } else if (uLibDir && uLibDir.exists() && uLibDir.isDirectory()) {
            uLibDir.append("Macromedia");
            macromediaDir = uLibDir.exists() && uLibDir.isDirectory() ? uLibDir : null;
          }

        } else { // Linux/Unix
          let homeDir = Services.dirsvc.get("Home", Ci.nsIFile);
          if (homeDir && homeDir.exists() && homeDir.isDirectory()) {
            homeDir.append(".macromedia");
            macromediaDir = homeDir.exists() && homeDir.isDirectory() ? homeDir : null;
          }
        }

        if (macromediaDir != null) {
          this._macromediaDirectory = macromediaDir.clone();
        }
      }

      if (this._macromediaDirectory != null) {
        if (this._cookiesDirectory != null && !this._cookiesDirectory.exists()) {
          // Due to the random part, update the value if it does not exists
          // (if it has been removed since FFx has been launched).
          this._cookiesDirectory = null;
        }

        if (this._settingsDirectory == null) {
          this._settingsDirectory = this._getSettingsDirectory(this._macromediaDirectory);
        }

        if (this._cookiesDirectory == null) {
          this._cookiesDirectory = this._getCookiesDirectory(this._macromediaDirectory);
        }
        return (this._settingsDirectory != null || this._cookiesDirectory != null);
      }
    } catch (ex) {
      this._settingsDirectory = null;
      this._cookiesDirectory  = null;
      log.e(ex.message);
      log.e(ex.stack);
    }

    log.i("Flash Player directory not found")
    return false;
  },

  _getSettingsDirectory: function(macromediaDir) {
    const settingsPaths = ["macromedia.com", "support", "flashplayer", "sys"];
    var settingsDir = null;
    var entries = macromediaDir.directoryEntries;

    while (entries && entries.hasMoreElements()) {
      let entry = entries.getNext();
      entry.QueryInterface(Ci.nsIFile);
      if (entry.isDirectory()) {
        let havePath = true;
        for (let i=0; i<settingsPaths.length; i++) {
          entry.append(settingsPaths[i]);
          if (!(entry.exists() && entry.isDirectory())) {
            havePath = false;
            break;
          }
        }

        if (havePath) {
          settingsDir = entry;
          break;
        }
      }
    }

    log.d("Adobe Flash Settings directory: " + (settingsDir == null ? "not found" : settingsDir.path));
    return settingsDir;
  },

  _getCookiesDirectory: function(macromediaDir) {
    function getPrerandDir(aDir) {
      let entries = aDir.directoryEntries;
      while (entries && entries.hasMoreElements()) {
        let entry = entries.getNext();
        entry.QueryInterface(Ci.nsIFile);
        if (entry.isDirectory()) {
          if (entry.leafName == LSO_PRERAND_DIR) {
            return entry;
          } else {
            let tmp = getPrerandDir(entry);
            if (tmp != null) {
              return tmp;
            }
          }
        }
      }

      return null;
    }

    var cookiesDir = null;
    var prerandDir = getPrerandDir(macromediaDir);
    if (prerandDir != null) {
      let entries = prerandDir.directoryEntries;
      while (entries && entries.hasMoreElements() && cookiesDir == null) {
        cookiesDir = entries.getNext();
        cookiesDir.QueryInterface(Ci.nsIFile);
        if (!cookiesDir.isDirectory()) {
          cookiesDir = null;
        }
      }
    }

    log.d("Adobe Flash Cookies directory: " + (cookiesDir == null ? "not found" : cookiesDir.path));
    return cookiesDir;
  },

  uninit: function() {
    if (typeof(cookieManagers) == "object" && cookieManagers.__proto__.constructor.name == "Array") {
      let index = cookieManagers.indexOf(gLocalSharedObject);
      if (index > -1) {
        cookieManagers.splice(index, 1);
      }
    }

    this._macromediaDirectory = null;
    this._cookiesDirectory    = null;
    this._settingsDirectory   = null;
  },

  makeProtectedObject: function(aHost, aName, aFilePath, _isSettings) {
    function lsoEquals(aItem) {
      return (typeof(aItem) == "object") &&
        COOKIE_TYPE_LSO == aItem.getType() &&
        this.isSettings == aItem.isSettings &&
        this.file       == aItem.file;
    }
    function getRawHost() {
      if (this.host == "localWithNet" || this.host.indexOf(".") == -1) {
        return "local";
      }
      return gCommon.makeStrippedHost(this.host);
    }
    function getHost() {
      return this.host;
    }
    function getName() {
      return this.name;
    }
    function getPath() {
      return this.file;
    }
    function getType() {
      return COOKIE_TYPE_LSO;
    }

    let plso = {
      // Properties
      name       : aName,
      host       : aHost,
      file       : aFilePath,
      isSettings : _isSettings,
      // Helper functions:
      equals     : lsoEquals,
      getRawHost : getRawHost,
      getHost    : getHost,
      getName    : getName,
      getPath    : getPath,
      getType    : getType
    }

    return plso;
  },

  makeLSOObject: function(aSolFile) {
    function cookieEquals(aItem) {
      return (typeof(aItem) == "object") &&
        COOKIE_TYPE_LSO == aItem.getType() &&
        this.isSettings == aItem.isSettings &&
        this.file       == aItem.file;
    }
    function getHost() {
      return this.host;
    }
    function getName() {
      return this.name;
    }
    function getPath() {
      return this.path;
    }
    function getType() {
      return COOKIE_TYPE_LSO;
    }
    function updateFrom(aCookie) {
      if (aCookie.getType() == COOKIE_TYPE_LSO) {
      }
    }

    let path       = aSolFile.parent.path;
    let isSettings = aSolFile.path.indexOf(this._settingsDirectory.path) > -1;
    let isDefault  = false;
    let rawHost    = null;
    let host       = null;
    let file       = null;
    let name       = null;

    if (isSettings) {
      let parentName = aSolFile.parent.leafName;
      if (parentName == "sys") {
        host      = LSO_KEY_MAIN_SETTINGS;
        rawHost   = LSO_KEY_MAIN_SETTINGS;
        isDefault = true;
        path      = "/";
        file      = path + aSolFile.leafName;
        name      = "settings";
      } else {
        path = path.substr(this._settingsDirectory.path.length);
        file = aSolFile.path.substr(this._settingsDirectory.path.length);
      }
    } else {
      path = path.substr(this._cookiesDirectory.path.length);
      file = aSolFile.path.substr(this._cookiesDirectory.path.length);
    }

    file = file.replace(LSO_REGEX_WINSEP, "/");
    path = path.replace(LSO_REGEX_WINSEP, "/");

    if (!isDefault) {
      // Deal with max path len of 32 char, which end with a '#'
      // name example: se.video#/plaza.kit.adplayer.adplayer.sol -> se.videoplaza.kit.adplayer.adplayer
      let iFileTmp = aSolFile.clone();
      let paths    = path.split("/");

      // first the cookie name
      name = iFileTmp.leafName;
      iFileTmp = iFileTmp.parent;
      while (LSO_REGEX_SPLITPATH.test(iFileTmp.leafName)) {
        let prefix = iFileTmp.leafName;
        name = prefix.substring(0, prefix.length-1) + name;
        iFileTmp = iFileTmp.parent;
      }
      if (LSO_REGEX_SOLEXT.test(name)) {
        // Remove the .sol extension
        name = name.substring(0, name.length-4);
      }

      // For the path start at the "root" of the cookies type
      let i = 1;
      host = paths[i];
      while (LSO_REGEX_SPLITPATH.test(host) && i<(paths.length-1)) {
        host = host.substring(0, host.length-1);
        i++;
        host = host + paths[i];
      }
      if (host.charAt(0) == "#") {
        // settings domain cookies start with a '#'
        host = host.substr(1);
      }

      // Local cookies (for some Flash games like botanicula)
      if (host == "localWithNet" || host.indexOf(".") == -1) {
        rawHost = "local";
      } else {
        rawHost = gCommon.makeStrippedHost(host);
      }
    }

    var item = {
      // Properties
      name        : name,
      isDomain    : true,
      host        : host,
      rawHost     : rawHost,
      path        : path,
      isSecure    : false,
      expires     : DATE_MAX,
      creationTime: aSolFile.lastModifiedTime,
      expiry      : DATE_MAX,
      isHttpOnly  : false,
      isSession   : false,
      lastAccessed: 1000,
      isProtected : false,
      isSettings  : isSettings,
      isDefault   : isDefault,
      file        : file,
      fileSize    : aSolFile.fileSize,
      level       : 1,
      parentIndex : -1,
      solFile     : aSolFile,
      _data       : null,
      // Helper functions
      get value() {
        if (this._data == null) {
          this._data = gSolReader.solToJson(this.solFile);
        }
        return this._data;
      },
      set value(val) {
        throw new Error("SOL datas are read only");
      },
      equals      : cookieEquals,
      updateFrom  : updateFrom,
      getHost     : getHost,
      getName     : getName,
      getType     : getType
    };

    item.isProtected = CookieKeeper.isProtectedItem(item);

    return item;
  },

  _fetchSolFiles: function(baseDir, files) {
    let entries = baseDir.directoryEntries;
    while (entries && entries.hasMoreElements()) {
      let entry = entries.getNext();
      entry.QueryInterface(Ci.nsIFile);
      if (entry.isDirectory()) {
        this._fetchSolFiles(entry, files);
      } else if (entry.isFile() && LSO_REGEX_SOLEXT.test(entry.leafName)) {
        files.push(entry);
      }
    }
    return files;
  },

  _fetchItemsSync: function() {
    try {
      if (!this.init()) {
        return [];
      }

      let files    = this._fetchSolFiles(this._macromediaDirectory, []);
      let cookies  = [];
      let nCookies = files.length;
      for (let i=0; i<nCookies; i++) {
        cookies.push(this.makeLSOObject(files[i]));
      }

      return cookies;
    } catch (ex) {
      log.e(ex.message);
      return [];
    }
  },

  /**
   * @brief Fetch the LSO datas (async load)
   * @param aCallbackFunction a function which accept one arg: an array of the loaded items.
   */
  fetchItems: function(aCallbackFunction) {
    setTimeout(function() {
      let cookies = gLocalSharedObject._fetchItemsSync();
      aCallbackFunction(cookies);
    }, 0);
  },

  performDeletion: function (deleteItems) {
    var deletedItems = []; // Put the deleted items here, used by observers
    var error        = null;

    if (deleteItems == null) {
      return false;
    }
    if (deleteItems.length == 0) {
      return true;
    }

    try {
      if (!this.init()) {
        return false;
      }

      let delSettings = Services.prefs.getPrefType(PREF_LSO_KEEP_SETTINGS) == Services.prefs.PREF_BOOL ?
                          !Services.prefs.getBoolPref(PREF_LSO_KEEP_SETTINGS) : false;

      for (let i=0; i<deleteItems.length; i++) {
        let item = deleteItems[i];
        if (item.getType() != COOKIE_TYPE_LSO) {
          continue;
        }

        if (item.isDefault == false || delSettings) {
          let iFile;
          let paths = item.file.split("/");
          let baseDirLen = 0;
          if (item.isSettings) {
            iFile = this._settingsDirectory.clone();
          } else {
            iFile = this._cookiesDirectory.clone();
          }
          baseDirLen = iFile.path.length;
          for (let j=0; j<paths.length; j++) {
            iFile.append(paths[j]);
          }

          if (iFile.exists() && iFile.isFile()) {
            iFile.remove(false);
            deletedItems.push(item);
            // If the item is protected, remove the protection too.
            if (item.isProtected) {
              CookieKeeper.removeProtectedItem(item);
            }
            let parentDir = iFile.parent;
            while (!parentDir.directoryEntries.hasMoreElements() && parentDir.path.length > baseDirLen) {
              // remove the empty directories, up to the LSO dir at max.
              // The length test is just a safety measure,
              // even if should not be a problem without
              parentDir.remove(false);
              parentDir = parentDir.parent;
            }
          }
        } else if (item.isDefault && typeof(gCookieKeeper) != "undefined" && gCookieKeeper._bundle) {
          error = gCookieKeeper._bundle.getString("notification.lso.default_protected");
        }
      }


    } catch (ex) {
      log.e(ex, ex.stack);
      error = ex.message;
    }

    // Item are deleted or an error occured, notify observers
    if (deletedItems.length > 0 || error != null) {
      let subject = {
        error  : error,
        deleted: deletedItems
      }
      Services.obs.notifyObservers({ wrappedJSObject: subject }, CK_OBS_PCOOKIE_TOPIC, CK_OBS_PCOOKIE_LSO_DELETE);
    }

    return error == null;
  },

  /// @brief Remove all unwanted Flash cookies.
  /// @note This function is used at program exit,
  /// or via the main menus (tools menu/toolbar menu)
  /// Must only be called by CookieKeeper.jsm
  clearUnwanted: function() {
    var cookies = this._fetchItemsSync();
    var deleteItems = [];

    for (let i=0; i<cookies.length; i++) {
      let item = cookies[i];
      if (!item.isProtected) {
        deleteItems.push(item);
      }
    }
    this.performDeletion(deleteItems);
  }
};
