/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Observer:
 *  - Cookie add/remove/change.
 *  - Protected cookie add/remove.
 *
 * ************************************************************************
 */

"use strict";

var gCookiesObserver = {
  _blockCookiesEvents     : false,
  _blockProtectionsEvents : false,

  set blockCookiesEvents(value) {
    if (typeof(value) === "boolean") {
      this._blockCookiesEvents = value;
    }
    return value;
  },

  set blockProtectionsEvents(value) {
    if (typeof(value) === "boolean") {
      this._blockProtectionsEvents = value;
    }
    return value;
  },

  init: function() {
    this._blockCookiesEvents     = false;
    this._blockProtectionsEvents = false;
    Services.obs.addObserver(this, "cookie-changed", false);
    // Services.obs.addObserver(this, "dom-storage-changed", false);
    // Services.obs.addObserver(this, "dom-storage2-changed", false);
    Services.obs.addObserver(this, CK_OBS_PCOOKIE_TOPIC, false);
  },

  uninit: function() {
    Services.obs.removeObserver(this, CK_OBS_PCOOKIE_TOPIC);
    // Services.obs.removeObserver(this, "dom-storage2-changed");
    // Services.obs.removeObserver(this, "dom-storage-changed");
    Services.obs.removeObserver(this, "cookie-changed");
  },

  observe: function(subject, topic, action) {
    // if (topic != "cookie-changed") {
    //   log.d("observe: ", topic, action, subject);
    // }

    if (!this._blockProtectionsEvents && topic === CK_OBS_PCOOKIE_TOPIC) {
      // somewhat a strange object: wrappedJSObject ...
      if (action === CK_OBS_PCOOKIE_DEL) {
        this._handleProtItemAddRemoved(subject.wrappedJSObject, false);
      } else if (action === CK_OBS_PCOOKIE_ADD) {
        this._handleProtItemAddRemoved(subject.wrappedJSObject, true);
      } else if (action == CK_OBS_PCOOKIE_DOM_INSERT) {
        this._handleDomStorageInserted(subject.wrappedJSObject);
      } else if (action == CK_OBS_PCOOKIE_DOM_DELETE || action == CK_OBS_PCOOKIE_LSO_DELETE) {
        if (subject.wrappedJSObject.error != null) {
          // Show a warning notification
          let notification    = document.createElementNS(XUL_NS, "notification");
          let notificationBox = document.getElementById("notificationBox");
          notification.setAttribute("id", "notifyDelError");
          notification.setAttribute("type", "warning");
          notification.setAttribute("image", "chrome://cookiekeeper/skin/warning16.png");
          notification.setAttribute("label", subject.wrappedJSObject.error);
          notificationBox.appendChild(notification);
        }
        this._handleItemsDeleted(subject.wrappedJSObject.deleted);
      }
    } else if (!this._blockCookiesEvents && topic === "cookie-changed") {
      if (subject instanceof Ci.nsICookie2) {
        if (action == "changed") {
          this._handleCookieChanged(subject);
        } else if (action === "added") {
          this._handleCookieAdded(subject);
        } else if (action === "deleted") {
          this._handleCookieDeleted(subject);
        }
      } else if (action === "cleared") {
        gCookieKeeper.reload(true);
      } else if (action === "reload") {
        gCookieKeeper.reload(true);
      } else if (action === "batch-deleted") {
        log.e("TODO: batch-deleted " + subject, subject);
        try {
          log.i("length=" + subject.length);
          // log.d(subject.QueryInterface(Ci.nsICookie));
          // for (let i=0; i<subject.length; i++) {
          //   log.i("["+i+"]", subject[i]);
          // }
          // for (let cookie of subject) {
          //   log.i("batch-deleted: ", cookie);
          // }
          for (let i = 0; i < subject.length; i++) {
            log.i("batch-deleted: ", subject.queryElementAt(i, Ci.nsICookie2));
          }
        } catch (ex) {
          log.e(ex);
        }
      } else {
        log.e("Observer: Unknown data", topic, action, subject);
      }
    // } else if (topic == "dom-storage-changed") {
    //   log.e("TODO: dom-storage-changed", action, subject);
    // } else if (topic == "dom-storage2-changed") {
      // log.e("TODO: dom-storage2-changed", action, subject);
    }

    return null;
  },

  _handleCookieChanged: function(changedCookie) {
    var mozCookie = gCookieStorage.makeCookieObject(changedCookie);
    gCookiesTreeView.updateItem(mozCookie);
  },

  _handleCookieAdded: function(addedCookie) {
    var rowCountImpact = 0;
    var addedHost = { value: 0 };
    var cookie = gCookieStorage.makeCookieObject(addedCookie);

    gCookiesTreeView.addItem(cookie);
  },

  _handleCookieDeleted: function(deletedCookie) {
    let aItem = gCookieStorage.makeCookieObject(deletedCookie);
    let rowCount = gCookiesTreeView.rowCount;
    gCookieKeeper.blockSelectEvent();
    gCookiesTreeView.deleteItem(aItem);
    gCookieKeeper.unblockSelectEvent();

    if (rowCount != gCookiesTreeView.rowCount) {
      gCookieKeeper.onItemSelected();
    }
  },

  _handleProtItemAddRemoved: function(aObject, isAdded) {
    if (aObject == null || typeof(isAdded) != "boolean") {
      log.e("ck.observer.js:_handleProtCookieAddRemoved: invalid parameters", aObject, isAdded);
      return;
    }
    var aPItem = null;
    if (aObject.pcookie) {
      aPItem = aObject.pcookie;
    } else if (aObject.pstorage) {
      aPItem = aObject.pstorage;
    } else if (aObject.plso) {
      aPItem = aObject.plso;
    } else {
      log.e("ck.observer.js:_handleProtItemAddRemoved: Unknown protected object type", aObject, isAdded);
      return;
    }

    var strippedHost = aPItem.getRawHost();
    if (gCookiesTreeView._hosts.hasOwnProperty(strippedHost)) {
      let host = gCookiesTreeView._hosts[strippedHost];
      const nItems = host.items.length;
      for (let i=0; i<nItems; i++) {
        let item = host.items[i];
        if (aPItem.equals(item)) {
          item.isProtected = isAdded;

          // Update ui:
          let index = gCookiesTreeView.getIndexOfItem(item);
          if (index > -1) {
            gCookiesTreeView._treeBoxObj.invalidateRow(index);
            let currentSelectionIdx = gCookiesTreeView.selection.currentIndex;
            if (currentSelectionIdx == index || currentSelectionIdx == item.parentIndex) {
              gCookieKeeper.onItemSelected();
            }
          }
          i = nItems;
        }
      }
    } // else: A protected item was added (import?) but the cookie does not exists

  }, // _handleProtCookieAddRemoved

  _handleItemsDeleted: function(aItems) {
    // Item are deleted, update the UI
    if (aItems == null || aItems.length == 0) {
      // nothing to do
    } else if (aItems.length == 1) {
      let rowCount = gCookiesTreeView.rowCount;
      gCookieKeeper.blockSelectEvent();
      gCookiesTreeView.deleteItem(aItems[0]);
      gCookieKeeper.unblockSelectEvent();

      if (rowCount != gCookiesTreeView.rowCount) {
        gCookieKeeper.onItemSelected();
      }
    } else {
      gCookieKeeper.blockSelectEvent();
      gCookiesTreeView.saveState();
      for (let i=0; i<aItems.length; i++) {
        let item = aItems[i];
        if (item.getType() == COOKIE_TYPE_STORAGE || item.getType() == COOKIE_TYPE_LSO) {
          gCookiesTreeView.deleteItemFromCache(item);
        }
      }
      gCookiesTreeView.refreshView(true);
      gCookieKeeper.unblockSelectEvent();
      gCookieKeeper.onItemSelected(null);
    }
  },

  _handleDomStorageInserted: function(aSubject) {
    // If there is an error, no change have been made
    if (aSubject.error == null) {
      // Insert or update items
      let inserted = aSubject.insert;
      for (let i=0; i<inserted.length; i++) {
        let item = inserted[i];
        let strippedHost = item.rawHost;
        if (!(strippedHost in gCookiesTreeView._hosts)) {
          // Create the _hosts "directory" if it do not yet exists
          gCookiesTreeView._hosts[strippedHost] = gCookiesTreeView._makeContainer(strippedHost);
          // Add the new item
          gCookiesTreeView._hosts[strippedHost].items.push(item);
        } else {
          // Search if we have to update it
          let hostItems = gCookiesTreeView._hosts[strippedHost].items;
          let updated   = false;
          for (let j=0; j<hostItems.length && updated==false; j++) {
            let hItem = hostItems[j];
            if (hItem.equals(item)) {
              // It's an update
              hItem.updateFrom(item);
              updated = true;
            }
          }
          if (!updated) {
            // It's a new item
            hostItems.push(item);
          }
        }
      }

      // Update the UI
      gCookieKeeper.blockSelectEvent();
      gCookiesTreeView.saveState();
      gCookiesTreeView.refreshView(true);
      gCookieKeeper.unblockSelectEvent();
      gCookieKeeper.onItemSelected(null);

    } else {
      log.e("_handleDomStorageInserted: do nothing ", aSubject);
    }
  }
};
