/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Permissions glue code.
 * Interract with the browser cookies permissions system.
 *
 * ************************************************************************
 */

"use strict";

const NOTIFICATION_FLUSH_PERMISSIONS = "flush-pending-permissions";
const MOZILLA_PERMISSION_TYPE        = "cookie";

// Manage a permission
function CKPermission(oNativePermission, oCKPermManager) {
  if (typeof(oNativePermission) != "object" || oNativePermission == null) {
    throw "Invalid parameter oNativePermission";
  }
  if (typeof(oCKPermManager) != "object" || oCKPermManager == null) {
    throw "Invalid parameter oCKPermManager";
  }

  var host             = null;
  var origin           = null;
  var havePrincipal    = false;
  var capability       = oNativePermission.capability; // read-only property
  var rawHost          = null;

  this.__defineGetter__("isCKPermission",function() { return true; });
  this.__defineGetter__("host",          function() { return host; });
  this.__defineGetter__("origin",        function() { return origin; });
  this.__defineGetter__("havePrincipal", function() { return havePrincipal; });
  this.__defineGetter__("nativePerm",    function() { return oNativePermission; });
  this.__defineGetter__("type",          function() { return MOZILLA_PERMISSION_TYPE; });
  this.__defineGetter__("capability",    function() { return capability; });
  this.__defineSetter__("capability",    function(val) {
    if (typeof(val) == "number" && (
      val == MOZ_COOKIE_PERM_DEFAULT ||
      val == MOZ_COOKIE_PERM_ALLOW   ||
      val == MOZ_COOKIE_PERM_DENY    ||
      val == MOZ_COOKIE_PERM_FIRST   ||
      val == MOZ_COOKIE_PERM_SESSION)) {

      if (val != capability) {
        capability = val;
        // Update the Mozilla permission:
        Services.perms.addFromPrincipal(oNativePermission.principal, this.type, capability);
      }

    } else {
      throw "Invalid capability";
    }
  });
  this.__defineGetter__("rawHost", function() {
    if (rawHost == null) {
      rawHost = gCommon.makeStrippedHost(host);
    }
    return rawHost;
  });
  this.__defineGetter__("principal", function() {
    if (havePrincipal) {
      return oNativePermission.principal;
    }
    return null;
  });

  if (oNativePermission.principal) {
    host          = oNativePermission.principal.baseDomain;
    origin        = oNativePermission.principal.origin;
    havePrincipal = true;
  } else {
    host = oNativePermission.host;
    origin = "http://" + host;
  }
}

CKPermission.prototype.equals = function(oCKPerm) {
  return this.same(oCKPerm) && this.capability == oCKPerm.capability;
}

CKPermission.prototype.same = function(oCKPerm) {
  if (typeof(oCKPerm) != "object" || oCKPerm == null || !oCKPerm.isCKPermission) {
    return false;
  }

  if (this.havePrincipal && oCKPerm.havePrincipal) {
    return this.principal.equals(oCKPerm.principal);
  } else {
    return this.origin == oCKPerm.origin;
  }
}

CKPermission.prototype.compare = function(oCKPerm) {
  if (typeof(oCKPerm) != "object" || oCKPerm == null || !oCKPerm.isCKPermission) {
    throw new Error("CKPermission.compare: Invalid parameter");
  }

  let cmp = gCommon.cmpHostName(this.host, oCKPerm.host);
  if (cmp == 0) {
    let thisProto = this.origin.substring(0, this.origin.indexOf("://"));
    let oProto    = oCKPerm.origin.substring(0, oCKPerm.origin.indexOf("://"));
    cmp = thisProto > oProto;
  }
  return cmp;
}

// Permissions manager
function CKPermissionsManager() {
  // Constants:
  this.__defineGetter__("PERM_DEFAULT", function() { return Ci.nsICookiePermission.ACCESS_DEFAULT; });
  this.__defineGetter__("PERM_ALLOW",   function() { return Ci.nsICookiePermission.ACCESS_ALLOW; });
  this.__defineGetter__("PERM_DENY",    function() { return Ci.nsICookiePermission.ACCESS_DENY; });
  this.__defineGetter__("PERM_FIRST",   function() { return Ci.nsICookiePermission.ACCESS_ALLOW_FIRST_PARTY_ONLY; });
  this.__defineGetter__("PERM_SESSION", function() { return Ci.nsICookiePermission.ACCESS_SESSION; });
  this.__defineGetter__("PERM_TYPE",    function() { return MOZILLA_PERMISSION_TYPE; });

  // Permissions list
  this.aPermissions = [];
  // Array of change listeners
  this.aListeners = [];

  // I don't know really is this is needed, but mozilla do it before fetching the permissions:
  Services.obs.notifyObservers(null, NOTIFICATION_FLUSH_PERMISSIONS, MOZILLA_PERMISSION_TYPE);
}

CKPermissionsManager.prototype.destructor = function() {
  if (this.aListeners.length > 0) {
    Services.obs.removeObserver(this, "perm-changed");
  }
  delete this.aListeners;
  delete this.aPermissions;
}

/**
 * @brief Fetch current cookies permissions.
 * @return an array of CKPermission.
 */
CKPermissionsManager.prototype.fetchCookiesPermissions = function() {
  let aPerms = [];
  let enumerator = Services.perms.enumerator;

  while (enumerator.hasMoreElements()) {
    let mozPerm = enumerator.getNext().QueryInterface(Ci.nsIPermission);
    if (mozPerm.type == MOZILLA_PERMISSION_TYPE) {
      try {
        let ckPerm = new CKPermission(mozPerm, this);
        aPerms.push(ckPerm);
      } catch (ex) {
        log.e(ex);
      }
    }
  }

  this.aPermissions = aPerms;
  return aPerms;
}

/**
 * @brief Add a permission.
 * @param sHost string hostname ("example.org")
 * @param nPermission permission value, as an integer (PERM_ALLOW, PERM_DENY, PERM_SESSION...)
 * @return An intance of the newly created permission, or null if there was an error.
 */
CKPermissionsManager.prototype.add = function(sHost, nPermission) {

  // Check parameters
  let sHostType = "";
  if (typeof(sHost) == "string" && sHost.length > 0) {
    sHostType = "string";
  } else if (sHost instanceof Ci.nsIURI) {
    sHostType = "uri";
  }

  if (sHostType.length == 0) {
    throw new Error("Invalid sHost param type");
  }
  if (typeof(nPermission) != "number") {
    throw new Error("Invalid nPermission param type");
  }
  if (!(nPermission == MOZ_COOKIE_PERM_DEFAULT ||
        nPermission == MOZ_COOKIE_PERM_ALLOW   ||
        nPermission == MOZ_COOKIE_PERM_DENY    ||
        nPermission == MOZ_COOKIE_PERM_FIRST   ||
        nPermission == MOZ_COOKIE_PERM_SESSION)) {
    throw new Error("Invalid nPermission param value");
  }

  var oCKPerm   = null;
  var fakePerm  = null;
  var principal = null;
  var uri;

  if (sHostType == "uri") {
    uri = sHost;
  } else if (sHost.indexOf("http://") == 0 || sHost.indexOf("https://") == 0) {
    uri = Services.io.newURI(sHost, null, null);
  } else {
    let domain = (sHost.charAt(0) == ".") ? sHost.substring(1, sHost.length) : sHost;
    uri = Services.io.newURI("http://" + domain, null, null);
  }

  try {
    if (typeof(Services.scriptSecurityManager.createCodebasePrincipal) == "function") {
      // Firefox
      principal = Services.scriptSecurityManager.createCodebasePrincipal(uri, {});
    } else {
      // Palemoon
      principal = Services.scriptSecurityManager.getSimpleCodebasePrincipal(uri);
    }
    // If we have ended up with an unknown scheme, the following will throw.
    principal.origin;
    // If we have ended up with an invalid domain, the following will throw.
    principal.baseDomain;
  } catch (ex) {
    log.e("CKPermissionsManager.add(): Invalid principal", ex.toString());
    return null;
  }

  Services.perms.addFromPrincipal(principal, MOZILLA_PERMISSION_TYPE, nPermission);

  if (this instanceof CKPermissionsManager) {
    fakePerm = {
      principal : principal,
      capability: nPermission
    };

    // If there is the perm observer active, this permission may already be added
    oCKPerm = this.findCKPerm(fakePerm, false);
    if (oCKPerm == null) {
      oCKPerm = new CKPermission(fakePerm, this);
      this.aPermissions.push(oCKPerm);
    }

    return oCKPerm;
  }

  return principal;
}

/**
 * @brief Remove a permission.
 * @param oPermission a CKPermission object.
 */
CKPermissionsManager.prototype.remove = function(oPermission) {
  // Check if parameter is an URI:
  if (oPermission instanceof Ci.nsIURI) {
    let principal;
    if (typeof(Services.scriptSecurityManager.createCodebasePrincipal) == "function") {
      // Firefox
      principal = Services.scriptSecurityManager.createCodebasePrincipal(oPermission, {});
    } else {
      // Palemoon
      principal = Services.scriptSecurityManager.getSimpleCodebasePrincipal(oPermission);
    }
    Services.perms.removeFromPrincipal(principal, MOZILLA_PERMISSION_TYPE);
    return true;
  }

  // Else:
  // Check if the permission exists
  if (typeof(oPermission) != "object" || oPermission == null || !oPermission.isCKPermission) {
    throw "Invalid permission object";
  }

  if (this.findCKPerm(oPermission.nativePerm, true)) {
    if (oPermission.havePrincipal) {
      // New permissions system in gecko
      Services.perms.removeFromPrincipal(oPermission.principal, MOZILLA_PERMISSION_TYPE);
    } else try {
      Services.perms.remove(oPermission.host, MOZILLA_PERMISSION_TYPE);
    } catch (ex) {
      let uri = Services.io.newURI("http://" + oPermission.host, null, null);
      Services.perms.remove(uri, MOZILLA_PERMISSION_TYPE);
    }
    return true;
  }

  return false;
}

CKPermissionsManager.prototype.getDefaultPerm = function(aParam) {
  if (typeof(aParam) == "string") {
    let uri = Services.io.newURI("http://" + aParam, null, null);
    return Services.perms.testPermission(uri, this.PERM_TYPE);
  } else if (aParam instanceof Ci.nsIURI) {
    return Services.perms.testPermission(aParam, this.PERM_TYPE);
  }
  return -1;
}

CKPermissionsManager.prototype.getCurrentPerm = function(aParam) {
  if (typeof(aParam) == "string") {
    let uri = Services.io.newURI("http://" + aParam, null, null);
    return Services.perms.testExactPermission(uri, this.PERM_TYPE);
  } else if (aParam instanceof Ci.nsIURI) {
    return Services.perms.testPermission(aParam, this.PERM_TYPE);
  }
  return -1;
}

/**
 * @brief Add a permission change observer.
 * @param aCallback function(CKPermission, changeType)
 * @par
 * changeType string: "added" / "changed" / "deleted".
 */
CKPermissionsManager.prototype.addPermChangeObserver = function(aCallback) {
  if (typeof(aCallback) == "function" && this.aListeners.indexOf(aCallback) == -1) {
    this.aListeners.push(aCallback);
    if (this.aListeners.length == 1) {
      Services.obs.addObserver(this, "perm-changed", false);
    }

    return true;
  }
  return false;
}

CKPermissionsManager.prototype.removePermChangeObserver = function(aCallback) {
  if (typeof(aCallback) == "function") {
    let idx = this.aListeners.indexOf(aCallback);
    if (idx != -1) {
      this.aListeners.splice(idx, 1);

      if (this.aListeners.length == 0) {
        Services.obs.removeObserver(this, "perm-changed");
      }

      return true;
    }
  }
  return false;
}

CKPermissionsManager.prototype.findCKPerm = function(oNativePermission, bRemove) {
  let oPerm  = new CKPermission(oNativePermission, this);
  let ckPerm = null;

  for (let i=0; i<this.aPermissions.length && ckPerm == null; i++) {
    ckPerm = this.aPermissions[i];
    if (!ckPerm.same(oPerm)) {
      ckPerm = null;
    } else if (bRemove) {
      this.aPermissions.splice(i, 1);
    }
  }

  return ckPerm;
}

CKPermissionsManager.prototype.observe = function(aSubject, aTopic, aData) {
  if (this.aListeners.length == 0) {
    // We have a bug here, should not happen.
    log.e("CKPermissionsManager.observe: no listener registered");
    return;
  }

  if (aTopic == "perm-changed") {
    let ckPerm = null;
    let permission = aSubject.QueryInterface(Ci.nsIPermission);

    if (aData == "added") {
      ckPerm = this.findCKPerm(permission, false);
      if (ckPerm == null) {
        ckPerm = new CKPermission(permission, this);
        this.aPermissions.push(ckPerm);
      }

    } else if (aData == "changed") {
      ckPerm = this.findCKPerm(permission, false);
      if (ckPerm == null) {
        // No perm fetch, create a new CKPermission
        ckPerm = new CKPermission(permission, this);
        this.aPermissions.push(ckPerm);
      } else {
        // Update our CKPermission
        ckPerm.capability = permission.capability;
      }

    } else if (aData == "deleted") {
      ckPerm = this.findCKPerm(permission, true);
      if (ckPerm == null) {
        ckPerm = new CKPermission(permission, this);
      }

    } else {
      log.d("CKPermissionsManager.observe: TODO: " + aData);
    }

    for (let i=0; i<this.aListeners.length && ckPerm != null; i++) {
      this.aListeners[i](ckPerm, aData);
    }
  }
}
