/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Public API: Private part, used by the bootstrap code too.
 *
 * ************************************************************************
 */

"use strict";

// Protected list format version
const PROTECTED_LIST_VERSION = 1;

var CookieKeeperPrivate = {
  _pListLastSave : 0,
  _protectedList : null,

  init: function() {
    this._protectedList = {
      lastUpdate : 0,
      version    : PROTECTED_LIST_VERSION,
      cookies    : [],
      storage    : [],
      lso        : []
    };
    this._pListLastSave = 0;
  },

  uninit: function() {
    this._pListLastSave = 0;
    this._protectedList = null;
  },

  getCookieWindows: function() {
    var e    = Services.wm.getEnumerator(null);
    var wins = [];

    while (e.hasMoreElements()) {
      let win = e.getNext().QueryInterface(Ci.nsIDOMChromeWindow);
      let id  = win.document.documentElement.getAttribute("id");

      for (let i=0; i<CK_WINDOWS_ID.length; i++) {
        if (CK_WINDOWS_ID[i] == id) {
          wins.push({id: id, window:win});
        }
      }
    }

    return wins;
  },

  /// Load the protection list, from the file: cookiekeeper.json (UTF-8 / JSON)
  /// In the user profile directory.
  /// @param doSync if true, dot not used async load.
  loadProtectedList: function(doSync) {

    function jsonFileLoaded(jsonString) {
      var pListsVersion = 0;
      if (typeof(jsonString) != "string") {
        log.e("loadProtectedList.jsonFileLoaded: parameter is not a string", jsonString);
        return;
      }
      if (jsonString.length > 0) {
        let pLists = JSON.parse(jsonString);

        // Add the helper functions:

        // Standard cookies
        if (typeof(pLists.cookies) != "undefined") {
          let sList = pLists.cookies; // source
          let dList = CookieKeeperPrivate._protectedList.cookies; //destination
          for (let i=0; i<sList.length; i++) {
            let item = sList[i];
            if (typeof(item.host) == "string" && typeof(item.name) == "string" && typeof(item.path) == "string") {
              dList.push(gCookieStorage.makeProtectedObject(item.host, item.name, item.path));
            } else {
              log.e("Found an invalid protected cookie at pos:" + i, item);
            }
          }
        }

        // HTML5 Dom Storage
        if (typeof(pLists.storage) != "undefined") {
          let sList = pLists.storage;
          let dList = CookieKeeperPrivate._protectedList.storage;
          for (let i=0; i<sList.length; i++) {
            let item = sList[i];
            if (typeof(item.scope) == "string" && typeof(item.key) == "string") {
              dList.push(gDomStorage.makeProtectedObject(item.scope, item.key));
            } else {
              log.e("Found an invalid protected dom storage at pos:" + i, item);
            }
          }
        }

        // Local shared object (LSOs) / Flash cookies.
        if (typeof(pLists.lso) != "undefined") {
          let sList = pLists.lso;
          let dList = CookieKeeperPrivate._protectedList.lso;
          for (let i=0; i<sList.length; i++) {
            let item = sList[i];
            if (typeof(item.host) == "string" && typeof(item.name) == "string" &&
                typeof(item.file) == "string" && typeof(item.isSettings) == "boolean") {
              dList.push(gLocalSharedObject.makeProtectedObject(item.host, item.name, item.file, item.isSettings));
            } else {
              log.e("Found an invalid protected lso at pos:" + i, item);
            }
          }
        }

        if (typeof(pLists.version) == "number") {
          pListsVersion = pLists.version;
        }

        CookieKeeperPrivate._protectedList.lastUpdate = pLists.lastUpdate;

      }

      // Check version format:
      if (pListsVersion != PROTECTED_LIST_VERSION) {
        // Tell to save to the new format version
        CookieKeeperPrivate._pListLastSave = CookieKeeperPrivate._protectedList.lastUpdate;
        CookieKeeperPrivate._protectedList.lastUpdate = Date.now();
      } else {
        CookieKeeperPrivate._pListLastSave = CookieKeeperPrivate._protectedList.lastUpdate;
      }
    }

    function asyncLoaded(istream, status) {
      try {
        if (Components.isSuccessCode(status)) {
          let jsonString = NetUtil.readInputStreamToString(istream, istream.available(), {charset: "UTF-8"});
          jsonFileLoaded(jsonString);
        } else {
          log.e("error on file read isSuccessCode = " + status);
        }
      } catch (ex) {
        log.e("loadProtectedList.asyncLoaded: " + ex, ex);
      }
    }

    function syncLoad(jsonFile) {
      try {
        let istream = Cc["@mozilla.org/network/file-input-stream;1"].createInstance(Ci.nsIFileInputStream);
        istream.init(jsonFile, -1, -1, false);
        let jsonString = NetUtil.readInputStreamToString(istream, istream.available(), {charset: "UTF-8"});
        istream.close();

        jsonFileLoaded(jsonString);
      } catch (ex) {
        log.e("loadProtectedList.syncLoad: " + ex, ex);
      }
    }

    if (this._protectedList.lastUpdate != 0) {
      // Already loaded
      return;
    }

    let jsonFile = FileUtils.getFile("ProfD", [PROT_JSON_FILE], false);
    if (!jsonFile.exists()) {
      // Nothing to load
      this._protectedList.lastUpdate = Date.now();
      return;
    }

    if (doSync) {
      syncLoad(jsonFile);
    } else {
      try {
        NetUtil.asyncFetch(jsonFile, asyncLoaded);
      } catch (ex) {
        log.e(ex);
        syncLoad(jsonFile);
      }
    }
  },

  /// Save the protected cookies/storage to the file: cookiekeeper.json (UTF-8 / JSON)
  /// In the user profile directory
  /// @param doSync [boolean] if true, save in sync mode, if false, async.
  /// @note The sync mode should only be used during add-on shutdown
  saveProtectedList: function(doSync) {
    function writeDoneCallback(status) {
      if (!Components.isSuccessCode(status)) {
        // Log the error:
        log.e("cookiekeeper.json: error on write file (" + status + ")");
      } else {
        // Data has been written to the file.
        log.i("saveProtectedList: cookiekeeper.json file saved");
        CookieKeeperPrivate._pListLastSave = Date.now();
      }
    }

    if (this._pListLastSave < this._protectedList.lastUpdate) {
      let openFlags  = FileUtils.MODE_WRONLY | FileUtils.MODE_TRUNCATE;
      let jsonFile   = FileUtils.getFile("ProfD", [PROT_JSON_FILE], false);
      let jsonString = JSON.stringify(this._protectedList);

      if (!jsonFile.exists()) {
        // To keep some privacy... not making the file readable by everybody
        jsonFile.create(Ci.nsIFile.NORMAL_FILE_TYPE, parseInt("600", 8)); // Unix file perm: -rw-------
      }
      let ostream = FileUtils.openFileOutputStream(jsonFile, openFlags);

      if (doSync) {
        let converter = Cc["@mozilla.org/intl/converter-output-stream;1"].createInstance(Ci.nsIConverterOutputStream);
        converter.init(ostream, "UTF-8", 0, 0);
        converter.writeString(jsonString);
        converter.close();
        this._pListLastSave = Date.now();
        log.i("saveProtectedList: cookiekeeper.json file saved");
      } else {
        let converter = Cc['@mozilla.org/intl/scriptableunicodeconverter'].createInstance(Ci.nsIScriptableUnicodeConverter);
        converter.charset = 'UTF-8';
        let istream = converter.convertToInputStream(jsonString);
        NetUtil.asyncCopy(istream, ostream, writeDoneCallback);
      }
    }
  }

}; // CookieKeeperPrivate
