/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Cookies tree view.
 *
 * ************************************************************************
 */

"use strict";

// View mode
const VIEW_MODE = {
  TREE : 0, /// Tree view -> sort by host
  FLAT : 1, /// Flat view -> sort by another property or filtered by "textBoxSearch"
};

// Protection filter values (show only all/protected/unprotected cookies):
const FILTER = {
  NONE        : 0,
  PROTECTED   : 1,
  UNPROTECTED : 2,
  DOM_HTML5   : 3,
  LAST        : 3  // Update this value to the last (previous) filter value. Internal use only.
};

const SORT_INFOS = {
  HOST          : {property: "host",         column: "domainCol"},
  NAME          : {property: "name",         column: "nameCol"},
  PATH          : {property: "path",         column: "pathCol"},
  VALUE         : {property: "value",        column: "valueCol"},
  PROTECTED     : {property: "isProtected",  column: "protectedCol"},
  HTTP          : {property: "isHttpOnly",   column: "httpCol"},
  SECURE        : {property: "isSecure",     column: "secureCol"},
  SESSION       : {property: "isSession",    column: "sessionCol"},
  CREATION      : {property: "creationTime", column: "creationCol"},
  LAST_ACCESSED : {property: "lastAccessed", column: "lastAccessedCol"},
  EXPIRES       : {property: "expires",      column: "expiresCol"},
};
// -- Default sort property:
const SORT_INFOS_DEFAULT = SORT_INFOS.HOST;

var gCookiesTreeView = {
  _hosts        : {},             // cookies list, by hosts
  _dHosts       : {},             // Displayed host when display a tree view
  _dHostsOrder  : [],             // Displayed hosts list, to be able to sort the tree view
  _flatSet      : [],             // Items when display a flat view
  _viewMode     : VIEW_MODE.TREE, // The display mode: tree or flat ()
  _searchValue  : "",
  _gFilter      : FILTER.NONE,    // global filter (show all/protected/unprotected)
  _rowCount     : 0,
  _cacheValid   : 0,
  _cacheItems   : [],
  _selection    : null,            // setup automatically when this view model is attach to the tree view
  _treeBoxObj   : null,
  _sortInfos    : SORT_INFOS_DEFAULT,
  _sortAscending: true,
  _selectedItems: [],
  _selectedIndex: -1,
  _openHosts    : [],
  _strDateMax   : null,
  _strDateUnknown: null,
  _openState    : false,

  // -------------------------------------------------
  // Customs functions
  // -------------------------------------------------

  get rowCount() {
    return this._rowCount;
  },

  set rowCount(val) {
    var oldCount = this._rowCount;
    if (typeof(val) == "number" && val >= 0) {
      if (val != oldCount) {
        this._rowCount = val;
        if (oldCount > 0) {
          this._treeBoxObj.rowCountChanged(0, -oldCount);
        }
        if (val > 0) {
          this._treeBoxObj.rowCountChanged(0, val);
        }
      }
    } else {
      log.e("gCookiesTreeView.rowCount: invalid value ", val);
    }

    return val;
  },

  get viewMode() {
    return this._viewMode;
  },

  set viewMode(val) {
    let ok = false;
    if (typeof(val) == "number") {
      switch (val) {
      case VIEW_MODE.TREE:
        ok = true;
        document.getElementById("cookiesTree").setAttribute("seltype", "single");
        break;
      case VIEW_MODE.FLAT:
        ok = true;
        document.getElementById("cookiesTree").setAttribute("seltype", "multiple");
        break;
      }
    }

    if (ok) {
      this._viewMode = val;
    } else {
      log.e("gCookiesTreeView.viewMode: invalid value ", val);
    }

    return val;
  },

  get globalFilter() {
    return this._gFilter;
  },

  set globalFilter(val) {
    if (typeof(val) == "number" && val >= 0 && val <= FILTER.LAST) {
      this._gFilter = val;
      this.refreshView(true);
    } else {
      log.e("gCookiesTreeView.globalFilter: invalid value ", val);
    }
    return val;
  },

  get searchValue() {
    return this._searchValue;
  },

  set searchValue(val) {
    if (typeof(val) == "string" && this._searchValue != val) {
      this._searchValue = val;
      if (val.length == 0 && this._sortInfos === SORT_INFOS.HOST) {
        this.viewMode = VIEW_MODE.TREE;
      } else {
        this.viewMode = VIEW_MODE.FLAT;
      }
      this.refreshView(true);
    } else {
      log.e("gCookiesTreeView.searchValue: invalid value ", val);
    }
    return val;
  },

  get sortInfos() {
    return this._sortInfos;
  },
  set sortInfos(val) {
    this._sortInfos = val;
    return this._sortInfos;
  },
  get sortAscending() {
    return this._sortAscending;
  },
  set sortAscending(val) {
    if (typeof(val) == "boolean") {
      this._sortAscending = val;
    }
    return this._sortAscending;
  },

  init: function(bundle) {
    this._strDateMax     = bundle.getString("date.max");
    this._strDateUnknown = bundle.getString("date.unknown");
  },

  reset: function() {
    let oldCount = this._rowCount;

    this._hosts         = {};
    this._dHosts        = {}
    this._dHostsOrder   = [];
    this._flatSet       = [];
    this._cacheItems    = [];
    this._cacheValid    = 0;
    this._searchValue   = "";
    this._viewMode      = VIEW_MODE.TREE;
    this._gFilter       = FILTER.NONE;
    this._rowCount      = 0;
    this._sortInfos     = SORT_INFOS_DEFAULT;
    this._sortAscending = true;

    this._treeBoxObj.rowCountChanged(0, -oldCount);
    this._selection.clearSelection();
  },

  _makeContainer: function(strippedHost) {
    function containerEquals(aItem) {
      return (typeof(aItem) == "object") &&
        aItem.getType() == COOKIE_TYPE_CONTAINER  &&
        aItem.rawHost   == this.rawHost;
    }
    function getHost() {
      return this.rawHost;
    }
    function getType() {
      return COOKIE_TYPE_CONTAINER;
    }

    let host = {
      items   : [],
      rawHost : strippedHost,
      level   : 0,
      open    : false,
      type    : COOKIE_TYPE_CONTAINER,
      equals  : containerEquals,
      getHost : getHost,
      getType : getType
    };

    return host;
  },

  updateItem: function(aItem) {
    if (typeof(aItem) != "object") {
      return;
    }
    let strippedHost = aItem.rawHost;
    if (typeof(strippedHost) != "string") {
      return;
    }

    if (strippedHost in this._hosts) {
      var rowIndex = this.getIndexOfItem(aItem);
      if (rowIndex > -1) {
        var oItem = this.getItemAtIndex(rowIndex);
        oItem.updateFrom(aItem);
        this._treeBoxObj.invalidateRow(rowIndex);
        if (this.selection.currentIndex == rowIndex) {
          gCookieKeeper.onItemSelected();
        }
      } else {
        // Item not display
        let items = this._hosts[strippedHost].items;
        for (let i=0; i<items.length; i++) {
          if (items[i].equals(aItem)) {
            items[i].updateFrom(aItem);
            i = items.length;
          }
        }
      }
    }
  },

  /**
   * @brief Add a single item (cookie/dom storage)
   * @param aItem The item to add
   * @note This function will update the UI.
   */
  addItem: function(aItem) {
    function cmpItemName(a, b) {
      let aName = a.getName();
      let bName = b.getName();
      let c = aName.toLowerCase().localeCompare(bName.toLowerCase());
      if (c == 0) {
        c = a.getHost().toLowerCase().localeCompare(b.getHost().toLowerCase());
      }
      if (gCookiesTreeView._sortAscending) { return c; }
      else { return -c; }
    }

    if (typeof(aItem) != "object") {
      return;
    }

    let strippedHost = aItem.rawHost;
    if (typeof(strippedHost) != "string") {
      return;
    }
    let host = null;
    let isNewHost = false;
    if (!(strippedHost in this._hosts)) {
      host = this._makeContainer(strippedHost);
      this._hosts[strippedHost] = host;
      isNewHost = true;
    } else {
      host = this._hosts[strippedHost];
    }

    // Safety check if the item already exists
    // Can happend with appId != 0 or observer "batch-deleted"
    let il = host.items.length;
    for (let i=0; i<il; i++) {
      let item = host.items[i];
      if (item.equals(aItem)) {
        // This cookie has been refreshed
        log.w("gCookiesTreeView: addItem -> updateItem", aItem);
        this.updateItem(aItem);
        return;
      }
    }

    let currentSelIdx  = this.selection.currentIndex;
    let currentSelItem = currentSelIdx == -1 ? null : this.getItemAtIndex(currentSelIdx);

    host.items.push(aItem);
    host.items.sort(cmpItemName);

    let tbo = this._treeBoxObj;
    tbo.beginUpdateBatch();

    // Add it to the display if it match the filter
    if (this.matchFilter(aItem)) {
      if (this.viewMode == VIEW_MODE.TREE) {
        let dHost = this._dHosts[strippedHost];
        if (isNewHost) {
          dHost = this._makeContainer(strippedHost);
          dHost.items.push(aItem);
          this._dHosts[strippedHost] = dHost;
          this._dHostsOrder.push(strippedHost);
          this._dHostsOrder.sort();
          let idx = this._dHostsOrder.indexOf(strippedHost);
          this._invalidateCache(idx);
          this.rowCount += 1;
          tbo.invalidateRow(idx);
        } else if (dHost.open) {
          dHost.items.push(aItem);
          dHost.items.sort(cmpItemName);
          let idx = this._dHostsOrder.indexOf(strippedHost);
          this._invalidateCache(idx);
          this.rowCount += 1;
          tbo.invalidateRow(idx);
        } else {
          dHost.items.push(aItem);
        }

      } else {
        // this.viewMode == VIEW_MODE.FLAT
        this._flatSet.push(aItem);
        this.doSort();
        let idx = this._flatSet.indexOf(aItem);
        this.rowCount += 1;
        tbo.invalidateRow(idx);
      }
    }

    if (currentSelItem != null) {
      currentSelIdx = this.getIndexOfItem(currentSelItem);
      if (currentSelIdx != -1) {
        this.selection.select(currentSelIdx);
      }
    }
    tbo.endUpdateBatch();
  },

  /**
   * @brief Add items (cookies/dom storage) to the list.
   * @par itemsArray An array with the items
   * @note This function is for the initial load only
   */
  addItems: function(itemsArray) {
    const len = itemsArray.length;
    for (let i=0; i<len; i++) {
      let item = itemsArray[i];
      let strippedHost = item.rawHost;

      if (!(strippedHost in this._hosts)) {
        // Create the _hosts "directory" if it do not yet exists
        this._hosts[strippedHost] = this._makeContainer(strippedHost);
      }

      this._hosts[strippedHost].items.push(item);
    }
  },

  /**
   * Delte an item from the cache, without updating the UI.
   * @return true if the item was deleted, false otherwise.
   */
  deleteItemFromCache: function(aItem) {
    let strippedHost = aItem.rawHost;
    if (strippedHost in this._hosts) {
      let host = this._hosts[strippedHost];
      let il = host.items.length;
      for (let i=0; i<il; i++) {
        if (host.items[i].equals(aItem)) {
          // log.d("_deleteFromHost: found host!");
          host.items.splice(i, 1);
          i = il;
        }
      }
      if (host.items.length == 0) {
        delete this._hosts[strippedHost];
      }
      return true;
    }
    return false;
  },

  /**
   * @brief Delete a single item (cookie/dom storage)
   * @param aItem The item deleted, to be remove from our view
   * @note This function will update the UI.
   * Maintain a good selection is complicated if the view mode is a tree...
   */
  deleteItem: function(aItem) {
    let currentSelIdx  = this.selection.currentIndex;

    if (this.deleteItemFromCache(aItem)) {
      let strippedHost = aItem.rawHost;
      let invalidIdx   = 0;
      let itemIdx      = 0;
      let rcc          = 0; // row count change
      if (this.viewMode == VIEW_MODE.TREE) {
        if (this.matchFilter(aItem)) {
          let currentSelItem = currentSelIdx == -1 ? null : this.getItemAtIndex(currentSelIdx);
          let dHost  = this._dHosts[strippedHost];
          let selRem = false; // is the current selection removed ?
          invalidIdx = this.getIndexOfItem(dHost);

          if (currentSelItem != null && (currentSelItem.equals(dHost) || currentSelItem.equals(aItem))) {
            selRem = true;
          }

          let il = dHost.items.length;
          for (let i=0; i<il; i++) {
            if (dHost.items[i].equals(aItem)) {
              dHost.items.splice(i, 1);
              itemIdx = invalidIdx + i + 1;
              i = il;
            }
          }

          if (dHost.open) {
            rcc--;
            if (itemIdx == currentSelIdx) {
              // The selection must stay with the current host
              itemIdx = (itemIdx-invalidIdx == il) ? (itemIdx - 1) : itemIdx;
              currentSelIdx = itemIdx;
            }
          }

          if (dHost.items.length == 0) {
            itemIdx = invalidIdx == currentSelIdx ? invalidIdx : itemIdx;
            rcc--;
            let ho = this._dHostsOrder;
            ho.splice(ho.indexOf(strippedHost), 1);
            delete this._dHosts[strippedHost];
          }

          if (!selRem && currentSelItem != null) {
            currentSelIdx = this.getIndexOfItem(currentSelItem);
          }
        } // else not display

      } else if (this.matchFilter(aItem)) {
        let fl = this._flatSet.length;
        for (let i=0; i<fl; i++) {
          if (this._flatSet[i].equals(aItem)) {
            this._flatSet.splice(i, 1);
            rcc--;
            invalidIdx = itemIdx = i;
            i = fl;
          }
        }
        if (rcc != -1) {
          log.e("item not found in flatSet, but should be");
        }
      }

      if (rcc < 0) {
        let tbo = this._treeBoxObj;
        tbo.beginUpdateBatch();
        this._invalidateCache(invalidIdx);
        this.rowCount += rcc;
        let rc = this.rowCount;

        if (invalidIdx < rc) {
          tbo.invalidateRange(invalidIdx, rc);
        }

        // Handle selection
        if (currentSelIdx < rc) {
          if (itemIdx == currentSelIdx) {
            this.selection.select(itemIdx);
          } else {
            this.selection.select(currentSelIdx);
          }
        } else if (rc > 0) {
          this.selection.select(rc - 1);
        } else {
          // log.d("no selection");
        }

        tbo.endUpdateBatch();
      }

    } else {
      log.i("deleteItem: host not found");
    }
  },

  matchFilter: function(aItem) {
    var ok = false;

    if (aItem != null && aItem.getType() != COOKIE_TYPE_CONTAINER) {
      switch (this._gFilter) {
      case FILTER.NONE:
        ok = true;
        break;
      case FILTER.PROTECTED:
        ok = aItem.isProtected;
        break;
      case FILTER.UNPROTECTED:
        ok = !aItem.isProtected;
        break;
      case FILTER.DOM_HTML5:
        ok = aItem.getType() == COOKIE_TYPE_STORAGE;
        break;
      }

      if (ok && this.searchValue.length > 0) {
        let name = aItem.getName();
        let host = aItem.getHost();
        ok = host.indexOf(this.searchValue)        != -1 ||
             name.indexOf(this.searchValue)        != -1 ||
             aItem.value.indexOf(this.searchValue) != -1;
      }
    }

    return ok;
  },

  refreshView: function(doRestoreState) {
    this._dHosts      = {};
    this._dHostsOrder = [];
    this._flatSet     = [];
    this.rowCount     = 0;

    this._invalidateCache(0);

    let newCount = 0;

    if (this.viewMode == VIEW_MODE.TREE) {
      let dHosts = {};
      let hOrder = [];
      for (let key in this._hosts) {
        let host = this._hosts[key];
        let il = host.items.length;
        for (let i=0; i<il; i++) {
          let item = host.items[i];
          if (this.matchFilter(item)) {
            if (!(host.rawHost in dHosts)) {
              dHosts[host.rawHost] = this._makeContainer(host.rawHost);
              hOrder.push(host.rawHost);
            }
            dHosts[host.rawHost].items.push(item);
          }
        }
      }

      newCount          = hOrder.length;
      this._dHosts      = dHosts;
      this._dHostsOrder = hOrder;
    } else {
      // Flat view
      let flatSet = [];
      for (let key in this._hosts) {
        let host = this._hosts[key];
        let il = host.items.length;
        for (let i=0; i<il; i++) {
          let item = host.items[i];
          if (this.matchFilter(item)) {
            flatSet.push(item);
          }
        }
      }

      newCount = flatSet.length;
      this._flatSet = flatSet;
    }

    this._treeBoxObj.beginUpdateBatch();
    this.selection.clearSelection();
    this.doSort(this._sortInfos, this._sortAscending);
    this.rowCount = newCount;
    if (doRestoreState) {
      this.restoreState();
    }
    this._treeBoxObj.invalidate();
    this._treeBoxObj.endUpdateBatch();
  },

  /**
   * @brief Perform the sort of the cookies
   * @param aSortInfo The property to use for comparison (SORT_INFOS)
   * @param ascending [boolean] true if for in ascending order
   */
  doSort: function(aSortInfo, ascending) {
    if (aSortInfo == null) {
      aSortInfo = this._sortInfos;
    } else {
      this._sortInfos = aSortInfo;
    }
    if (ascending == null || typeof(ascending) != "boolean") {
      ascending = this._sortAscending;
    } else {
      this._sortAscending = ascending;
    }

    function cmpItemHost(a, b) {
      let c = gCommon.cmpHostName(a.getHost(), b.getHost());
      if (c == 0) {
        let c = a.getName().toLowerCase().localeCompare(b.getName().toLowerCase());
      }
      if (ascending) { return c; }
      else { return -c; }
    }

    function cmpItemName(a, b) {
      let c = a.getName().toLowerCase().localeCompare(b.getName().toLowerCase());
      if (c == 0) { return cmpItemHost(a, b); }
      if (ascending) { return c; }
      else { return -c; }
    }

    function cmpItemString(a, b) {
      let c = a[aSortInfo.property].toLowerCase().localeCompare(b[aSortInfo.property].toLowerCase());
      if (c == 0) { return cmpItemHost(a, b); }
      else if (ascending) { return c; }
      else { return -c; }
    }

    function cmpItemBool(a, b) {
      let c = a[aSortInfo.property] - b[aSortInfo.property];
      if (c == 0) { return cmpItemHost(a, b); }
      else if (ascending) { return c; }
      else { return -c; }
    }

    function cmpItemDate(a, b) {
      let c = a[aSortInfo.property] - b[aSortInfo.property];
      if (c == 0) { return cmpItemHost(a, b); }
      else if (ascending) { return c; }
      else { return -c; }
    }

    function cmpItemExpireDate(a, b) {
      let da = a.isSession ? a.expiry : a.expires;
      let db = b.isSession ? b.expiry : b.expires;
      let c = da - db;
      if (c == 0) { return cmpItemHost(a, b); }
      else if (ascending) { return c; }
      else { return -c; }
    }

    if (this.viewMode == VIEW_MODE.TREE) {
      this._dHostsOrder.sort();
      if (!ascending) {
        this._dHostsOrder.reverse()
      }
      for (let key in this._dHosts) {
        let host = this._dHosts[key];
        host.items.sort(cmpItemHost);
      }
      this._invalidateCache(0);
    } else {
      // VIEW_MODE.FLAT
      let cmpMethod = cmpItemBool;
      if (aSortInfo === SORT_INFOS.HOST) {
        cmpMethod = cmpItemHost;
      } else if (aSortInfo === SORT_INFOS.NAME) {
        cmpMethod = cmpItemName;
      } else if (aSortInfo === SORT_INFOS.PATH || aSortInfo === SORT_INFOS.VALUE) {
        cmpMethod = cmpItemString;
      } else if (aSortInfo === SORT_INFOS.CREATION_TIME || aSortInfo === SORT_INFOS.LAST_ACCESSED) {
        cmpMethod = cmpItemDate;
      } else if (aSortInfo === SORT_INFOS.EXPIRES) {
        cmpMethod = cmpItemExpireDate;
      }

      this._flatSet.sort(cmpMethod);
    }
  },

  /**
   * Save the open hosts and selections
   */
  saveState: function () {
    // Save selection
    var seln            = this.selection;
    this._selectedItems = [];

    if (this.viewMode == VIEW_MODE.FLAT) {
      var rangeCount = seln.getRangeCount();
      for (let i=0; i<rangeCount; i++) {
        let min = {}; let max = {};
        seln.getRangeAt(i, min, max);

        for (let j = min.value; j <= max.value; j++) {
          let item = this.getItemAtIndex(j);
          if (item == null) {
            // log.d("No item at index " + j);
          } else {
            // log.d("saveState: selected ", item);
            this._selectedItems.push(item);
          }
        }
      }
    } else {
      let idx  = seln.currentIndex;
      let item = idx == -1 ? null : this.getItemAtIndex(idx);
      if (item == null) {
        // log.d("saveState: nothing selected: " + idx);
      } else {
        this._selectedItems.push(item);
        // log.d("saveState: selected " + idx, item);
      }
    }
    this._selectedIndex = seln.currentIndex;

    // Save open states
    if (this.viewMode == VIEW_MODE.TREE) {
      this._openHosts = [];
      for (let key in this._dHosts) {
        let host = this._dHosts[key];
        if (host.open) {
          // log.d("saveState: open " + host.rawHost);
          this._openHosts.push(host.rawHost);
        }
      }
    }
  },

  /**
   * Restore a previously saved state
   */
  restoreState: function() {
    // Restore open states
    var rc = this.rowCount;
    if (this.viewMode == VIEW_MODE.TREE && this._openHosts.length > 0) {
      for (let i=0; i<rc; i++) {
        // log.d("restoreState open: test item at " + i);
        var item = this.getItemAtIndex(i);
        if (item != null && item.getType() == COOKIE_TYPE_CONTAINER && this._openHosts.indexOf(item.rawHost) > -1) {
          if (item.open == false) {
            this.toggleOpenState(i);
            rc = this.rowCount;
          }
        }
      }
      this._openHosts = [];
    }

    var seln      = this.selection;
    var nSelected = this._selectedItems.length;
    rc            = this.rowCount;
    seln.clearSelection();

    if (nSelected == 0) {
      // nothing previously selected, no need to go further
      if (rc > 0) {
        seln.select(0)
      }
      return;
    }

    if (this.viewMode == VIEW_MODE.TREE) {
      let selItem = this._selectedItems[0];
      let parent  = -1;
      let found = false;
      if (selItem != null) {
        for (let i=0; i<rc; i++) {
          // log.d("restoreState selection: test item at " + i);
          let item = this.getItemAtIndex(i);
          if (item != null) {
            if (item.equals(selItem)) {
              // log.d("restoreState Found: ", item);
              this._treeBoxObj.ensureRowIsVisible(i);
              seln.select(i);
              i = rc; // end loop
              found = true;
            } else if (item.getType() == COOKIE_TYPE_CONTAINER && item.rawHost == selItem.rawHost) {
              parent = i; // Save the parent index, in case the item is not displayed
              if (!item.open) {
                // If the item may be in this container, open it
                this.toggleOpenState(i);
                rc = this.rowCount;
              }
            }
          }
        }
      }
      // Fallback if we do not have found our happiness
      if (!found) {
        let idx = 0;
        if (parent != -1) {
          // We have the parent, but not the item itself
          idx = parent;
        } else if (this._selectedIndex > -1) {
          // We have nothing, just reselect the same index
          if (this._selectedIndex < this.rowCount) {
            idx = this._selectedIndex;
          } else if (this.rowCount > 0) {
            idx = this.rowCount-1;
          }
        }
        seln.select(idx);
        this._treeBoxObj.ensureRowIsVisible(idx);
      }
    } else {
      // Using the _flatSet[]
      if (rc != this._flatSet.length) {
        log.e("Restore state: rowCount != _flatSet.length", rc, this._flatSet.length);
        rc = this._flatSet.length;
        this.rowCount = rc;
      }

      let nDone = 0;
      let lastIndex = this._selectedIndex;
      for (let i=0; i<rc && nDone < nSelected; i++) {
        let item = this._flatSet[i];
        for (let j=0; j<nSelected; j++) {
          let selItem = this._selectedItems[j];
          // log.d("restoreState: test ", item, " with selected ", selItem);
          if (selItem != null && selItem.equals(item)) {
            // log.d("restoreState select " + i);
            seln.rangedSelect(i, i, true);
            j = nSelected; // end loop
            nDone++;
            lastIndex = i;
          }
        }
      }
      if (nDone == 0 && this._selectedIndex > -1) {
        let idx = 0;
        if (this._selectedIndex < this.rowCount) {
          idx = this._selectedIndex;
        } else if (this.rowCount > 0) {
          idx = this.rowCount-1;
        }
        seln.select(idx);
        this._treeBoxObj.ensureRowIsVisible(idx);
      } else {
        this._treeBoxObj.ensureRowIsVisible(lastIndex);
      }
    }
    this._selectedIndex = -1;
  },

  getItemAtIndex: function (aIndex) {
    if (DEBUG) {
      if (typeof(aIndex) != "number" || aIndex < 0) {
        log.e("gCookiesTreeView.getItemAtIndex: invalide parameter aIndex: ", aIndex, typeof(aIndex));
        var err = new Error();
        log.e(err.stack); // print stack trace
        return null;
      } else if (this !== gCookiesTreeView) {
        log.e("gCookiesTreeView.getItemAtIndex: invalide context");
        return null;
      }
    }

    if (this.viewMode == VIEW_MODE.FLAT) {
      if (aIndex >= this._flatSet.length) {
        return null;
      }
      return this._flatSet[aIndex];
    }

    var start = 0;
    var count = 0;
    var hostIndex = 0;

    var cacheIndex = Math.min(this._cacheValid, aIndex);
    if (cacheIndex > 0) {
      let cacheItem = this._cacheItems[cacheIndex];
      start = cacheItem['start'];
      count = hostIndex = cacheItem['count'];
    }

    const hoLength = this._dHostsOrder.length;
    for (let i = start; i < hoLength; i++) {
      let currHost = this._dHosts[this._dHostsOrder[i]];
      if (count == aIndex) { return currHost; }
      hostIndex = count;

      let cacheEntry = { 'start' : i, 'count' : count };
      let cacheStart = count;

      if (currHost.open) {
        if (count < aIndex && aIndex <= (count + currHost.items.length)) {
          // We are looking for an entry within this host's children,
          // enumerate them looking for the index.
          count++;
          for (let j=0; j<currHost.items.length; j++) {
            if (count == aIndex) {
              let item = currHost.items[j];
              item.parentIndex = hostIndex;
              return item;
            }
            count++;
          }
        } else {
          // A host entry was open, but we weren't looking for an index
          // within that host entry's children, so skip forward over the
          // entry's children. We need to add one to increment for the
          // host value too.
          count += currHost.items.length + 1;
        }
      } else {
        count++;
      }

      for (let j = cacheStart; j < count; j++) {
        this._cacheItems[j] = cacheEntry;
      }
      this._cacheValid = count - 1;
    }
    return null;
  },

  getIndexOfItem: function(aItem) {
    if (aItem == null || typeof(aItem) == "undefined") {
      // param check
      return -1;
    }

    if (this.viewMode == VIEW_MODE.TREE) {
      let hostOrder = this._dHostsOrder
      let hosts     = this._dHosts;
      let nHosts = hostOrder.length;
      var index = 0;
      for (let cHost = 0; cHost < nHosts; cHost++) {
        let aHost = hosts[hostOrder[cHost]];
        let isHost = aItem.rawHost == aHost.rawHost;
        if (aItem.getType() == COOKIE_TYPE_CONTAINER && isHost) {
          return index;
        }
        index++;
        if (isHost) {
          if (!aHost.open) {
            return -1;
          }
          let nItems = aHost.items.length;
          for (let cItem=0; cItem<nItems; cItem++) {
            let item = aHost.items[cItem];
            if (item.equals(aItem)) {
              return index;
            }
            index++;
          }
          // not found!
          return -1;
        } else if (aHost.open) {
          index += aHost.items.length;
        }
      }
    } else if (aItem.getType() == COOKIE_TYPE_CONTAINER) {
      return -1;
    } else {
      const nItems = this._flatSet.length;
      for (let cItem = 0; cItem < nItems; cItem++) {
        let item = this._flatSet[cItem];
        if (item.equals(aItem)) {
          return cItem;
        }
      }
    }
    return -1;
  },

  _invalidateCache: function (aIndex) {
    this._cacheValid = Math.min(this._cacheValid, aIndex);
    if (this._cacheValid <= 0) {
      this._cacheItems = [];
    }
  },

  _getCookieCellValue: function(aItem, aColumnId) {
    if (aColumnId == SORT_INFOS.HOST.column) {
      let host = aItem.getHost();
      if (gCommon.isSpecialHost(host)) {
        host = gCookieKeeper._bundle.getString(host.substring(1, host.length-1));
      }
      return host;

    } else if (aColumnId == SORT_INFOS.NAME.column) {
      return aItem.getName();
    } else if (aColumnId == SORT_INFOS.PROTECTED.column) {
      return aItem.isProtected ? gCookieKeeper._yes : gCookieKeeper._no;
    } else if (aColumnId == SORT_INFOS.HTTP.column) {
      return aItem.isHttpOnly ? gCookieKeeper._yes : gCookieKeeper._no;
    } else if (aColumnId == SORT_INFOS.SECURE.column) {
      return aItem.isSecure ? gCookieKeeper._yes : gCookieKeeper._no;
    } else if (aColumnId == SORT_INFOS.SESSION.column) {
      return aItem.isSession ? gCookieKeeper._yes : gCookieKeeper._no;
    } else if (aColumnId == SORT_INFOS.VALUE.column) {
      return aItem.value.length > 64 ? aItem.value.substring(0, 61) + "..." : aItem.value;
    } else if (aColumnId == SORT_INFOS.EXPIRES.column) {
      let expDate = aItem.isSession ? aItem.expiry : aItem.expires;
      return expDate >= DATE_MAX ?
              this._strDateMax :
              gCommon.formatDateString(expDate, null); // gCookieKeeper._bundle.getString("expireAtEndOfSession")
    } else if (aColumnId == SORT_INFOS.CREATION.column) {
      switch (aItem.getType()) {
      case COOKIE_TYPE_STANDARD:
      case COOKIE_TYPE_LSO:
        return gCommon.formatDateString(aItem.creationTime, null);
      case COOKIE_TYPE_STORAGE:
        return this._strDateUnknown;
      default:
        log.e("Unknown cookie type: " + aItem.getType());
        return "<Error>";
      }
    } else if (aColumnId == SORT_INFOS.LAST_ACCESSED.column) {
      switch (aItem.getType()) {
      case COOKIE_TYPE_STANDARD:
        return gCommon.formatDateString(aItem.lastAccessed, null);
      case COOKIE_TYPE_STORAGE:
      case COOKIE_TYPE_LSO:
        return this._strDateUnknown;
      default:
        log.e("Unknown cookie type: " + aItem.getType());
        return "<Error>";
      }
    } else if (aColumnId == SORT_INFOS.PATH.column) {
      return aItem.path;
    } else {
      return "n/a";
    }
  },

  // -------------------------------------------------
  // Tree view implementation
  // -------------------------------------------------

  getCellText: function (aIndex, aColumn) {
    if (this.viewMode == VIEW_MODE.TREE) {
      var item = this.getItemAtIndex(aIndex);

      if (item == null) {
        return "";
      }

      if (item.getType() == COOKIE_TYPE_CONTAINER) {
        if (aColumn.id == SORT_INFOS.HOST.column) {
          let host = item.rawHost;
          if (gCommon.isSpecialHost(host)) {
            host = gCookieKeeper._bundle.getString(host.substring(1, host.length-1));
          }
          return host;
        } else {
          return "";
        }
      } else {
        return this._getCookieCellValue(item, aColumn.id);
      }

    } else if (aIndex < this._flatSet.length) {
      var item = this._flatSet[aIndex];
      return this._getCookieCellValue(item, aColumn.id);
    } else {
      log.e("aIndex " + aIndex + " >= this._flatSet.length " + this._flatSet.length + " rowCount=" + this.rowCount);
    }
    return "";
  },

  get selection () {
    return this._selection;
  },
  set selection (val) {
    this._selection = val;
    return val;
  },

  getRowProperties: function (aIndex) {
    /*
     // FIXME set protected icon with CSS. Don't work for now, use getImageSrc() instead
    var item = this.getItemAtIndex(aIndex);
    if (item == null || item.getType() == COOKIE_TYPE_CONTAINER || !item.isProtected) {
      return "";
    }
    if (item.isProtected) {
      return "protected";
    }*/
    return "";
  },
  getCellProperties: function (aIndex, aColumn) { return ""; },
  getColumnProperties: function (aColumn) { return ""; },
  isContainer: function (aIndex) {
    let ret = false;
    if (this.viewMode == VIEW_MODE.TREE) {
      var item = this.getItemAtIndex(aIndex);
      if (item != null) { ret = item.getType() == COOKIE_TYPE_CONTAINER; }
    }
    return ret;
  },
  isContainerOpen: function (aIndex) {
    let ret = false;
    if (this.viewMode == VIEW_MODE.TREE) {
      var item = this.getItemAtIndex(aIndex);
      if (item != null) { ret = item.open; }
    }
    return ret;
  },
  isContainerEmpty: function (aIndex) {
    let ret = true;
    if (this.viewMode == VIEW_MODE.TREE) {
      var item = this.getItemAtIndex(aIndex);
      if (item != null && item.getType() == COOKIE_TYPE_CONTAINER) { ret = (item.items.length == 0); }
    }
    return ret;
  },
  isSeparator: function (aIndex) { return false; },
  isSorted: function (aIndex) { return false; },
  canDrop: function (aIndex, aOrientation) { return false; },
  drop: function (aIndex, aOrientation) {},
  getParentIndex: function (aIndex) {
    if (this.viewMode == VIEW_MODE.TREE) {
      var item = this.getItemAtIndex(aIndex);
      // If an item has no parent index (i.e. it is at the top level) this
      // function MUST return -1 otherwise we will go into an infinite loop.
      // Containers are always top level items in the cookies tree, so make
      // sure to return the appropriate value here.
      if (item != null && item.getType() != COOKIE_TYPE_CONTAINER) {
        return item.parentIndex;
      }
    }
    return -1;
  },
  hasNextSibling: function (aParentIndex, aIndex) {
    if (this.viewMode == VIEW_MODE.TREE) {
      // |aParentIndex| appears to be bogus, but we can get the real
      // parent index by getting the entry for |aIndex| and reading the
      // parentIndex field.
      // The index of the last item in this host collection is the
      // index of the parent + the size of the host collection, and
      // aIndex has a next sibling if it is less than this value.
      var item = this.getItemAtIndex(aIndex);
      if (item != null) {
        if (item.getType() == COOKIE_TYPE_CONTAINER) {
          for (let i = aIndex + 1; i < this.rowCount; i++) {
            var subsequent = this.getItemAtIndex(i);
            if (subsequent.getType() == COOKIE_TYPE_CONTAINER) { return true; }
          }
          return false;
        } else {
          var parent = this.getItemAtIndex(item.parentIndex);
          if (parent && parent.getType() == COOKIE_TYPE_CONTAINER) {
            return aIndex < (item.parentIndex + parent.items.length);
          }
        }
      }
    }
    return aIndex < (this.rowCount - 1);
  },
  hasPreviousSibling: function (aIndex) {
    if (this.viewMode == VIEW_MODE.TREE) {

      var item = this.getItemAtIndex(aIndex);
      if (item == null) { return false; }

      var parent = this.getItemAtIndex(item.parentIndex);
      if (parent != null && parent.getType() == COOKIE_TYPE_CONTAINER) {
        return aIndex > (item.parentIndex + 1);
      }
    }
    return aIndex > 0;
  },
  getLevel: function (aIndex) {
    if (this.viewMode == VIEW_MODE.TREE) {
      var item = this.getItemAtIndex(aIndex);
      if (item != null) { return item.level; }
    }
    return 0;
  },
  getImageSrc: function (aIndex, aColumn) {
    if (aColumn.index == 0) {
      var item = this.getItemAtIndex(aIndex);
      if (item != null && item.getType() != COOKIE_TYPE_CONTAINER) {
        if (item.getType() == COOKIE_TYPE_LSO) {
          if (item.isProtected) {
            return "chrome://cookiekeeper_os/skin/lsoSecure16.png";
          }
          return "chrome://cookiekeeper/skin/lso16.png";
        } else if (item.getType() == COOKIE_TYPE_STORAGE) {
          if (item.isProtected) {
            return "chrome://cookiekeeper_os/skin/domSecure16.png";
          }
          return "chrome://cookiekeeper/skin/html5badge16.png";
        } else if (item.isProtected && item.isSession) {
          return "chrome://cookiekeeper/skin/warning16.png";
        } else if (item.isProtected) {
          return "chrome://cookiekeeper_os/skin/secure16.png";
        }
      }
    }
    return "";
  },
  getProgressMode: function (aIndex, aColumn) {},
  getCellValue: function (aIndex, aColumn) {},
  setTree: function (aTree) {
    this._treeBoxObj = aTree;
  },
  toggleOpenState: function (aIndex) {
    if (this.viewMode == VIEW_MODE.TREE) {
      let item = this.getItemAtIndex(aIndex);
      if (item != null && item.getType() == COOKIE_TYPE_CONTAINER) {
        this._invalidateCache(aIndex);
        let multiplier = item.open ? -1 : 1;
        let delta = multiplier * item.items.length;
        this._rowCount += delta;
        item.open = !item.open;
        this._treeBoxObj.rowCountChanged(aIndex + 1, delta);
        this._treeBoxObj.invalidateRow(aIndex);
      }
    }
  },
  toggleAllOpenStates: function () {
    if (this.viewMode == VIEW_MODE.TREE) {
      this._openState = !this._openState;
      for (let key in this._dHosts) {
        let item = this._dHosts[key];
        item.open = this._openState;
      }
      this.saveState();
      this.refreshView(true);

      return true;
    }
    return false;
  },
  cycleHeader: function (aColumn) {},
  selectionChanged: function () {},
  cycleCell: function (aIndex, aColumn) {},
  isEditable: function (aIndex, aColumn) {
    return false;
  },
  isSelectable: function (aIndex, aColumn) {
    return false;
  },
  setCellValue: function (aIndex, aColumn, aValue) {},
  setCellText: function (aIndex, aColumn, aValue) {},
  performAction: function (aAction) { log.d("performAction", aAction); },
  performActionOnRow: function (aAction, aIndex) { log.d("performActionOnRow", aAction, aIndex); },
  performActionOnCell: function (aAction, aindex, aColumn) { log.d("performActionOnCell", aAction, aIndex, aColumn); }
};
