/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Common definitions
 *
 * ************************************************************************
 */

"use strict";

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Cu = Components.utils;

// show debug message in browser log window
const DEBUG                       = true;
/// Change only if the preferences need to be updated
/// If this value is changed, don't forget to update the default pref value in:
/// cookiekeeper/defaults/preferences.js
const CK_PREF_VERSION             = 4;

// https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Date
// 100,000,000 days relative to 01 January, 1970 UTC * 86,400,000 milliseconds
const DATE_MAX                    = 100000000 * 86400000;
const DATE_MAX_SECONDS            = 100000000 * 86400;

// Preferences constants:
const PREF_BRANCH                 = "extensions.cookiekeeper.";
const PREF_VERSION                = "extensions.cookiekeeper.version";
const PREF_CLEAR_LSO_ON_DEMAND    = "extensions.cookiekeeper.clear_on_demand.lso";
const PREF_CLEAR_HSTS_ON_DEMAND   = "extensions.cookiekeeper.clear_on_demand.hsts";
const PREF_CLEAR_COOKIES_AT_EXIT  = "extensions.cookiekeeper.clear_at_exit.cookies";
const PREF_CLEAR_DOM_AT_EXIT      = "extensions.cookiekeeper.clear_at_exit.dom";
const PREF_CLEAR_LSO_AT_EXIT      = "extensions.cookiekeeper.clear_at_exit.lso";
const PREF_CLEAR_HSTS_AT_EXIT     = "extensions.cookiekeeper.clear_at_exit.hsts";
const PREF_LSO_KEEP_SETTINGS      = "extensions.cookiekeeper.lso.always_keep_main_settings";
const PREF_NO_ASK_AGAIN_DEL_PROT  = "extensions.cookiekeeper.no_ask_again.del_prot";
const PREF_NO_ASK_AGAIN_CHEXT     = "extensions.cookiekeeper.no_ask_again.change_ext";
const PREF_EXPORT_CHANGE_EXT      = "extensions.cookiekeeper.export.change_ext";
const PREF_EXPORT_FORMAT          = "extensions.cookiekeeper.export.format";
// show the boxFixPrefClear if mozilla pref can messed up our config:
const PREF_WARN_MOZ               = "extensions.cookiekeeper.warn_moz_pref";
const PREF_WARN_PROT_SESSION      = "extensions.cookiekeeper.warn_prot_session";
const PREF_NO_AUSTRALIS_UI        = "extensions.cookiekeeper.no_australis_ui"; // Firefox only
const PREF_ENABLE_PRIV_BROWSE     = "extensions.cookiekeeper.enable_in_private_browsing";
// Old protected cookie string format: |host;name;path|host;name;path|...
const PREF_PROT_COOKIES           = "extensions.cookiekeeper.protected";
const PROT_SEP_COOKIE             = "|";
const PROT_SEP_DATA               = ";";
// Version 1.8.0: JSON file in the user profile:
// If you change the filename, don't forget to update the
// bootstrap.js uninstall() function.
const PROT_JSON_FILE              = "cookiekeeper.json";
// Exported formats:
const CK_EXPORT_CSV               = "CSV";
const CK_EXPORT_JSON              = "JSON";

// Moz App names:
const FIREFOX                     = "Firefox";
const THUNDERBIRD                 = "Thunderbird";
const SEAMONKEY                   = "SeaMonkey";
const PALEMOON                    = "Palemoon";
const KMELEON                     = "K-Meleon";

const CK_BUTTON_ID                = "cookiekeeper-button";
const CK_MENU_ID                  = "cookiekeeper-tools-menu";
const CK_MENU_PANEL_ID            = "PanelUI-cookiekeeper";
const CK_WINDOWS_ID               = ["cookieKeeperWindow",
                                    "cookieKeeperPrefsWindow",
                                    "cookieKeeperEditorCookie",
                                    "cookieKeeperEditorDomStorage",
                                    "cookieKeeperEditorLSO"];
const CK_WINDOW_MAIN_IDX          = 0;
const CK_WINDOW_PREFS_IDX         = 1;

const CK_PREFPANE_OPTIONS         = "cookiekeeper-prefs-pane-cookiekeeper";
const CK_PREFPANE_PCOOKIES        = "cookiekeeper-prefs-pane-pcookies";
const CK_PREFPANE_PERMISSIONS     = "cookiekeeper-prefs-pane-permissions";
const CK_PREFPANE_FLASH           = "cookiekeeper-prefs-pane-flash";
const CK_PREFPANE_BROWSER         = "cookiekeeper-prefs-pane-browser";

const CK_MODULE_NAME              = "CookieKeeper";
const CK_MODULE_VERSION           = "1.9.0";

// Observer ProtectedCookies:
const CK_OBS_PCOOKIE_TOPIC        = "cookiekeeper-change";
const CK_OBS_PCOOKIE_STARTUP      = "startup";  // The add-on is starting
const CK_OBS_PCOOKIE_SHUTDOWN     = "shutdown"; // The add-on is shutdown
const CK_OBS_PCOOKIE_ADD          = "protection-added";
const CK_OBS_PCOOKIE_DEL          = "protection-deleted";
const CK_OBS_PCOOKIE_DOM_CLEAN    = "dom-storage-clean";
const CK_OBS_PCOOKIE_DOM_DELETE   = "dom-storage-batch-deleted";
const CK_OBS_PCOOKIE_DOM_INSERT   = "dom-storage-batch-inserted-or-updated";
// const CK_OBS_PCOOKIE_DOM_UPDATE   = "dom-storage-batch-updated";
const CK_OBS_PCOOKIE_LSO_DELETE   = "lso-batch-deleted";

// Mozilla preferences:
// from mozilla-release/netwerk/cookie/nsCookieService.cpp (Firefox 24.0):
// values for network.cookie.cookieBehavior:
const MOZ_BEHAVIOR_ENABLED        = 0; // allow all cookies (default)
const MOZ_BEHAVIOR_SRV_ONLY       = 1; // reject all third-party cookies
const MOZ_BEHAVIOR_DISABLED       = 2; // reject all cookies
const MOZ_BEHAVIOR_P3P            = 3; // reject third-party cookies unless the eTLD already has at least one cookie
// from mozilla-release/extensions/cookie/nsCookiePermission.cpp (Firefox 24.0)
// values for network.cookie.lifetimePolicy:
const MOZ_ACCEPT_NORMALLY         = 0; // accept normally
const MOZ_ASK_BEFORE_ACCEPT       = 1; // ask before accepting
const MOZ_ACCEPT_SESSION          = 2; // downgrade to session
const MOZ_ACCEPT_FOR_N_DAYS       = 3; // limit lifetime to N days

// Cookies permissions
const MOZ_COOKIE_PERM_DEFAULT     = Ci.nsICookiePermission.ACCESS_DEFAULT;
const MOZ_COOKIE_PERM_ALLOW       = Ci.nsICookiePermission.ACCESS_ALLOW;
const MOZ_COOKIE_PERM_DENY        = Ci.nsICookiePermission.ACCESS_DENY;
const MOZ_COOKIE_PERM_FIRST       = Ci.nsICookiePermission.ACCESS_ALLOW_FIRST_PARTY_ONLY;
const MOZ_COOKIE_PERM_SESSION     = Ci.nsICookiePermission.ACCESS_SESSION;

// Dom Storage database
const DOM_DB_NAME                 = "webappsstore.sqlite";

// Types of supported cookies
const COOKIE_TYPE_CONTAINER       = 0; // A container (directory/folder)
const COOKIE_TYPE_STANDARD        = 1; // A standard cookie
const COOKIE_TYPE_STORAGE         = 2; // DOM Storage
const COOKIE_TYPE_LSO             = 3; // Local shared object (LSOs) / Flash cookies.

/*
// pref string constants
static const char kPrefCookieBehavior[]     = "network.cookie.cookieBehavior";
static const char kPrefMaxNumberOfCookies[] = "network.cookie.maxNumber";
static const char kPrefMaxCookiesPerHost[]  = "network.cookie.maxPerHost";
static const char kPrefCookiePurgeAge[]     = "network.cookie.purgeAge";
static const char kPrefThirdPartySession[]  = "network.cookie.thirdparty.sessionOnly";

// default limits for the cookie list. these can be tuned by the
// network.cookie.maxNumber and network.cookie.maxPerHost prefs respectively.
static const uint32_t kMaxNumberOfCookies = 3000;
static const uint32_t kMaxCookiesPerHost  = 150;
static const uint32_t kMaxBytesPerCookie  = 4096;
static const uint32_t kMaxBytesPerPath    = 1024;
*/

const XUL_NS  = "http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul";
const HTML_NS = "http://www.w3.org/1999/xhtml";

// XUL Accessible Type
const XULMenuitem      = "100E";
const XULMenupopup     = "100F";
const XULMenuSeparator = "1010";
const XULToolbarButton = "1023";

const APP_MOZ_CONST = {
  // Firefox
  "{ec8030f7-c20a-464f-9b0e-13a3a9e97384}": {
    NAME:                       FIREFOX,
    PRIVATEBROWSING:            "browser.privatebrowsing.autostart",
    SANITIZE_ON_SHUTDOWN:       "privacy.sanitize.sanitizeOnShutdown",
    COOKIES_CLEAR_ON_SHUTDOWN:  "privacy.clearOnShutdown.cookies",
    COOKIES_BEHAVIOR:           "network.cookie.cookieBehavior",
    COOKIES_LIFETIME:           "network.cookie.lifetimePolicy",
    COOKIES_LIFETIME_DAYS:      "network.cookie.lifetime.days",
    WINDOW_TYPE:                "navigator:browser", // main window attribute "windowtype"
    TOOLBAR:                    "nav-bar",  // default toolbar
    TOOLBOX:                    "navigator-toolbox",
    PALETTE:                    "BrowserToolbarPalette",
    TOOLSMENU:                  "menu_ToolsPopup",
    PANELUI:                    "PanelUI-multiView" // "mainPopupSet"
  },

  // SeaMonkey
  "{92650c4d-4b8e-4d2a-b7eb-24ecf4f6b63a}": {
    NAME:                       SEAMONKEY,
    SANITIZE_ON_SHUTDOWN:       "privacy.sanitize.sanitizeOnShutdown",
    COOKIES_CLEAR_ON_SHUTDOWN:  "privacy.item.cookies",
    COOKIES_BEHAVIOR:           "network.cookie.cookieBehavior",
    COOKIES_LIFETIME:           "network.cookie.lifetimePolicy",
    COOKIES_LIFETIME_DAYS:      "network.cookie.lifetime.days",
    WINDOW_TYPE:                "navigator:browser",
    TOOLBAR:                    "nav-bar",
    TOOLBOX:                    "navigator-toolbox",
    PALETTE:                    "BrowserToolbarPalette",
    TOOLSMENU:                  "taskPopup",
    PANELUI:                    null
  },

  // Thunderbird
  "{3550f703-e582-4d05-9a08-453d09bdfdc6}": {
    NAME:                       THUNDERBIRD,
    COOKIES_BEHAVIOR:           "network.cookie.cookieBehavior",
    COOKIES_LIFETIME:           "network.cookie.lifetimePolicy",
    COOKIES_LIFETIME_DAYS:      "network.cookie.lifetime.days",
    WINDOW_TYPE:                "mail:3pane",
    TOOLBAR:                    "mail-bar3",
    TOOLBOX:                    "mail-toolbox",
    PALETTE:                    "MailToolbarPalette",
    TOOLSMENU:                  "taskPopup",
    PANELUI:                    null
  },

  // Palemoon
  "{8de7fcbb-c55c-4fbe-bfc5-fc555c87dbc4}": {
    NAME:                       PALEMOON,
    PRIVATEBROWSING:            "browser.privatebrowsing.autostart",
    SANITIZE_ON_SHUTDOWN:       "privacy.sanitize.sanitizeOnShutdown",
    COOKIES_CLEAR_ON_SHUTDOWN:  "privacy.clearOnShutdown.cookies",
    COOKIES_BEHAVIOR:           "network.cookie.cookieBehavior",
    COOKIES_LIFETIME:           "network.cookie.lifetimePolicy",
    COOKIES_LIFETIME_DAYS:      "network.cookie.lifetime.days",
    WINDOW_TYPE:                "navigator:browser", // main window attribute "windowtype"
    TOOLBAR:                    "nav-bar",  // default toolbar
    TOOLBOX:                    "navigator-toolbox",
    PALETTE:                    "BrowserToolbarPalette",
    TOOLSMENU:                  "menu_ToolsPopup",
    PANELUI:                    "PanelUI-multiView" // "mainPopupSet"
  },

  // K-Meleon
  "{944df56f-c0ec-4d7b-94b6-f625c8cc96ea}": {
    NAME:                       KMELEON,
    PRIVATEBROWSING:            "browser.privatebrowsing.autostart",
    SANITIZE_ON_SHUTDOWN:       "privacy.sanitize.sanitizeOnShutdown",
    COOKIES_CLEAR_ON_SHUTDOWN:  "privacy.clearOnShutdown.cookies",
    COOKIES_BEHAVIOR:           "network.cookie.cookieBehavior",
    COOKIES_LIFETIME:           "network.cookie.lifetimePolicy",
    COOKIES_LIFETIME_DAYS:      "network.cookie.lifetime.days",
    WINDOW_TYPE:                "navigator:browser", // main window attribute "windowtype"
    TOOLBAR:                    "nav-bar",  // default toolbar
    TOOLBOX:                    "navigator-toolbox",
    PALETTE:                    "BrowserToolbarPalette",
    TOOLSMENU:                  "menu_ToolsPopup",
    PANELUI:                    "PanelUI-multiView" // "mainPopupSet"
  }
}

/**
 * Logging facility (thanks to CustomizableUI.jsm)
 */
var log = {

  d: function(...args) {
    if (DEBUG) {
      if (args) {
        args.unshift("[" + CK_MODULE_NAME + " " + CK_MODULE_VERSION + "]");
        console.log.apply(console, args);
      } else {
        console.log("null");
      }
    }
  },

  i: function(...args) {
    if (DEBUG) {
      args.unshift("[" + CK_MODULE_NAME + " " + CK_MODULE_VERSION + "]");
      console.info.apply(console, args);
    }
  },

  w: function(...args) {
    if (DEBUG) {
      args.unshift("[" + CK_MODULE_NAME + " " + CK_MODULE_VERSION + "]");
      console.warn.apply(console, args);
    }
  },

  e: function(...args) {
    if (DEBUG) {
      args.unshift("[" + CK_MODULE_NAME + " " + CK_MODULE_VERSION + "]");
      // let err = new Error();
      // args.push("\n" + err.stack);
      console.error.apply(console, args);
    }
  }
};

/**
 * Some functions used by others CookieKeeper objects
 */
var gCommon = {
  _srvDFormat   : Cc["@mozilla.org/intl/scriptabledateformat;1"].getService(Ci.nsIScriptableDateFormat),
  // quick but innacurate ip regex:
  _reHostIsIPv4 : /^[0-9]{0,3}\.[0-9]{0,3}\.[0-9]{0,3}\.[0-9]{0,3}$/,
  _reHostIsIPv6 : /^(::){0,1}([0-9a-fA-F]{1,4}:)+[0-9a-fA-F]{1,4}$/,

  isSpecialHost: function(aHost) {
    return typeof(aHost) == "string" && aHost.charAt(0) == "<";
  },

  /**
   * @brief make a stipped host (www.foo-bar.org -> foo-bar.org)
   * @param [string] aHost The host to strip
   * @return [string] the stripped host
   * @note IP adresse are unchanged.
   */
  makeStrippedHost: function(aHost) {
    if (this.isSpecialHost(aHost)) {
      // Special cookie, like flash main settings
      return aHost;
    } else if (gCommon._reHostIsIPv4.test(aHost)) {
      return aHost;
    } else if (gCommon._reHostIsIPv6.test(aHost)) {
      return "[" + aHost + "]";
    } else {
      try {
        let strippedHost = aHost;
        if (aHost.indexOf(".") >= 0) {
          strippedHost = aHost.indexOf(".") == 0 ? aHost.substring(1) : aHost;
          strippedHost = Services.eTLD.getBaseDomainFromHost(strippedHost);
        } // else maybe a local domain, like 'localhost'
        return strippedHost;
      } catch (ex) {
        let arr = aHost.split(".").reverse();
        return arr[1] + "." + arr[0];
      }
    }
  },

  reverseStr: function(str) {
    let rts = "";
    if (typeof(str) != "string") {
      str = str.toString();
    }
    for (let i=str.length-1; i>=0; i--) {
      rts = rts + str[i];
    }
    return rts;
  },

  cmpHostName: function(aHost, bHost) {
    let c = 0;
    let ha  = aHost.split(".");
    let hb  = bHost.split(".");
    ha = ha.reverse();
    hb = hb.reverse();
    let min = ha.length < hb.length ? ha.length : hb.length;

    for (let i=1; i<min && c == 0; i++) {
      c = ha[i].toLowerCase().localeCompare(hb[i].toLowerCase());
    }
    if (c == 0) {
      c = ha[0].toLowerCase().localeCompare(hb[0].toLowerCase());
    }
    if (c == 0) {
      c = ha.length - hb.length;
    }
    return c;
  },

  formatDateString: function(aDate, defaultString) {
    if (typeof(aDate) == "number") {
      try {
        var date = new Date(aDate);
        return this._srvDFormat.FormatDateTime("", this._srvDFormat.dateFormatLong,
                                       this._srvDFormat.timeFormatSeconds,
                                       date.getFullYear(),
                                       date.getMonth() + 1,
                                       date.getDate(),
                                       date.getHours(),
                                       date.getMinutes(),
                                       date.getSeconds());
      } catch (e) {
        log.e("_formatDateString(" + aDate + ") exception:" + e);
        return "" + aDate;
      }
    } else if (typeof(defaultString) == "string") {
      return defaultString;
    } else {
      return "<error " + aDate + ">";
    }
  },

  /// Transform a size in a human readable size (like: 105.3 Kib)
  getHumanSize: function(aSize, aBundle) {
    const BUNDLE_SIZE = ["size.byte", "size.bytes", "size.kilobytes", "size.megabytes"];
    if (!typeof(aSize) == "number") {
      return "";
    }
    if (aSize <= 1) {
      let str = aBundle.getString(BUNDLE_SIZE[0]);
      return str.replace("%n", aSize.toString(10));
    }

    let tmpSize = aSize;
    let tmpMod  = 0|0;
    let nDiv    = 1;
    while (tmpSize > 1024 && nDiv < BUNDLE_SIZE.length) {
      tmpMod  = (tmpSize % 1024)|0; // Force 32bits integer (asm.js tip)
      tmpSize = (tmpSize / 1024);
      nDiv++;
    }
    let str = aBundle.getString(BUNDLE_SIZE[nDiv]);

    if (tmpMod != 0) {
      tmpSize = tmpSize.toFixed(1);
    }
    return str.replace("%n", tmpSize.toLocaleString(10));
  }
};
