/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Preferences tab script.
 *
 * ************************************************************************
 */
"use strict";

Components.utils.import("resource://gre/modules/Services.jsm");
Components.utils.import("resource://gre/modules/Promise.jsm");
Components.utils.import("resource://gre/modules/NetUtil.jsm");
Components.utils.import("resource://cookiekeeper/cookiekeeper.jsm");

// Topic for the preferences change observer
const NS_PREFBRANCH_PREFCHANGE_TOPIC_ID  = "nsPref:changed";
// Custom DOM event, used by prefs.browser.js, fired by prefsObserver
const EVENT_CUSTOM_CHANGED = "changed";

const TAB1_CKPREFS           = 1;
const TAB2_PROTECTED_COOKIES = 2;
const TAB3_PERMISSIONS       = 3;
const TAB4_BROWSER_PREFS     = 4;
const TAB_A_PREFIX           = "a_tab";
const TAB_DIV_ID             = ["div_content_ck", "div_content_protections", "div_content_permissions", "div_content_browser"];

const PREF_MATCH_OS_LOCALE  = "intl.locale.matchOS";
const PREF_SELECTED_LOCALE  = "general.useragent.locale";
const PREF_ACCEPT_LANGUAGES = "intl.accept_languages";

// code mostly from the Mozilla's addon sdk
var addonSdk = {
  readURI: function(_uri) {
    let channel;
    let { promise, resolve, reject } = Promise.defer();

    try {
      // Current Gecko code
      channel = NetUtil.newChannel({
        uri : NetUtil.newURI(_uri, 'UTF-8', null),
        loadUsingSystemPrincipal : true});
    } catch (ex) {
      // For older gecko, like palemoon
      channel = NetUtil.newChannel(_uri, 'UTF-8', null);
    }

    try {
      NetUtil.asyncFetch(channel, function (stream, result) {
        if (Components.isSuccessCode(result)) {
          let count = stream.available();
          let data = NetUtil.readInputStreamToString(stream, count, { charset : 'UTF-8' });
          resolve(data);
        } else {
          reject("Failed to read: '" + _uri + "' (Error Code: " + result + ")");
        }
      });
    } catch (e) {
      reject("Failed to read: '" + _uri + "' (Error: " + e.message + ")");
    }

    return promise;
  },
  parseJsonURI: function(uri) {
    return this.readURI(uri).
      then(JSON.parse).
      then(null, function (error) {
        throw Error("Failed to parse locale file:\n" + uri + "\n" + error);
      });
  },
  prefs : {
    get: function(name, defaultValue) {
      switch (Services.prefs.getPrefType(name)) {
        case Services.prefs.PREF_INVALID:
          return defaultValue;
        case Services.prefs.PREF_STRING:
          return Services.prefs.getCharPref(name);
        case Services.prefs.PREF_INT:
          return Services.prefs.getIntPref(name);
        case Services.prefs.PREF_BOOL:
          return Services.prefs.getBoolPref(name);
        default:
          throw "Unknown pref type for " + name;
      }
    },
    set: function(name, value) {
      if (typeof(value) == "boolean") {
        Services.prefs.setBoolPref(name, value);
      } else if (typeof(value) == "number") {
        Services.prefs.setIntPref(name, value);
      } else {
        log.d("TODO set pref of type " + typeof(value));
      }
    },
    getLocalized: function(name, defaultValue) {
      let value = null;
      try {
        value = Services.prefs.getComplexValue(name, Ci.nsIPrefLocalizedString).data;
      }
      finally {
        return value || defaultValue;
      }
    }
  }
};

/// A Google chrome API "chrome", partial impl.
/// Here to ease the future transition to the google chrome addon format in FFx.
var gChrome = {
  jsonLocale: null, // chrome://cookiekeeper_locales/content/en/preferences.json

  i18n : {
    getUILanguage: function() {
      let locales = [];
      function addLocale(locale) {
        locale = locale.trim();
        let hyphen = locale.indexOf("-");
        if (hyphen != -1) {
          let family  = locale.substring(0, hyphen);
          let variant = locale.substring(hyphen+1);
          locale = family.toLowerCase() + "-" + variant.toUpperCase();
        }
        if (locales.indexOf(locale) === -1) {
          locales.push(locale);
        }
      }

      // Most important locale is OS one. But we use it, only if
      // "intl.locale.matchOS" pref is set to `true`.
      // Currently only used for multi-locales mobile builds.
      // http://mxr.mozilla.org/mozilla-central/source/mobile/android/installer/Makefile.in#46
      if (addonSdk.prefs.get(PREF_MATCH_OS_LOCALE, false)) {
        let localeService = Services.locale;
        let osLocale = localeService.getLocaleComponentForUserAgent();
        addLocale(osLocale);
      }

      // In some cases, mainly on Fennec and on Linux version,
      // `general.useragent.locale` is a special 'localized' value, like:
      // "chrome://global/locale/intl.properties"
      let browserUiLocale = addonSdk.prefs.getLocalized(PREF_SELECTED_LOCALE, "") ||
                            addonSdk.prefs.get(PREF_SELECTED_LOCALE, "");

      if (browserUiLocale) {
        addLocale(browserUiLocale);
      }

      // Third priority is the list of locales used for web content
      let contentLocales = addonSdk.prefs.getLocalized(PREF_ACCEPT_LANGUAGES, "") ||
                           addonSdk.prefs.get(PREF_ACCEPT_LANGUAGES, "");
      if (contentLocales) {
        // This list is a string of locales seperated by commas.
        // There is spaces after commas, so strip each item
        for (let locale of contentLocales.split(",")) {
          addLocale(locale.replace(/(^\s+)|(\s+$)/g, ""));
        }
      }

      // Finally, we ensure that en-US is the final fallback if it wasn't added
      addLocale("en");
      addLocale("en-US");

      return locales;
    },

    init: function(aCallback) {
      var langs;

      function updatePlaceholders() {
        // partially from Firefox source code: resource://gre/modules/ExtensionUtils.jsm
        for (let i=0; i<gChrome.jsonLocale.length; i++) {
          let messages = gChrome.jsonLocale[i];
          for (let key of Object.keys(messages)) {
            let msg = messages[key];
            if (typeof(msg) != "object" || typeof(msg.message) != "string") {
              log.e("Invalid locale message data for " + langs[i]);
              continue;
            }
            if (typeof(msg.placeholders) != "object") {
              continue;
            }
            let placeholders = new Map();
            for (let phKey of Object.keys(msg.placeholders)) {
              let placeholder = msg.placeholders[phKey];
              if (typeof(placeholder) != "object" || typeof(placeholder["content"]) != "string") {
                log.e("Invalid placeholder '" + key + "'/'" + phKey + "' for locale " + langs[i]);
                continue;
              }
              placeholders.set(phKey.toLowerCase(), placeholder["content"]);
            }

            let replacer = (match, name) => {
              let replacement = placeholders.get(name.toLowerCase());
              if (typeof(replacement) == "string") {
                return replacement;
              }
              return "";
            };
            let value = msg.message.replace(/\$([A-Za-z0-9@_]+)\$/g, replacer);
            msg.message = value;
          }
        }
        aCallback(null);
      }

      function addJson(counter, langs) {
        if (typeof(aCallback) != "function") {
          throw "aCallback is undefined";
        }
        if (counter < langs.length) {
          let locJson = "chrome://cookiekeeper_locales/content/" + langs[counter] + "/preferences.json";
          let promise = addonSdk.parseJsonURI(locJson);

          promise.then(function(loc) {
            gChrome.jsonLocale.push(loc);
            if (counter+1 < langs.length) {
              addJson(counter+1, langs);
            } else {
              updatePlaceholders();
            }
          }, function(ex) {
            // Ignore if a file is missing, not very important
            if (counter+1 < langs.length) {
              addJson(counter+1, langs);
            } else {
              updatePlaceholders();
            }

            throw ex;
          }).then(undefined, function(error) { // .catch() function is not defined on palemoon
            log.e(error.message);
          });
        } else {
          throw "OutOfBoundException! " + counter.toString(10) + " >= " + langs.length.toString(10);
        }
      }
      if (gChrome.jsonLocale == null) {
        gChrome.jsonLocale = [];
        langs = gChrome.i18n.getUILanguage();
        addJson(0, langs, aCallback);
      }
    },

    getMessage: function(messageName,  substitutions = []) {
      let aSubstitutions = null;
      if (!Array.isArray(substitutions)) {
        if (typeof(substitutions) == "undefined" || substitutions == null) {
          aSubstitutions = [];
        } else {
          aSubstitutions = [substitutions];
        }
      } else {
        aSubstitutions = substitutions;
      }
      for (let i=0; i<gChrome.jsonLocale.length; i++) {
        if (typeof(gChrome.jsonLocale[i][messageName]) != "undefined") {
          let msg = gChrome.jsonLocale[i][messageName].message;
          // Check the substitutions
          let replacer = (matched, index, dollarSigns) => {
            if (index) {
              // This is not quite Chrome-compatible. Chrome consumes any number
              // of digits following the $, but only accepts 9 substitutions. We
              // accept any number of substitutions.
              index = parseInt(index, 10) - 1;
              return index in aSubstitutions ? aSubstitutions[index] : "";
            } else {
              // For any series of contiguous `$`s, the first is dropped, and
              // the rest remain in the output string.
              return dollarSigns;
            }
          };
          return msg.replace(/\$(?:([1-9]\d*)|(\$+))/g, replacer);
        }
      }
      return "";
    }
  }
};

// Preferences observers:
// Update the UI when some browser preferences changes, or when a CK notification pop-up.
var prefsObserver = {
  prefsList: null,

  init: function() {
    if (this.prefsList != null) {
      log.e("prefsObserver already initialized");
      return;
    }
    this.prefsList = new Map();
    Services.obs.addObserver(this, CookieKeeper.OBS_TOPIC, false);
    this.initInputs(document.getElementById("cookiekeeper_body_preferences"));
  },

  uninit: function() {
    if (this.prefsList == null) {
      log.e("prefsObserver.uninit(): prefsList is null");
      return;
    }

    Services.obs.removeObserver(this, CookieKeeper.OBS_TOPIC);

    for (let key of this.prefsList.keys()) {
      let hElem = this.prefsList.get(key);
      let type  = hElem.getAttribute("type");
      if (type == "checkbox") {
        hElem.removeEventListener("change", prefsObserver.onCheckboxChanged, false);
      } else if (type == "number") {
        hElem.removeEventListener("change", prefsObserver.onInputNumberChanged, false);
      } else if (type == "radio") {
        hElem.removeEventListener("change", prefsObserver.onRadioChanged, false);
      }

      Services.prefs.removeObserver(key, this);
    }

    this.prefsList.clear();
    this.prefsList = null;
  },

  // search for <input> items of type "checkbox" and "number", and "form" elements for radio buttons
  // in the DOM tree, to set the current preferences
  initInputs: function(domItem) {
    if (typeof(domItem) != "object" || !domItem.children) {
      log.e("prefsObserver.initInputs: Invalid DOM item ", domItem);
      return;
    }

    let nChildren = domItem.children.length;
    for (let i=0; i<nChildren; i++) {
      let prefName;
      let prefDefault;
      let prefValue;
      let type;
      let child = domItem.children[i];

      if (!child.hasAttribute("preference")) {
        this.initInputs(child);
        continue;
      }

      type = child.hasAttribute("type") ? child.getAttribute("type") : null;
      if (type == null) {
        log.w("Preference element with no type");
        continue;
      }

      // Check the default value
      if (child.hasAttribute("default")) {
        let defaultValue = child.getAttribute("default");
        if (type == "checkbox") {
          prefDefault = defaultValue == "true" ? true : false;
        } else {
          prefDefault = Number.parseInt(defaultValue, 10);
        }
      }

      try {
        prefName  = child.getAttribute("preference");
        prefValue = addonSdk.prefs.get(prefName, prefDefault);
      } catch (ex) {
        log.e(ex);
        continue;
      }

      if (child.nodeName == "FORM" && type == "radio" && typeof(prefValue) == "number") {
        this.selectRadioButton(child, prefValue);
        child.addEventListener("change", prefsObserver.onRadioChanged, false);

      } else if (child.nodeName == "INPUT" && type == "checkbox" && typeof(prefValue) == "boolean") {
        child.checked = prefValue;
        child.addEventListener("change", prefsObserver.onCheckboxChanged,  false);

      } else if (child.nodeName == "INPUT" && type == "number" && typeof(prefValue) == "number") {
        child.value = prefValue.toString(10);
        child.addEventListener("change", prefsObserver.onInputNumberChanged,  false);

      } else {
        log.d("TODO: prefsObserver.initInputs " + child.nodeName + " / " + type + " (" + child.id + ")");
        log.d("pref (" + typeof(prefValue) + ") = ", prefValue);
        continue;
      }

      prefsObserver.prefsList.set(prefName, child);
      Services.prefs.addObserver(prefName, prefsObserver, false);
    }
  },

  observe: function(aSubject, aTopic, aData) {
    if (aTopic == NS_PREFBRANCH_PREFCHANGE_TOPIC_ID) {
      let hElem = prefsObserver.prefsList.get(aData);
      if (typeof(hElem) != "object") {
        log.e("No element for " + aData);
        return;
      }

      let type  = hElem.getAttribute("type");
      let value = addonSdk.prefs.get(aData);
      if (type == "checkbox" && typeof(value) == "boolean") {
        // Avoid infinite event loop:
        hElem.removeEventListener("change", prefsObserver.onCheckboxChanged, false);
        hElem.checked = value;
        hElem.addEventListener("change", prefsObserver.onCheckboxChanged, false);

      } else if (type == "number" && typeof(value) == "number") {
        hElem.removeEventListener("change", prefsObserver.onInputNumberChanged, false);
        hElem.value = value.toString(10);
        hElem.addEventListener("change", prefsObserver.onInputNumberChanged, false);

      } else if (type == "radio" && typeof(value) == "number") {
        hElem.removeEventListener("change", prefsObserver.onRadioChanged, false);
        prefsObserver.selectRadioButton(hElem, value);
        hElem.addEventListener("change", prefsObserver.onRadioChanged, false);

      } else {
        log.e("Invalid preference type: HTML Element type " + type +
          " do not match value type " + typeof(value) + ":", value);
      }

      let evtChanged = new Event(EVENT_CUSTOM_CHANGED, {"bubbles":true, "cancelable":false});
      hElem.dispatchEvent(evtChanged);

    } else if (aTopic == CookieKeeper.OBS_TOPIC) {
      // CookieKeeper specific events
      if (aData == CookieKeeper.OBS_SHUTDOWN) {
        // Add-on shutdown: Close this window.
        window.close();

      } else if (ckPrefs.currentTab == 2 && (aData == CookieKeeper.OBS_ADD || aData == CookieKeeper.OBS_DEL)) {
        // Protected items list changed (add/removed)
        // Update the table content if displayed.
        gCookieKeeperProtItemsHTML.init();
      }
    }
  },
  selectRadioButton: function(hElem, value) {
    let nRadio = hElem.children.length;
    hElem.value = value.toString(10);
    for (let i=0; i<nRadio; i++) {
      let hChild = hElem.children[i];
      if (hChild.nodeName == "INPUT" && hChild.value == hElem.value) {
        hChild.checked = true;
        break;
      }
    }
  },
  onCheckboxChanged: function(evtChange) {
    let node     = evtChange.target;
    let prefName = node.getAttribute("preference");
    let eChanged = new Event(EVENT_CUSTOM_CHANGED, {"bubbles":true, "cancelable":false});
    // Avoid infinite loop with events
    Services.prefs.removeObserver(prefName, prefsObserver);
    // Set the new preference value
    addonSdk.prefs.set(prefName, node.checked);
    // Add again our preference observer
    Services.prefs.addObserver(prefName, prefsObserver, false);
    // Dispach a changed event, needed for other part of the code.
    this.dispatchEvent(eChanged);
  },
  onRadioChanged: function(evtChange) {
    let prefName = this.getAttribute("preference");
    let sValue   = evtChange.target.value;
    let nValue   = Number.parseInt(sValue, 10);
    this.value   = sValue;
    let eChanged = new Event(EVENT_CUSTOM_CHANGED, {"bubbles":true, "cancelable":false});

    Services.prefs.removeObserver(prefName, prefsObserver);
    addonSdk.prefs.set(prefName, nValue);
    Services.prefs.addObserver(prefName, prefsObserver, false);
    this.dispatchEvent(eChanged);
  },
  onInputNumberChanged: function(evtChange) {
    let prefName = this.getAttribute("preference");
    let minValue = this.hasAttribute("min") ? Number.parseInt(this.getAttribute("min"), 10) : Number.MIN_SAFE_INTEGER;
    let maxValue = this.hasAttribute("max") ? Number.parseInt(this.getAttribute("max"), 10) : Number.MAX_SAFE_INTEGER;
    let sValue   = evtChange.target.value;
    let nValue   = Number.parseInt(sValue, 10);

    // Check if the entered value is a valid number
    // <input> element of type number allow to enter any string value...
    if (Number.isInteger(nValue) && nValue >= minValue && nValue <= maxValue) {
      let eChanged = new Event(EVENT_CUSTOM_CHANGED, {"bubbles":true, "cancelable":false});

      Services.prefs.removeObserver(prefName, prefsObserver);
      addonSdk.prefs.set(prefName, nValue);
      Services.prefs.addObserver(prefName, prefsObserver, false);
      this.dispatchEvent(eChanged);
    } else {
      // Error reset to the previous value
      nValue     = addonSdk.prefs.get(prefName);
      sValue     = nValue.toString(10);
      this.value = sValue;
    }
  }
};

/**
 * Translate <span> item with the attribute "data-l10n-id".
 */
function translateUI(rootElem) {
  var defaultSubs = [Services.appinfo.name];

  function translateDomItem(domItem) {
    let childrenElems;
    if (typeof(domItem) != "object") {
      log.e("translateDomItem(domItem): domItem is not an object: ", domItem);
      return;
    }
    if (typeof(domItem.children) == "object" && typeof(domItem.children.length) == "number") {
      childrenElems = domItem.children;
    } else if (typeof(domItem.childNodes) == "object" && typeof(domItem.childNodes.length) == "number") {
      childrenElems = domItem.childNodes;
    } else {
      log.e("translateDomItem(domItem): Invalid domItem ", domItem);
      return;
    }

    let nChildren = childrenElems.length;
    for (let i=0; i<nChildren; i++) {
      let child = childrenElems[i];
      if (typeof(child.hasAttribute) != "function") {
        continue;
      }
      if (child.hasAttribute("data-l10n-id")) {
        let attr = child.getAttribute("data-l10n-id");
        let trText = gChrome.i18n.getMessage(attr, defaultSubs);
        if (trText.length != 0) {
          child.innerHTML = trText;
        } else {
          log.w("Missing translation for " + attr);
        }
        child.removeAttribute("data-l10n-id");
      }
      if (child.hasAttribute("data-l10n-id-tooltiptext")) {
        let attr = child.getAttribute("data-l10n-id-tooltiptext");
        let trText = gChrome.i18n.getMessage(attr, defaultSubs);
        if (trText.length != 0) {
          child.setAttribute("title", trText);
        } else {
          log.w("Missing translation for " + attr);
        }
        child.removeAttribute("data-l10n-id-tooltiptext");
      }
      if (child.hasAttribute("data-l10n-id-label")) {
        let attr = child.getAttribute("data-l10n-id-label");
        let trText = gChrome.i18n.getMessage(attr, defaultSubs);
        if (trText.length != 0) {
          child.setAttribute("label", trText);
        } else {
          log.w("Missing translation for " + attr);
        }
        child.removeAttribute("data-l10n-id-label");
      }
      if (child.hasAttribute("data-l10n-id-href")) {
        let attr = child.getAttribute("data-l10n-id-href");
        let hrefText = gChrome.i18n.getMessage(attr, defaultSubs);
        if (hrefText.length != 0) {
          child.setAttribute("href", hrefText);
        } else {
          log.w("Missing translation for " + attr);
        }
        child.removeAttribute("data-l10n-id-href");
      }
      translateDomItem(child);
    }
  }
  let toElem = typeof(rootElem);
  if (toElem == "undefined" || (toElem == "object" && rootElem == null)) {
    translateDomItem(document);
  } else if (toElem == "string") {
    translateDomItem(document.getElementById(rootElem));
  } else {
    translateDomItem(rootElem);
  }
}

var ckPrefs = {
  currentTab: 0,
  /**
   * @brief Call when the "onload" event is fired.
   */
  init: function() {
    this.currentTab = TAB4_BROWSER_PREFS;
    document.getElementById(TAB_DIV_ID[this.currentTab - 1]).setAttribute("style", "display: block;");

    if (APP_MOZ_CONST[Services.appinfo.ID].NAME != FIREFOX) {
      // Australis UI is only available with Firefox right now
      let cbNoAustralisUI = document.getElementById("cbNoAustralisUI");
      cbNoAustralisUI.parentElement.removeChild(cbNoAustralisUI);
    }

    if (window.location.hash.length > 1) {
      let tabName = window.location.hash.substring(1);
      if (tabName == CK_PREFPANE_PCOOKIES) {
        this.showTab(TAB2_PROTECTED_COOKIES);
      } else if (tabName == CK_PREFPANE_PERMISSIONS) {
        this.showTab(TAB3_PERMISSIONS);
      } else if (tabName == CK_PREFPANE_BROWSER) {
        this.showTab(TAB4_BROWSER_PREFS);
      } else {
        this.showTab(TAB1_CKPREFS);
      }
    } else {
      this.showTab(TAB1_CKPREFS);
    }

    try {
      gChrome.i18n.init(translateUI);
      prefsObserver.init();
      gCookieKeeperBrowserHTMLPrefs.init();
    } catch (ex) {
      log.e(ex.toString(), ex);
    }
  },

  uninit: function() {
    log.d("HTML prefs tab closing");
    gCookieKeeperProtItemsHTML.uninit();
    gCookieKeeperPermissionsHTML.uninit();
    gCookieKeeperBrowserHTMLPrefs.uninit();
    prefsObserver.uninit();
  },

  onKeypress: function(evtKeypress) {
    if (evtKeypress.key == "ArrowDown") {
      if (this.currentTab + 1 > TAB4_BROWSER_PREFS) {
        this.showTab(TAB1_CKPREFS);
      } else {
        this.showTab(this.currentTab + 1);
      }
      return false;
    } else if (evtKeypress.key == "ArrowUp") {
      if (this.currentTab - 1 < TAB1_CKPREFS) {
        this.showTab(TAB4_BROWSER_PREFS);
      } else {
        this.showTab(this.currentTab - 1);
      }
      return false;
    }
    return true;
  },

  /**
   * @brief Change the display tab.
   */
  showTab: function(tabId) {
    if (tabId >= 1 && tabId < 6 && tabId != this.currentTab) {
      let oldTabLink = document.getElementById(TAB_A_PREFIX + this.currentTab.toString(10));
      let newTabLink = document.getElementById(TAB_A_PREFIX + tabId.toString(10));
      let oldTabDiv  = document.getElementById(TAB_DIV_ID[this.currentTab - 1]);
      let newTabDiv  = document.getElementById(TAB_DIV_ID[tabId - 1]);

      this.currentTab = tabId;

      oldTabLink.classList.remove("tab_active");
      newTabLink.classList.add("tab_active");
      oldTabDiv.setAttribute("style", "display: none;");
      newTabDiv.setAttribute("style", "display: block;");

      switch (tabId) {
      case TAB2_PROTECTED_COOKIES:
        gCookieKeeperProtItemsHTML.init();
        break;
      case TAB3_PERMISSIONS:
        gCookieKeeperPermissionsHTML.init();
        break;
      case TAB4_BROWSER_PREFS:
        // Update UI, sometimes may help: Events not dispatched correctly,
        // When the tab is not displayed
        gCookieKeeperBrowserHTMLPrefs.onPermChange();
        gCookieKeeperBrowserHTMLPrefs.onLifetimeChanged();
        break;
      }
    }

    return true;
  }

};
