/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Preferences: Show/remove protected cookies list.
 *
 * ************************************************************************
 */

"use strict";

// XUL Protected cookies management
var gCookieKeeperProtItemsXUL = {
  _strCookie    : "Cookie",
  _strDomStorage: "DOM storage",
  _strLSO       : "Flash cookie",
  _buttonRemove : null,
  _treeView     : null,
  _treeChildren : null,
  _treeCols     : [],
  _pItems       : [],
  _sortColumn   : 0,
  _sortOrderAsc : false,
  _observer     : null,

  init: function() {
    var bundle         = document.getElementById("bundlePreferences");
    this._buttonRemove = document.getElementById("buttonRemove");
    this._treeView     = document.getElementById("treeProtectedItem");
    this._treeChildren = document.getElementById("treeProtItemsChildren");

    this._treeCols.push(document.getElementById("tiType"));
    this._treeCols.push(document.getElementById("tiDomain"));
    this._treeCols.push(document.getElementById("tiName"));
    this._treeCols.push(document.getElementById("tiPath"));
    var tiCookieExists = document.getElementById("tiCookieExists");
    this._treeCols.push(tiCookieExists);

    this._strCookie     = bundle.getString("type_cookie");
    this._strDomStorage = bundle.getString("type_dom_storage");
    this._strLSO        = bundle.getString("type_lso");

    let pItems = CookieKeeper.getProtectedItems();
    this._pItems = pItems.cookies.concat(pItems.storage);
    this._pItems = this._pItems.concat(pItems.lso);

    if (this._pItems.length == 0) {
      this._buttonRemove.disabled = true;
      this._treeView.disabled     = true;
    } else {
      for (let i=0; i<this._pItems.length; i++) {
        this._pItems[i].haveItem = false;
      }

      this._treeView.seltype = "single";
      this.onSort(null, "tiType");
      this._treeView.view.selection.select(0);
    }

    if (tiCookieExists.hidden) {
      // Source: https://developer.mozilla.org/en-US/docs/Web/API/MutationObserver
      // Display if a protected item really have a cookie or a dom storage may be costly.
      // So only fetch the data if the column is displayed.

      // create an observer instance
      this._observer = new MutationObserver(function(mutations) {
        mutations.forEach(function(mutation) {
          if (mutation.type == "attributes" && mutation.attributeName == "hidden") {
            if (!mutation.target.hidden) {
              CookieKeeper.loadCookiesASync(gCookieKeeperProtItemsXUL.onStorageLoaded);
              // One time fetch is enough
              gCookieKeeperProtItemsXUL._observer.disconnect();
              gCookieKeeperProtItemsXUL._observer = null;
            }
          }
        });
      });

      // configuration of the observer:
      var config = { attributes: true, childList: false, characterData: false };

      // pass in the target node, as well as the observer options
      this._observer.observe(tiCookieExists, config);
    } else {
      // The column is displayed by default, so load the datas
      CookieKeeper.loadCookiesASync(this.onStorageLoaded);
    }
  },

  uninit: function() {
    if (this._observer != null) {
      this._observer.disconnect();
      this._observer = null;
    }
    this._buttonRemove = null;
    this._treeView     = null;
    this._treeChildren = null;
    this._treeCols     = null;
    this._pItems       = null;
  },

  onStorageLoaded: function(aItemsArray) {
    // Set if the protected item have a cookie/dom storage stored
    var bundle = document.getElementById("bundleCookieKeeper");
    var yes    = bundle.getString("yes");

    if (aItemsArray && typeof(aItemsArray.length) == "number") {
      let pItems = gCookieKeeperProtItemsXUL._pItems;

      for (let j=0; j<pItems.length; j++) {
        let pc = pItems[j];
        if (pc.haveItem) continue;
        for (let i=0; i<aItemsArray.length; i++) {
          if (pc.equals(aItemsArray[i])) {
            pc.haveItem = true;
            pc.treeCellHaveItem.setAttribute("label", yes);
          }
        }
      }
    }
  },

  _fillList: function() {
    var bundle = document.getElementById("bundleCookieKeeper");
    var yes    = bundle.getString("yes");
    var no     = bundle.getString("no");

    for (var i=0; i<this._pItems.length; i++) {
      let pc   = this._pItems[i];
      let item = document.createElement("treeitem");
      let row  = document.createElement("treerow");
      let cell = null;

      switch (pc.getType()) {
      case COOKIE_TYPE_STANDARD: {
        cell = document.createElement("treecell");
        cell.setAttribute("label", this._strCookie);
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label", pc.getHost());
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  pc.getName());
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  pc.getPath());
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label", pc.haveItem ? yes : no);
        this._pItems[i].treeCellHaveItem = cell;
        row.appendChild(cell);
        break;
      }

      case COOKIE_TYPE_STORAGE: {
        cell = document.createElement("treecell");
        cell.setAttribute("label", this._strDomStorage);
        row.appendChild(cell);

        let scope = pc.scope.split(":");
        let host  = gCommon.reverseStr(scope[0]);
        cell = document.createElement("treecell");
        cell.setAttribute("label", host + " (" + scope[1] + ":" + scope[2] + ")");
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  pc.getName());
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  "");
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label", pc.haveItem ? yes : no);
        this._pItems[i].treeCellHaveItem = cell;

        row.appendChild(cell);
        break;
      }

      case COOKIE_TYPE_LSO: {
        cell = document.createElement("treecell");
        cell.setAttribute("label", this._strLSO);
        row.appendChild(cell);

        let host = pc.getHost();
        if (gCommon.isSpecialHost(host))
          host = bundle.getString(host.substring(1, host.length-1));
        cell = document.createElement("treecell");
        cell.setAttribute("label", host);
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  pc.getName());
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label",  pc.getPath());
        row.appendChild(cell);

        cell = document.createElement("treecell");
        cell.setAttribute("label", pc.haveItem ? yes : no);
        this._pItems[i].treeCellHaveItem = cell;
        row.appendChild(cell);
        break;
      }

      default:
        log.e("Unknown cookie type " + pc.getType(), pc);
        break;
      }

      if (cell != null) {
        item.appendChild(row);
        item.setAttribute("accessibleType", "101F")
        this._treeChildren.appendChild(item);

        // To easily remove this item when a protected cookie is deleted
        pc.view = item;
      }
    }
  },

  onRemove: function() {
    var seln = this._treeView.view.selection;
    var selIndexes = [];

    if (seln.count == 1) {
      selIndexes.push(seln.currentIndex);
    } else if (seln.count > 1) {
      let rangeCount = seln.getRangeCount();
      for (let i = 0; i < rangeCount; i++) {
        var min = {}; var max = {};
        seln.getRangeAt(i, min, max);
        for (let j = min.value; j <= max.value; j++) {
          selIndexes.push(j);
        }
      }
    }

    for (let i=selIndexes.length-1; i>=0; i--) {
      let index = selIndexes[i];
      var pCookie = this._pItems[index];

      this._pItems.splice(index, 1);

      // Remove protected cookie:
      CookieKeeper.removeProtectedItem(pCookie);

      // Update ui:
      this._treeChildren.removeChild(pCookie.view);
      var rowCount = this._pItems.length;
      if (rowCount == 0) {
        this._buttonRemove.disabled = true;
        this._treeView.disabled     = true;
      } else {
        var newIndex = (index < rowCount ? index : rowCount - 1);
        seln.select(newIndex);
      }
    }

    this._treeView.focus();
  },

  // Sort the _pItems list
  _sort: function(a, b) {
    var c = 0;
    var loop = 0|0; // prevent infinte loop... we never know
    var strTypes = [
      "",
      gCookieKeeperProtItemsXUL._strCookie,
      gCookieKeeperProtItemsXUL._strDomStorage,
      gCookieKeeperProtItemsXUL._strLSO
    ];

    let sortColumn = gCookieKeeperProtItemsXUL._sortColumn;

    while (c == 0 && loop < 2) {
      loop++;
      switch (sortColumn) {
        case 0: // Type (Cookie/Dom storage)
          c = strTypes[a.getType()].localeCompare(strTypes[b.getType()]);
          if (c != 0) { break; }
          // fall through
        case 1: { // Host/Domain
          c = gCommon.cmpHostName(a.getHost(), b.getHost());
          if (c != 0) { break; }
        } // fall through
        case 2: { // Name/Key
          c = a.getName().toLowerCase().localeCompare(b.getName().toLowerCase());
          if (c != 0) { break; }
        } // fall through
        case 3: { // Path
          c = a.getPath().toLowerCase().localeCompare(b.getPath().toLowerCase());
          if (c != 0) { break; }
        } // fall through
        case 4: // Item present
          c = a.haveItem - b.haveItem;
          // fall through
        default:
          if (c == 0 && sortColumn>0) { sortColumn = 0; }
      }
    }

    if (gCookieKeeperProtItemsXUL._sortOrderAsc) {
      return c;
    } else {
      return -c;
    }
  },

  getColumnId: function(aColumnName) {
    for (let i=0; i<this._treeCols.length; i++) {
      if (aColumnName == this._treeCols[i].getAttribute("id")) {
        return i;
      }
    }

    log.e("Unknown column id " + aColumnName);
    return -1;
  },

  onSort: function(aEvent, columnName) {
    var columnId = this.getColumnId(columnName);
    if ((aEvent != null && aEvent.button != 0) || columnId < 0 || columnId >= this._treeCols.length) {
      // param check (not usefull, but to be sure)
      return;
    }

    var colHeader = this._treeCols[columnId];
    var column    = null;

    if (this._sortColumn == columnId) {
      this._sortOrderAsc = !this._sortOrderAsc;
    } else {
      var oldColHeader = this._treeCols[this._sortColumn];

      this._sortOrderAsc = true;
      this._sortColumn   = columnId;

      oldColHeader.removeAttribute("sortDirection");
    }
    colHeader.setAttribute("sortDirection", this._sortOrderAsc ? "ascending" : "descending");

    try {
      this._pItems.sort(this._sort);
    } catch (ex) {
      log.e(ex);
    }

    // Not pretty, but easy to do
    while (this._treeChildren.firstChild) {
      this._treeChildren.removeChild(this._treeChildren.firstChild);
    }

    this._fillList();
  }
};


// HTML Protected cookies management

var gCookieKeeperProtItemsHTML = {
  _oDynTable   : null,

  init: function() {
    try {
      this.uninit();

      let hTable = document.getElementById("pcookies_table");
      this._oDynTable = new DynTable(hTable, "pcookies");

      let hTBody  = document.getElementById("pcookies_tbody");
      let oPItems = CookieKeeper.getProtectedItems();
      let aPItems = oPItems.cookies.concat(oPItems.storage);
      aPItems     = aPItems.concat(oPItems.lso);

      for (let i=0; i<aPItems.length; i++) {
        let hTrElem = this.createLineElem(aPItems[i], i);
        if (hTrElem) {
          hTBody.appendChild(hTrElem);
        }
      }

      // Do the translation
      translateUI(hTBody);
      // Allow sorting
      this._oDynTable.enableSortColumn(this._compare);
      // Do initial sorting
      this._oDynTable.doSort(null);
      // Initial column resize
      this._oDynTable.resizeBodyColumns();
      // Allow selection
      this._oDynTable.setSelectListener(this.onSelectionChange);

      hTable.setAttribute("tabindex", "2");

      // Update the last column: indicator of the cookie existance
      window.setTimeout(function() {
        // Do this operation async, since it may take more time to complete
        CookieKeeper.loadCookiesASync(gCookieKeeperProtItemsHTML.onStorageLoaded);
      }, 0);

    } catch (ex) {
      log.e(ex);
    }
  },

  uninit: function() {
    let hTable  = document.getElementById("pcookies_table");
    let hTBody  = document.getElementById("pcookies_tbody");
    let hButton = document.getElementById("pcookies_button_delete");

    hTable.removeAttribute("tabindex");

    // Disable the "Remove" button
    hButton.setAttribute("disabled", "disabled");

    // Remove protected HTML items
    while (hTBody.firstElementChild) {
      let hTrElem = hTBody.firstElementChild;
      delete hTrElem.oPItem;
      hTBody.removeChild(hTrElem);
    }

    // Reset/Clean the _oDynTable object
    if (this._oDynTable != null) {
      this._oDynTable.destructor();
      this._oDynTable = null;
    }
  },

  createLineElem: function(oPItem, idx) {
    let typeName;
    let typeL10nId;
    switch (oPItem.getType()) {
    case COOKIE_TYPE_STANDARD:
      typeName = "Cookie";
      typeL10nId = "prot.type.cookie";
      break;
    case COOKIE_TYPE_STORAGE:
      typeName = "DOM Storage";
      typeL10nId = "prot.type.dom";
      break;
    case COOKIE_TYPE_LSO:
      typeName = "LSO Flash";
      typeL10nId = "prot.type.lso";
      break;
    default:
      log.e("Unknown protected item type :" + oPItem.getType(), oPItem);
      return null;
    }

    let idxStr = idx.toString(10);

    let hTrElem = document.createElement("tr");
    hTrElem.setAttribute("id", "pcookies_body_tr_" + idxStr);
    hTrElem.classList.add("table_scroll_contentrow");
    hTrElem.oPItem = oPItem;

    // Type of protected element
    let hTdElem = document.createElement("td");
    hTdElem.setAttribute("data-l10n-id", typeL10nId);
    hTdElem.setAttribute("data-l10n-id-tooltiptext", typeL10nId);
    hTdElem.classList.add("table_scroll_contentcell");
    hTdElem.innerHTML = typeName;
    hTrElem.appendChild(hTdElem);

    // Domain (site)
    hTdElem = document.createElement("td");
    hTdElem.setAttribute("title", oPItem.getHost());
    hTdElem.classList.add("table_scroll_contentcell");
    hTdElem.innerHTML = oPItem.getHost();
    hTrElem.appendChild(hTdElem);

    // Name
    hTdElem = document.createElement("td");
    hTdElem.setAttribute("title", oPItem.getName());
    hTdElem.classList.add("table_scroll_contentcell");
    hTdElem.innerHTML = oPItem.getName();
    hTrElem.appendChild(hTdElem);

    // Path
    hTdElem = document.createElement("td");
    if (oPItem.getType() == COOKIE_TYPE_STORAGE) {
      hTdElem.innerHTML = oPItem.getProtocol() + ":" + oPItem.getPort(); // fake space, to have the cell at the normal height
      hTdElem.setAttribute("title", hTdElem.innerHTML);
    } else {
      hTdElem.innerHTML = oPItem.getPath();
      hTdElem.setAttribute("title", oPItem.getPath());
    }
    hTdElem.classList.add("table_scroll_contentcell");
    hTrElem.appendChild(hTdElem);

    // Exists
    hTdElem = document.createElement("td");
    hTdElem.setAttribute("data-l10n-id", "no");
    hTdElem.classList.add("table_scroll_contentcell");
    hTdElem.innerHTML = "no";
    hTrElem.appendChild(hTdElem);

    return hTrElem;
  },

  onStorageLoaded: function(aItemsArray) {
    let hTBody = document.getElementById("pcookies_tbody");
    let nChildren = hTBody.children.length;
    let nItems = aItemsArray.length;

    for (let i=0; i<nItems; i++) {

      let oItem = aItemsArray[i];
      for (let j=0; j<nChildren; j++) {

        let hTrElem = hTBody.children[j];
        if (oItem.equals(hTrElem.oPItem)) {
          let hTdElem = hTrElem.children[4];
          hTdElem.innerHTML = "yes";
          hTdElem.setAttribute("data-l10n-id", "yes");
          translateUI(hTdElem.parentNode);
        }
      }
    }
  },

  // Compare 2 rows (<tr> elements) of the <tbody>.
  // Use to sort the table
  _compare: function(a, b) {
    let nSortColumn   = gCookieKeeperProtItemsHTML._oDynTable.aSortColumns.length;
    let sortDirection = gCookieKeeperProtItemsHTML._oDynTable.aSortColumns[0] > 0;

    for (let i=0; i<nSortColumn; i++) {
      let nCol   = Math.abs(gCookieKeeperProtItemsHTML._oDynTable.aSortColumns[i]) - 1;
      let aHTML  = a.children[nCol].innerHTML.trim();
      let bHTML  = b.children[nCol].innerHTML.trim();

      let cmpVal = 0;
      if (nCol == 1) {
        // Domain comparison:
        cmpVal = gCommon.cmpHostName(aHTML, bHTML);
      } else {
        // All other columns
        cmpVal = aHTML.toLowerCase().localeCompare(bHTML.toLowerCase());
      }
      if (cmpVal != 0) {
        return sortDirection ? cmpVal : -cmpVal;
      }
    }

    return 0;
  },

  onSelectionChange: function(aAddedItems, aRemovedItems, aCurrentItems) {
    // Enable the delete button
    let btn = document.getElementById("pcookies_button_delete");
    if (typeof(aCurrentItems) != "object" || aCurrentItems.length == 0) {
      btn.setAttribute("disabled", "disabled");
    } else {
      btn.removeAttribute("disabled");
    }
  },

  onDelete: function() {
    let hTBody = document.getElementById("pcookies_tbody");

    let aSelected = this._oDynTable.aSelected;
    if (aSelected.length == 0) {
      log.e("gCookieKeeperProtItemsHTML.onDelete: no item selected");
      return;
    }

    let nChildren = hTBody.children.length;
    let aTrElems  = new Array(aSelected.length);
    for (let i=0; i<aSelected.length; i++) {

      let selIdx = aSelected[i];
      if (selIdx < 0 || selIdx >= hTBody.children.length) {
        log.e("gCookieKeeperProtItemsHTML.onDelete: Invalid selection index: " + selIdx.toString(10) + " at " + i.toString(10));
        continue;
      }

      let hTrElem = hTBody.children[selIdx];
      let oPItem = hTrElem.oPItem;
      if (typeof(oPItem) != "object") {
        log.e("gCookieKeeperProtItemsHTML.onDelete: Invalid oPItem for <tr> element " + selIdx.toString(10));
      }

      if (CookieKeeper.removeProtectedItem(oPItem)) {
        aTrElems[i] = hTrElem;
      }
    }

    this._oDynTable.clearSelection();

    // Remove <tr> elements of the <tbody> children list
    for (let i=0; i<aTrElems.length; i++) {
      let hTrElem = aTrElems[i];
      if (typeof(hTrElem) != "undefined") {
        hTBody.removeChild(aTrElems[i]);
      }
    }

    // We now must update the <tr> class list for
    // "table_scroll_contentrow_even" and "table_scroll_contentrow_odd"
    nChildren = hTBody.children.length;
    for (let i=0; i<nChildren; i++) {
      let hTrElem = hTBody.children[i];
      hTrElem.classList.remove("table_scroll_contentrow_even");
      hTrElem.classList.remove("table_scroll_contentrow_odd");
      hTrElem.classList.add((i % 2 == 0) ? "table_scroll_contentrow_even" : "table_scroll_contentrow_odd");
    }
  }
};
