/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Preferences managment
 *
 * ************************************************************************
 */

"use strict";

var ckPrefs = {

  // Initializes default preferences
  // https://github.com/mozilla/addon-sdk/blob/85511f1c467a9fb5d2b8d2b6b4360e7598fce638/python-lib/cuddlefish/app-extension/bootstrap.js
  setDefaultPrefs: function() {
    const prefsDefault  = Cc["@mozilla.org/preferences-service;1"].getService(Ci.nsIPrefService).QueryInterface(Ci.nsIPrefBranch2);

    let branchD = prefsDefault.getDefaultBranch("");
    let branchS = Services.prefs.getDefaultBranch("");
    let prefLoaderScope = {
      pref: function(key, val) {
        // set the value only if it does not exists
        var exists = branchS.prefHasUserValue(key);
        switch (typeof(val)) {
          case "boolean":
            branchD.setBoolPref(key, val);
            if (!exists) branchS.setBoolPref(key, val);
            break;
          case "number":
            branchD.setIntPref(key, val);
            if (!exists) branchS.setIntPref(key, val);
            break;
          case "string":
            branchD.setCharPref(key, val);
            if (!exists) branchS.setCharPref(key, val);
            break;
        }
      }
    };

    let prefjs = addonData.resourceURI.spec + "defaults/preferences.js";
    // setup default prefs
    Services.scriptloader.loadSubScript(prefjs, prefLoaderScope);
  },

  _mergeProtCookieArrays: function(array1, array2) {
    if (array1.length == 0) {
      return array2;
    }
    if (array2.length == 0) {
      return array1;
    }

    let ret = [];
    ret = ret.concat(array1);

    // Add unique
    for (let i=0; i<array2.length; i++) {
      let pc2 = array2[i];
      let haveIt = false;
      for (let j=0; j<ret.length && haveIt==false; j++) {
        let pc1 = ret[j];
        haveIt = pc1.equals(pc2);
      }

      if (!haveIt) {
        ret.push(pc2);
      }
    }

    return ret;
  },

  // Migrate CookieCuller preferences
  migrateCCPrefs: function() {
    var psvc = Services.prefs;

    // Migrate & delete old cookie culler (v1.4) preferences
    const PREF_BRANCH_COOKIE_CULLER = "extensions.cookieculler.";
    const PREF_DEL_AT_STARTUP       = "extensions.cookieculler_options.delUnprotAtStartup";

    // Get protected cookies
    var bsvc   = Cc["@mozilla.org/preferences-service;1"].getService(Components.interfaces.nsIPrefService);
    var branch = bsvc.getBranch(PREF_BRANCH_COOKIE_CULLER);
    var delAtStartup = false;

    if (psvc.prefHasUserValue(PREF_DEL_AT_STARTUP)) {
      delAtStartup = psvc.getBoolPref(PREF_DEL_AT_STARTUP);
    }

    if (branch.prefHasUserValue("count") || delAtStartup) {
      // Init the locale bundle, to get a localized message
      if (l10n == null) {
        l10n = new Localization(addonData.resourceURI.spec, "cookiekeeper.properties");
      }

      var doMigr = Services.prompt.confirm(null, CK_MODULE_NAME, l10n.tr("install.prefs.cookieculler"));

      if (doMigr && branch.prefHasUserValue("count")) {
        let count       = branch.getIntPref("count");
        let protCookies = [];
        let newProtList = CookieKeeperPrivate._protectedList;

        // Get CookieCuller protected cookies
        for (let i=0; i<count; i++) {
          const pname = "pcookie" + i.toString(10);
          if (branch.prefHasUserValue(pname)) {
            let pstr = branch.getCharPref(pname);
            let cData = pstr.split(";");
            if (cData.length == 3) {
              let pCookie = gCookieStorage.makeProtectedObject(cData[0], cData[1], cData[2]);
              protCookies.push(pCookie);
            }
          }
        }

        // Build the protection list
        if (protCookies.length > 0) {
          newProtList.cookies    = this._mergeProtCookieArrays(newProtList.cookies, protCookies);
          newProtList.lastUpdate = Date.now();
        }
      } // pcookies migration

      if (doMigr && delAtStartup) {
        psvc.setBoolPref(PREF_CLEAR_COOKIES_AT_EXIT, delAtStartup);
        psvc.setBoolPref(PREF_DEL_AT_STARTUP, false);
      } // delAtStartup migration
    } // have CookieCuller preferences
  }, // migrateCCPrefs

  /// Migration of "localstore.rdf" and "xulstore.json" datas
  migrateStores: function() {
    var xulStore;
    var localStore;

    var oldVersion = typeof(addonData.oldVersion) == "string" ? addonData.oldVersion : "1.5.0";

    if (cmpVersion(oldVersion, "1.8") <= 0) {
      // Theses old preferences should have been removed in 1.8:
      // "preferences.xul" and "cookie.editor.xul"
      let resources = [
        "chrome://cookiekeeper/content/preferences.xul",
        "chrome://cookiekeeper/content/cookie.editor.xul"
      ];

      localStore = cleanLocalStore(resources, localStore);
      xulStore   = cleanXulStore(resources, xulStore);
    }

    if (cmpVersion(oldVersion, "1.9") < 0) {
      // Clean stores of "prefs.main.xul" (clean the size mostly)
      let resources = [
        "chrome://cookiekeeper/content/prefs.main.xul"
      ];

      localStore = cleanLocalStore(resources, localStore);
      xulStore   = cleanXulStore(resources, xulStore);
    }
  },

  upgradePreferences: function(reason) {
    var psvc             = Services.prefs;
    var oldVer           = CK_PREF_VERSION;
    var havePrefVersion  = false;
    var pList            = null;

    CookieKeeperPrivate.init();
    CookieKeeperPrivate.loadProtectedList(true);

    if (psvc.prefHasUserValue(PREF_VERSION)) {
      havePrefVersion = true;
      oldVer = psvc.getIntPref(PREF_VERSION);
      if (oldVer == 150) {
        oldVer = 1;
      }
    }
    if (psvc.prefHasUserValue(PREF_PROT_COOKIES)) {
      pList = psvc.getCharPref(PREF_PROT_COOKIES);
    }
    if (pList != null && !havePrefVersion) {
      oldVer = 1; // bug lp:1305252
    }
    // Update version
    psvc.setIntPref(PREF_VERSION, CK_PREF_VERSION);

    log.d("upgradePreferences : " + addonData.oldVersion + ":" + oldVer + " -> " + addonData.version + ":" + CK_PREF_VERSION);

    switch (oldVer) {
    case 1:
    {
      // Add the PROT_SEP_COOKIE char "|" at the end of the cookies list.
      // See bug lp:1282328 https://bugs.launchpad.net/cookiekeeper/+bug/1282328
      if (pList != null) {
        if (pList.length > 1 && pList.charAt(pList.length - 1) != PROT_SEP_COOKIE) {
          pList = pList + PROT_SEP_COOKIE;
          psvc.setCharPref(PREF_PROT_COOKIES, pList);
        }
      }
    } // fall-through
    case 2:
      // Migrate protected list from preferences to JSON file
      try {
        if (pList != null) {
          let oldProtList = [];
          let newProtList = CookieKeeperPrivate._protectedList;

          let cList = pList.split(PROT_SEP_COOKIE);
          for (let i=0; i<cList.length; i++) {
            let cStr = cList[i]; // cookie string
            if (cStr.length > 0) {
              let cData = cStr.split(PROT_SEP_DATA); // cookie datas (host, name, path)
              if (cData.length == 3) {
                let pCookie = gCookieStorage.makeProtectedObject(cData[0], cData[1], cData[2]);
                oldProtList.push(pCookie);
              }
            }
          }

          if (oldProtList.length > 0) {
            newProtList.cookies    = this._mergeProtCookieArrays(newProtList.cookies, oldProtList);
            newProtList.lastUpdate = Date.now();
          }

          // Remove the old list
          psvc.clearUserPref(PREF_PROT_COOKIES);
        }
      } catch (ex) {
        log.e("Convert old pref data error:", ex);
      }
    // fall-through
    case 3: // 1.8.x
      // fall-through
    case 4: // 1.9.x
      // fall-through
    default:
      break;
    } // switch

    if (reason == ADDON_INSTALL) {
      // Only ask to migrate the CookieCuller preferences
      // when the user first install CK.
      this.migrateCCPrefs();
    }

    CookieKeeperPrivate.saveProtectedList(true);
    CookieKeeperPrivate.uninit();
  }, // upgradePreferences()

  downgradePreferences: function() {
    function toStr(pcookie) {
      return pcookie.host + PROT_SEP_DATA + pcookie.name + PROT_SEP_DATA + pcookie.path;
    }

    if (addonData == null) {
      log.e("addonData is not set");
      return;
    }

    let ov   = addonData.version;
    let nv   = addonData.newVersion;
    var psvc = Services.prefs;

    if (typeof(ov) != "string" || typeof(nv) != "string") {
      log.e("Error in version numbers oldVersion=" + ov + " newVersion=" + nv);
      return;
    }

    if (cmpVersion(ov, "1.9") >= 0 && cmpVersion(nv, "1.8") <= 0) {
      log.i("Downgrade preferences from 1.9.x to 1.8.0 (" + ov + " -> " + nv + ")");
      psvc.clearUserPref(PREF_CLEAR_LSO_ON_DEMAND);
      psvc.clearUserPref(PREF_CLEAR_HSTS_ON_DEMAND);
      psvc.clearUserPref(PREF_CLEAR_LSO_AT_EXIT);
      psvc.clearUserPref(PREF_CLEAR_HSTS_AT_EXIT);
      psvc.clearUserPref(PREF_LSO_KEEP_SETTINGS);
      psvc.clearUserPref(PREF_WARN_PROT_SESSION);
      psvc.clearUserPref(PREF_NO_AUSTRALIS_UI);
      psvc.setIntPref(PREF_VERSION, 3);
    }

    if (cmpVersion(ov, "1.8") >= 0 && cmpVersion(nv, "1.7") <= 0) {
      log.i("Downgrade preferences from 1.8.x to 1.7.0 (" + ov + " -> " + nv + ")");
      let jsonFile = FileUtils.getFile("ProfD", [PROT_JSON_FILE], false);
      if (jsonFile.exists()) {
        // move pref cookies from JSON file to preferences
        let pcListStr = PROT_SEP_COOKIE;
        CookieKeeperPrivate.init();
        CookieKeeperPrivate.loadProtectedList(true);

        let protList = CookieKeeperPrivate._protectedList.cookies;
        // Build the protection list
        for (let i=0; i<protList.length; i++) {
          let pcookie = protList[i];
          if (pcookie != null) {
            let pstr = toStr(protList[i]) + PROT_SEP_COOKIE;
            pcListStr = pcListStr + pstr;
          }
        }

        psvc.setCharPref(PREF_PROT_COOKIES, pcListStr);
        // Remove the file
        jsonFile.remove(false);
        CookieKeeperPrivate.uninit();
      }
      psvc.setIntPref(PREF_VERSION, 2);
    }
  } // downgradePreferences()

}; // ckPrefs
