/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2016 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * UI managment (add toolbar button, an menu items)
 *
 * Partially based on:
 *  - https://github.com/jvillalobos/Australis-View-Demo/blob/master/src/bootstrap.js
 *  - https://blog.mozilla.org/addons/2014/03/06/australis-for-add-on-developers-2/
 * ************************************************************************
 */

"use strict";

const CK_TOOLBAR_CSS  = "chrome://cookiekeeper_app/skin/toolbar.css";

function ckUnloader() {
  // Check if the last browser window is closed
  // Note: The Services.wm.getEnumerator() function can't be called within the
  // event onCloseWindow(). I got an infinite loop.
  // So use the "unload" event instead.

  this.removeEventListener("unload", ckUnloader, false);
  let enumerator = Services.wm.getEnumerator(APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE);

  if (!enumerator.hasMoreElements()) {
    // If so, close all cookiekeeper windows too.
    CookieKeeper.closeCookieWindow();
  }
}

function isPrivateBrowsingMode(aWindow) {
  try {
    var enablePrivBrowse = false;
    if (Services.prefs.getPrefType(PREF_ENABLE_PRIV_BROWSE) == Services.prefs.PREF_BOOL) {
      enablePrivBrowse = Services.prefs.getBoolPref(PREF_ENABLE_PRIV_BROWSE);
    }
    if (enablePrivBrowse) {
      return false;
    }

    // Firefox 20+
    Cu.import("resource://gre/modules/PrivateBrowsingUtils.jsm");
    if (typeof(PrivateBrowsingUtils.isContentWindowPrivate) == "function") {
      // for e10s
      return PrivateBrowsingUtils.isContentWindowPrivate(aWindow);
    } else {
      // deprecated for e10s
      return PrivateBrowsingUtils.isWindowPrivate(aWindow);
    }
  } catch (ex) {
    return false;
  }
}

var ckPermMenuManagement = {
  oPermManager : null,
  trAllow      : "Allow",
  trSession    : "Session",
  trBlock      : "Block",
  trDefault    : "%1 (Default)",
  trAsk        : "%1 (Ask)",
  trLimit      : "%1 (Limit)",

  init: function() {
    this.trAllow   = l10n.tr("menu.item.permissions.allow");
    this.trSession = l10n.tr("menu.item.permissions.session");
    this.trBlock   = l10n.tr("menu.item.permissions.block");
    this.trDefault = l10n.tr("menu.label.permissions.default");
    this.trAsk     = l10n.tr("menu.label.permissions.ask");
    this.trLimit   = l10n.tr("menu.label.permissions.limit");
  },

  uninit: function() {
    if (this.oPermManager != null) {
      this.oPermManager.destructor();
      this.oPermManager = null;
    }
  },

  onSetSitePerm: function(aEvent) {
    var that   = ckPermMenuManagement;
    var target = aEvent.target;

    if (target == null || !(target.ckURI instanceof Ci.nsIURI)) {
      log.e("Unknown/wrong event target ", aEvent);
      return;
    }

    var uri         = target.ckURI;
    var currentPerm = that.oPermManager.getCurrentPerm(uri);

    if (target.permAction == currentPerm) {
      that.oPermManager.remove(uri);
    } else {
      that.oPermManager.add(uri, target.permAction);
    }

    aEvent.stopPropagation();
    return false;
  },

  addMenuPermissions: function(aEvent, menupopup, place) {

    if (aEvent.target.getAttribute("id") != "cookiekeeper-menupopup-" + place) {
      // For some reason, the event is triggered for the subpopup too
      // don't need to re-add the items
      aEvent.stopPropagation();
      aEvent.stopImmediatePropagation();
      return true;
    }

    var doc        = aEvent.target.ownerDocument;
    var domWindow  = aEvent.target.ownerGlobal;
    {
      let selectedBrowser = doc.defaultView.gBrowser.selectedBrowser;
      var currentURI = selectedBrowser.currentURI;
      var iconURL    = selectedBrowser.mIconURL;
    }

    if (currentURI.schemeIs("http") || currentURI.schemeIs("https")) {
      let permManager    = ckPermMenuManagement.oPermManager;
      let ckMenuShowMain = l10n.tr("menu.showcookies.label");
      let domain         = currentURI.host;
      let scheme         = currentURI.scheme;
      let minDomain      = "";
      let domains        = [];

      if (ckPermMenuManagement.oPermManager == null) {
        ckPermMenuManagement.oPermManager = new CKPermissionsManager();
        permManager = ckPermMenuManagement.oPermManager;
      }

      try {
        minDomain = Services.eTLD.getBaseDomainFromHost(domain);
        domains   = domain.split(".");
      } catch (e) {
        // probably an IP address
        domains = [domain];
        minDomain = domain;
      }

      let currentBehavior = 0;
      let lifetimePolicy  = 0;

      let appPrefInfos = APP_MOZ_CONST[Services.appinfo.ID];
      if (Services.prefs.getPrefType(appPrefInfos.COOKIES_BEHAVIOR) == Services.prefs.PREF_INT) {
        currentBehavior = Services.prefs.getIntPref(appPrefInfos.COOKIES_BEHAVIOR);
      }
      if (Services.prefs.getPrefType(appPrefInfos.COOKIES_LIFETIME) == Services.prefs.PREF_INT) {
        lifetimePolicy = Services.prefs.getIntPref(appPrefInfos.COOKIES_LIFETIME);
      }

      let sep = doc.createElementNS(XUL_NS, "menuseparator");
      sep.setAttribute("id", "cookiekeeper-menuitem-" + place + "-perms-sep");
      sep.setAttribute("accessibleType", XULMenuSeparator);
      menupopup.appendChild(sep);

      for (let i=0; i<domains.length; i++) {
        let partialDomain = domains[i];
        for (let j=i+1; j<domains.length; j++) {
          partialDomain = partialDomain + "." + domains[j];
        }
        if (minDomain.length > partialDomain.length) {
          break;
        }

        let uri = Services.io.newURI(scheme + "://" + partialDomain, null, null);
        let defaultPerm = permManager.getDefaultPerm(uri);
        let exactPerm   = permManager.getCurrentPerm(uri);
        let currentPerm = defaultPerm == MOZ_COOKIE_PERM_DEFAULT ? exactPerm : defaultPerm;
        let isDefault   = exactPerm   == MOZ_COOKIE_PERM_DEFAULT;

        let miShowCookies = doc.createElementNS(XUL_NS, "menuitem");
        miShowCookies.setAttribute("id", "cookiekeeper-menuitem-" + place + "-perms-" + partialDomain + "-showcookies");
        miShowCookies.setAttribute("label", ckMenuShowMain);
        miShowCookies.setAttribute("accessibleType", XULMenuitem);
        miShowCookies.addEventListener("command", function(event) {CookieKeeper.showMainWindow(domWindow, partialDomain); return true;}, false);

        let miSep = doc.createElementNS(XUL_NS, "menuseparator");
        miSep.setAttribute("id", "cookiekeeper-menuitem-" + place + "-perms-" + partialDomain + "-sep");
        miSep.setAttribute("accessibleType", XULMenuSeparator);

        let miAllow = doc.createElementNS(XUL_NS, "menuitem");
        miAllow.setAttribute("id", "cookiekeeper-menuitem-" + place + "-perms-" + partialDomain + "-allow");
        miAllow.setAttribute("accessibleType", XULMenuitem);
        miAllow.setAttribute("type", "checkbox");
        miAllow.setAttribute("checked", "false");
        miAllow.ckURI      = uri;
        miAllow.permAction = MOZ_COOKIE_PERM_ALLOW;
        miAllow.addEventListener("command", ckPermMenuManagement.onSetSitePerm, false);

        let miSession = doc.createElementNS(XUL_NS, "menuitem");
        miSession.setAttribute("id", "cookiekeeper-menuitem-" + place + "-perms-" + partialDomain + "-session");
        miSession.setAttribute("accessibleType", XULMenuitem);
        miSession.setAttribute("type", "checkbox");
        miSession.setAttribute("checked", "false");
        miSession.ckURI      = uri;
        miSession.permAction = MOZ_COOKIE_PERM_SESSION;
        miSession.addEventListener("command", ckPermMenuManagement.onSetSitePerm, false);

        let miBlock = doc.createElementNS(XUL_NS, "menuitem");
        miBlock.setAttribute("id", "cookiekeeper-menuitem-" + place + "-perms-" + partialDomain + "-block");
        miBlock.setAttribute("accessibleType", XULMenuitem);
        miBlock.setAttribute("type", "checkbox");
        miBlock.setAttribute("checked", "false");
        miSession.ckURI    = uri;
        miBlock.permAction = MOZ_COOKIE_PERM_DENY;
        miBlock.addEventListener("command", ckPermMenuManagement.onSetSitePerm, false);

        miAllow.setAttribute("label", this.trAllow);
        miSession.setAttribute("label", this.trSession);
        miBlock.setAttribute("label", this.trBlock);

        switch (currentPerm) {
        case MOZ_COOKIE_PERM_ALLOW:
          miAllow.setAttribute("checked", "true");
          if (isDefault) {
            let labelAllow = this.trDefault.replace("%1", this.trAllow);
            miAllow.setAttribute("label", labelAllow);
          }
          break;
        case MOZ_COOKIE_PERM_SESSION:
          miSession.setAttribute("checked", "true");
          if (isDefault) {
            let labelSession = this.trDefault.replace("%1", this.trSession);
            miSession.setAttribute("label", labelSession);
          }
          break;
        case MOZ_COOKIE_PERM_DENY:
          miBlock.setAttribute("checked", "true");
          if (isDefault) {
            let labelBlock = this.trDefault.replace("%1", this.trBlock);
            miBlock.setAttribute("label", labelBlock);
          }
          break;
        case MOZ_COOKIE_PERM_DEFAULT:
          if (currentBehavior == MOZ_BEHAVIOR_DISABLED) {
            let labelBlock = this.trDefault.replace("%1", this.trBlock);
            miBlock.setAttribute("label", labelBlock);
            miBlock.setAttribute("checked", "true");
          } else if (lifetimePolicy == MOZ_ACCEPT_SESSION) {
            let labelSession = this.trDefault.replace("%1", this.trSession);
            miSession.setAttribute("label", labelSession);
            miSession.setAttribute("checked", "true");
          } else if (lifetimePolicy == MOZ_ASK_BEFORE_ACCEPT) {
            let labelAllow = this.trAsk.replace("%1", this.trAllow);
            miAllow.setAttribute("label", labelAllow);
            miAllow.setAttribute("checked", "true");
          } else if (lifetimePolicy == MOZ_ACCEPT_FOR_N_DAYS) {
            let labelAllow = this.trLimit.replace("%1", this.trAllow);
            miAllow.setAttribute("label", labelAllow);
            miAllow.setAttribute("checked", "true");
          } else {
            let labelAllow = this.trDefault.replace("%1", this.trAllow);
            miAllow.setAttribute("label", labelAllow);
            miAllow.setAttribute("checked", "true");
          }
          break;
        }

        let mpDomain = doc.createElementNS(XUL_NS, "menupopup");
        mpDomain.setAttribute("id", "cookiekeeper-menupopup-" + place + "-perms-" + partialDomain);
        mpDomain.setAttribute("accessibleType", XULMenupopup);
        mpDomain.appendChild(miShowCookies);
        mpDomain.appendChild(miSep);
        mpDomain.appendChild(miAllow);
        mpDomain.appendChild(miSession);
        mpDomain.appendChild(miBlock);

        let menuDomain = doc.createElementNS(XUL_NS, "menu");
        menuDomain.setAttribute("id", "cookiekeeper-menu-" + place + "-perms-" + partialDomain);
        menuDomain.setAttribute("label", partialDomain);
        menuDomain.setAttribute("accessibleType", XULMenuitem);
        if (i==0) {
          menuDomain.setAttribute("class", "menu-iconic");
          menuDomain.setAttribute("image", iconURL);
        }
        menuDomain.appendChild(mpDomain);
        menupopup.appendChild(menuDomain);
      }
    }

    aEvent.stopPropagation();
    aEvent.stopImmediatePropagation();

    return true;
  },

  removeMenuPermissions: function(aEvent, menupopup, place) {
    if (aEvent.target.getAttribute("id") != "cookiekeeper-menupopup-" + place) {
      // For some reason, the event is triggered for the subpopup too
      // don't need to remove the items
      return false;
    }
    var doc  = aEvent.target.ownerDocument;
    var elem = doc.getElementById("cookiekeeper-menuitem-" + place + "-perms-sep");
    while (elem != null) {
      let tmp = elem;
      elem = elem.nextSibling;
      menupopup.removeChild(tmp);
    }
    aEvent.stopPropagation();
    aEvent.stopImmediatePropagation();

    return true;
  }
}

/**
 * Australis UI (Firefox 29+)
 */
var ckAustralisWindowListener = {
  _buttonCreated: false,
  _l10nAustralisPanel : null,

  onShowMenuPanel: function(aEvent) {
    let doc  = aEvent.target.ownerDocument;
    let panel  = doc.getElementById(CK_MENU_PANEL_ID);
    let iframe = doc.getElementById("cookiekeeper-panelview-iframe");
    iframe.setAttribute("src", "chrome://cookiekeeper/content/australis_panel.html");
  },

  onHideMenuPanel: function(aEvent) {
    let doc  = aEvent.target.ownerDocument;
    let panelView = doc.getElementById(CK_MENU_PANEL_ID);
    let iframe = doc.getElementById("cookiekeeper-panelview-iframe");
    iframe.setAttribute("src", "about:blank");
    panelView.style.width  = l10n.tr("australis.panel.width");
    panelView.style.height = l10n.tr("australis.panel.height");

    // Usefull ?
    // aEvent.stopImmediatePropagation();
    // aEvent.stopPropagation();
  },

  addUI: function(domWindow) {
    let doc = domWindow.document;
    domWindow.addEventListener("unload", ckUnloader, false);

    let enablePrivBrowse = false;
    if (Services.prefs.getPrefType(PREF_ENABLE_PRIV_BROWSE) == Services.prefs.PREF_BOOL) {
      enablePrivBrowse = Services.prefs.getBoolPref(PREF_ENABLE_PRIV_BROWSE);
    }

    let ckLabel        = l10n.tr("cookiekeeper.label");
    let ckTooltip      = l10n.tr("cookiekeeper.tooltip");
    let ckMenuShowMain = l10n.tr("menu.showcookies.label");
    let ckMenuRemAll   = l10n.tr("menu.removecookies.label");
    let ckMenuShowPref = l10n.tr("menu.preferences.label");

    let panelView = doc.createElement("panelview");
    panelView.setAttribute("id", CK_MENU_PANEL_ID);
    panelView.setAttribute("flex", "1");
    panelView.setAttribute("style", "overflow: hidden;");
    panelView.style.width  = l10n.tr("australis.panel.width");
    panelView.style.height = l10n.tr("australis.panel.height");

    let iframe = doc.createElement("iframe");
    iframe.setAttribute("id", "cookiekeeper-panelview-iframe");
    iframe.setAttribute("type", "chrome");
    iframe.setAttribute("flex", "1");
    iframe.addEventListener("DOMContentLoaded", function(event) {
      if (event.target.URL != "about:blank") {
        let currentURI = doc.defaultView.gBrowser.selectedBrowser.currentURI;

        if (ckAustralisWindowListener._l10nAustralisPanel == null) {
          ckAustralisWindowListener._l10nAustralisPanel = new Localization(addonData.resourceURI.spec, "australis_panel.properties")
        }

        // Allow the australis_panel.html to access the panelview,
        // to hide the panel via CustomizableUI.hidePanelForNode();
        event.target.ckPanel  = panelView;
        // Set the current tab URI
        event.target.ckURI = currentURI.cloneIgnoringRef();
        // The addon URI for the l10n
        event.target.l10n = ckAustralisWindowListener._l10nAustralisPanel;
      }
    });

    panelView.appendChild(iframe);

    let panelUI = doc.getElementById("PanelUI-multiView");
    if (panelUI) {
      panelUI.appendChild(panelView);
    }

    if (!ckAustralisWindowListener._buttonCreated) {
      ckAustralisWindowListener._buttonCreated = true;

      let buttonCK = {
        id            : CK_BUTTON_ID,
        defaultArea   : CustomizableUI.AREA_NAVBAR,
        label         : ckLabel,
        tooltiptext   : ckTooltip,
        type          : "view",
        viewId        : CK_MENU_PANEL_ID,
        onViewShowing : ckAustralisWindowListener.onShowMenuPanel,
        onViewHiding  : ckAustralisWindowListener.onHideMenuPanel,
        showInPrivateBrowsing : enablePrivBrowse // this add-on is not relevant in private browsing mode
      };
      CustomizableUI.createWidget(buttonCK);
    }

    // Add the menu items (under the Tools menu)

    let menu = doc.createElement("menu");
    menu.setAttribute("id", CK_MENU_ID);
    menu.setAttribute("class", "menu-iconic");
    menu.setAttribute("label", ckLabel);
    menu.setAttribute("tooltiptext", ckTooltip);

    let menuPopup = doc.createElement("menupopup");
    menuPopup.setAttribute("id", "cookiekeeper-menupopup-toolsmenu");
    menuPopup.addEventListener("popupshowing", function(event) { ckPermMenuManagement.addMenuPermissions(event, menuPopup, "toolsmenu");    });
    menuPopup.addEventListener("popuphidden",  function(event) { ckPermMenuManagement.removeMenuPermissions(event, menuPopup, "toolsmenu"); });
    menu.appendChild(menuPopup);

    let menuItem = doc.createElement("menuitem");
    menuItem.setAttribute("id", "cookiekeeper-menuitem-toolsmenu-view-cookies");
    menuItem.setAttribute("label", ckMenuShowMain);
    menuItem.setAttribute("accessibleType", XULMenuitem);
    menuItem.addEventListener("command", function(event) {CookieKeeper.showMainWindow(domWindow);});
    menuPopup.appendChild(menuItem);

    menuItem = doc.createElement("menuitem");
    menuItem.setAttribute("id", "cookiekeeper-menuitem-toolsmenu-view-preferences");
    menuItem.setAttribute("label", ckMenuShowPref);
    menuItem.setAttribute("accessibleType", XULMenuitem);
    menuItem.addEventListener("command", function(event) {CookieKeeper.showPreferencesDialogue(domWindow, null);});
    menuPopup.appendChild(menuItem);

    menuItem = doc.createElement("menuitem");
    menuItem.setAttribute("id", "cookiekeeper-menuitem-toolsmenu-delete-cookies");
    menuItem.setAttribute("label", ckMenuRemAll);
    menuItem.setAttribute("accessibleType", XULMenuitem);
    menuItem.addEventListener("command", function(event) {CookieKeeper.clearAllCookies(false);});
    menuPopup.appendChild(menuItem);

    let menuToolsPopup = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLSMENU);
    if (menuToolsPopup) {
      menuToolsPopup.appendChild(menu);
    }
  },

  removeUI: function(domWindow) {
    // Removes all added UI elements.
    let doc   = domWindow.document;
    let panelUI = doc.getElementById("PanelUI-multiView");
    let panel   = doc.getElementById(CK_MENU_PANEL_ID);
    if (panelUI && panel) {
      panelUI.removeChild(panel);
    }

    let toolsMenu = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLSMENU);
    let ckMenu    = doc.getElementById(CK_MENU_ID);
    if (toolsMenu !== null && ckMenu !== null) {
      toolsMenu.removeChild(ckMenu);
    }
  },

  onOpenWindow: function(aXULWindow) {
    // A new window has opened.
    let that = this;
    let domWindow = aXULWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);

    // Wait for it to finish loading
    domWindow.addEventListener("DOMContentLoaded", function ckListener() {
        domWindow.removeEventListener("DOMContentLoaded", ckListener, false);

        if (!isPrivateBrowsingMode(domWindow)) {
          // If this is a browser window then setup its UI
          let docElem = domWindow.document.documentElement;
          if (docElem.getAttribute("windowtype") === APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE) {
            that.addUI(domWindow);
          }
        }
      }, false);
  },

  onCloseWindow: function(aXULWindow) {
    let domWindow = aXULWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);
    let docElem = domWindow.document.documentElement;
    if (docElem.getAttribute("windowtype") === APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE) {
      this.removeUI(domWindow);
    }
  },

  onWindowTitleChange: function(aXULWindow, aNewTitle) {
  }
};

/**
 * Legacy UI (Thunderbird/SeaMonkey/Palemoon)
 */
var ckLegacyWindowListener = {

  insertButton: function(aDocument, aButton, insertAndPersist) {
    let toolbox = aDocument.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLBOX);
    if (toolbox.palette == null) {
      // When a new window is opened, the toolbox do not yet exists, items need to be
      // placed in the BrowserToolbarPalette like for an overlay toolbar button.
      // If placed in the BrowserToolbarPalette, Mozilla will restore itself the button position.
      toolbox = aDocument.getElementById(APP_MOZ_CONST[Services.appinfo.ID].PALETTE);
      toolbox.appendChild(aButton);
    } else {
      // The window was already opened and loaded, we need to know where to place the button
      // Put it for now in the palette
      toolbox.palette.appendChild(aButton);

      if (APP_MOZ_CONST[Services.appinfo.ID].NAME == FIREFOX) {
        // Australis UI exists only on Firefox
        try {
          // If Australis UI is available, use it to place the CK toolbar button.
          Cu.import("resource:///modules/CustomizableUI.jsm");

          let place = CustomizableUI.getPlacementOfWidget(aButton.id);
          if (place == null && insertAndPersist) {
            // Default to menu
            CustomizableUI.addWidgetToArea(aButton.id, CustomizableUI.AREA_NAVBAR);
          } else {
            CustomizableUI.ensureWidgetPlacedInWindow(aButton.id, aDocument.ownerGlobal);
          }

          return;
        } catch (ex) {
          log.d(ex);
        }
      }

      let toolbar = null;
      let nextItem = null; // to replace the button at the right position

      // Search the toolbar which have the attribute "currentset"  having our button id.
      // rem: not using toolbox children, because some toolbar may not be a direct child
      // of the toolbox, like the add-on bar. See browser.xul in Firefox source code.
      let toolbars = aDocument.querySelectorAll("toolbar");
      for (let i=0; i<toolbars.length && toolbar === null; i++) {
        toolbar = toolbars[i];
        let currentSet      = toolbar.getAttribute("currentset");
        let currentSetArray = currentSet == null ? [] : currentSet.split(",");
        let idx = currentSetArray.indexOf(aButton.id);
        if (idx == -1) {
          toolbar = null;
        } else {
          // Found the toolbar, now get the button position
          // Do a while loop, if the next element do not exists
          // (For example, a disabled addon)
          idx++;
          let nextItemId = currentSetArray[idx];
          while (idx < currentSetArray.length && nextItem == null) {
            nextItem = aDocument.getElementById(currentSetArray[idx]);
            if (nextItem == null) {
              // Maybe a spring, a separator or a space?
              let nChildren = toolbar.childNodes.length;
              nextItem = nChildren > idx-1 ? toolbar.childNodes[idx-1] : null;
              if (nextItem != null && nextItem.id.indexOf(nextItemId, 0) == -1) {
                // Apparently not
                nextItem = null;
              }
            }
            idx++;
          }
        }
      }

      if (toolbar == null) {
        if (insertAndPersist) {
          // Not found, use the default toolbar (nav-bar for FFx)
          toolbar = aDocument.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLBAR);
          toolbar.insertItem(aButton.id);
          // Save to the localstore.rdf
          toolbar.setAttribute("currentset", toolbar.currentSet);
          aDocument.persist(toolbar.id, "currentset");
        } // else { log.d("do not insert button"); }
      } else {
        toolbar.insertItem(aButton.id, nextItem);
      }
    }
  },

  addUI: function(domWindow, startupReason) {
    var buttonCK        = null;
    var menuCK          = null;
    var ckLabel         = null;
    var ckTooltip       = null;
    var ckMenuShowMain  = null;
    var ckMenuRemAll    = null;
    var ckMenuShowPref  = null;

    if (typeof(domWindow) === "undefined" || domWindow === null) {
      log.e("Error window not defined");
      return;
    }

    let doc = domWindow.document;

    function openCookieWindow(event) {
      CookieKeeper.showMainWindow(domWindow);
    }

    function openPreferencesWindow(event) {
      CookieKeeper.showPreferencesDialogue(domWindow, null);
    }

    function clearCookies(event) {
      CookieKeeper.clearAllCookies(false);
    }

    function createMenu(doc, place) {
      let miShowCK = doc.createElementNS(XUL_NS, "menuitem");
      miShowCK.setAttribute("id", "cookiekeeper-menuitem-" + place + "-view-cookies");
      miShowCK.setAttribute("label", ckMenuShowMain);
      miShowCK.setAttribute("accessibleType", XULMenuitem);
      miShowCK.setAttribute("class", "menuitem-iconic");
      miShowCK.addEventListener("command", openCookieWindow, false);

      let miShowPref = doc.createElementNS(XUL_NS, "menuitem");
      miShowPref.setAttribute("id", "cookiekeeper-menuitem-" + place + "-view-preferences");
      miShowPref.setAttribute("label", ckMenuShowPref);
      miShowPref.setAttribute("accessibleType", XULMenuitem);
      miShowPref.setAttribute("class", "menuitem-iconic");
      miShowPref.addEventListener("command", openPreferencesWindow, false);

      let miClear = doc.createElementNS(XUL_NS, "menuitem");
      miClear.setAttribute("id", "cookiekeeper-menuitem-" + place + "-delete-cookies");
      miClear.setAttribute("label",  ckMenuRemAll);
      miClear.setAttribute("accessibleType", XULMenuitem);
      miClear.setAttribute("class", "menuitem-iconic");
      miClear.addEventListener("command", clearCookies, false);

      let mp = doc.createElementNS(XUL_NS, "menupopup");
      mp.setAttribute("id", "cookiekeeper-menupopup-" + place);
      mp.setAttribute("accessibleType", XULMenupopup);
      mp.appendChild(miShowCK);
      mp.appendChild(miShowPref);
      mp.appendChild(miClear);
      if (APP_MOZ_CONST[Services.appinfo.ID].NAME != THUNDERBIRD) {
        mp.addEventListener("popupshowing", function(event) { ckPermMenuManagement.addMenuPermissions(event, mp, place);    });
        mp.addEventListener("popuphidden",  function(event) { ckPermMenuManagement.removeMenuPermissions(event, mp, place); });
      }

      return mp;
    }

    ckLabel        = l10n.tr("cookiekeeper.label");
    ckTooltip      = l10n.tr("cookiekeeper.tooltip");
    ckMenuShowMain = l10n.tr("menu.showcookies.label");
    ckMenuRemAll   = l10n.tr("menu.removecookies.label");
    ckMenuShowPref = l10n.tr("menu.preferences.label");

    // Add toolbar button
    buttonCK = doc.createElementNS(XUL_NS, "toolbarbutton");
    buttonCK.setAttribute("id", CK_BUTTON_ID);
    buttonCK.setAttribute("label", ckLabel);
    buttonCK.setAttribute("class", "toolbarbutton-1 chromeclass-toolbar-additional");
    buttonCK.setAttribute("tooltiptext", ckTooltip);
    buttonCK.setAttribute("persist", "class");
    buttonCK.setAttribute("accessibleType", XULToolbarButton);
    buttonCK.setAttribute("type", "menu");

    buttonCK.appendChild(createMenu(doc, "toolbar"));

    let insertAndPersist = (currReason == ADDON_INSTALL); // currReason is set in bootstrap.js
    this.insertButton(doc, buttonCK, insertAndPersist);

    // Add menu item under the tools-menu
    menuCK = doc.createElementNS(XUL_NS, "menu");
    menuCK.setAttribute("id", CK_MENU_ID);
    menuCK.setAttribute("class", "menu-iconic");
    menuCK.setAttribute("label", ckLabel);
    menuCK.setAttribute("tooltiptext", ckTooltip);
    menuCK.setAttribute("accessibleType", XULMenuitem);
    menuCK.appendChild(createMenu(doc, "toolsmenu"));

    let toolsMenu = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLSMENU);
    toolsMenu.appendChild(menuCK);

    domWindow.addEventListener("unload", ckUnloader, false);

    return true;
  },

  removeUI: function(domWindow) {
    let doc       = domWindow.document;
    let ckButton  = doc.getElementById(CK_BUTTON_ID);
    let toolsMenu = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLSMENU);
    let ckMenu    = doc.getElementById(CK_MENU_ID);

    if (ckButton != null) {
      ckButton.parentNode.removeChild(ckButton);
    } else {
      // The button may be in the toolbox palette
      let palette = doc.getElementById(APP_MOZ_CONST[Services.appinfo.ID].TOOLBOX).palette;
      let elems   = palette.childNodes;
      for (let i=0; i<elems.length; i++) {
        // Do a try catch if some elements are not accessible, like disabled addon which do not
        // have removed their button from the palette
        try {
          let elem = elems[i];
          if (CK_BUTTON_ID == elem.id) {
            palette.removeChild(elem);
          }
        } catch (ex) {
          log.e(ex);
        }
      }
    }
    if (toolsMenu !== null && ckMenu !== null) {
      toolsMenu.removeChild(ckMenu);
    } else {
      log.w("tools menu item not found");
    }
  },

  onOpenWindow: function(aXULWindow) {
    // A new window has opened.
    let domWindow = aXULWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);

    // Wait for it to finish loading
    domWindow.addEventListener("DOMContentLoaded", function ckListener() {
        domWindow.removeEventListener("DOMContentLoaded", ckListener, false);
        // If not in private browsing mode
        if (!isPrivateBrowsingMode(domWindow)) {
          // If this is a browser window then setup its UI
          let docElem = domWindow.document.documentElement;
          if (docElem.getAttribute("windowtype") == APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE) {
            ckLegacyWindowListener.addUI(domWindow);
          }
        }
      }, false);
  },

  onCloseWindow: function(aXULWindow) {
    let domWindow = aXULWindow.QueryInterface(Ci.nsIInterfaceRequestor).getInterface(Ci.nsIDOMWindow);
    let docElem   = domWindow.document.documentElement;
    if (docElem.getAttribute("windowtype") == APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE) {
      this.removeUI(domWindow);
    }
  },

  onWindowTitleChange: function(aXULWindow, aNewTitle) {
  }
};

// Partial K-Meleon support
var ckKMeleonUI = {

  initKMeleon: function(aCounter) {

    if (aCounter > 10) {
      return;
    }

    let chromeRegistry = Cc["@mozilla.org/chrome/chrome-registry;1"].getService(Ci.nsIChromeRegistry);
    log.d("chromeRegistry:", chromeRegistry);

    // Delay menu init, KMeleon.jsm is not yet available
    setTimeout(function() {
      try {
        Cu.import("resource:///modules/KMeleon.jsm");

        KMeleon.SetMenuCallback("&Tools", "CookieKeeper", function(mode) {
          KMeleon.Open('chrome://cookiekeeper/content/cookiekeeper.xul', KMeleon.OPEN_NEW);
        });

        KMeleon.SetMenuCallback("&Tools", "CookieKeeper Preferences", function(mode) {
          KMeleon.Open("chrome://cookiekeeper/content/preferences.html", KMeleon.OPEN_NEWTAB);
        });
      } catch (ex) {
        log.e(ex.message, ex);
        log.e("UI init failed (" + aCounter.toString(10) + ")");
        this.initKMeleon(aCounter + 1);
      }

      // Listen the preference "cookiekeeper.kmeleon.exit" to know when kmeleon exit
      // The cookiekeeper.kmm must be present in the K-Meleon macro directory.
      Services.prefs.setBoolPref("cookiekeeper.kmeleon.exit", false);
      Services.prefs.addObserver("cookiekeeper.kmeleon.exit", ckKMeleonUI, false);
    }, 100 * aCounter);
  },

  observe: function(aSubject, aTopic, aData) {
    try {
      if (aTopic == "nsPref:changed") {
        Services.prefs.removeObserver("cookiekeeper.kmeleon.exit", ckKMeleonUI);
        currReason = APP_SHUTDOWN;
        shutdown(addonData, APP_SHUTDOWN);
      }
    } catch(ex) {
      log.d(ex);
    }
  }
}

var ckUi = {
  _useAustralis : false,
  _uriCSS       : null,

  init: function() {

    if (APP_MOZ_CONST[Services.appinfo.ID].NAME == KMELEON) {
      // K-Meleon require is a special case.
      ckKMeleonUI.initKMeleon(0);
      return;
    }

    // Inject the CSS using the stylesheet service, for the toolbar icon:
    let ssService = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
    let windowListener = ckLegacyWindowListener;

    this._uriCSS  = Services.io.newURI(CK_TOOLBAR_CSS, null, null);
    this._useAustralis = false;

    if (APP_MOZ_CONST[Services.appinfo.ID].NAME == FIREFOX) {
      let useAustralis = true;
      if (Services.prefs.getPrefType(PREF_NO_AUSTRALIS_UI) == Services.prefs.PREF_BOOL) {
        useAustralis = !Services.prefs.getBoolPref(PREF_NO_AUSTRALIS_UI);
      }

      if (useAustralis) {
        try {
          // Setup the toolbar icon for Australis UI (Firefox 29+)
          // https://blog.mozilla.org/addons/2014/03/06/australis-for-add-on-developers-2/
          // Will fail (throw an exeption) if not found:
          Cu.import("resource:///modules/CustomizableUI.jsm");

          windowListener = ckAustralisWindowListener;
          this._useAustralis = true;
        } catch (ex) {
          log.e(ex);
        }
      }
    }

    if (!this._useAustralis) {
      ckPermMenuManagement.init();
    }

    if (!ssService.sheetRegistered(this._uriCSS, ssService.AUTHOR_SHEET)) {
      ssService.loadAndRegisterSheet(this._uriCSS, ssService.AUTHOR_SHEET);
    }

    // Add the menu panel to existing windows:
    let enumerator = Services.wm.getEnumerator(APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE);
    while (enumerator.hasMoreElements()) {
      let domWindow = enumerator.getNext();
      if (!isPrivateBrowsingMode(domWindow)) {
        windowListener.addUI(domWindow);
      }
    }

    // Window listener to add the menu panel on new windows:
    Services.wm.addListener(windowListener);
  }, // init

  uninit: function() {
    if (APP_MOZ_CONST[Services.appinfo.ID].NAME == KMELEON) {
      return;
    }

    try {
      CookieKeeper.closeCookieWindow()
    } catch (ex) {
      log.e(ex);
    }

    let windowListener = null;
    if (this._useAustralis) {
      windowListener = ckAustralisWindowListener;
    } else {
      windowListener = ckLegacyWindowListener;
    }

    let enumerator = Services.wm.getEnumerator(APP_MOZ_CONST[Services.appinfo.ID].WINDOW_TYPE);

    Services.wm.removeListener(windowListener);

    while (enumerator.hasMoreElements()) {
      let aWindow = enumerator.getNext();
      aWindow.removeEventListener("unload", ckUnloader, false);
      windowListener.removeUI(aWindow);
    }

    if (this._useAustralis) {
      CustomizableUI.destroyWidget(CK_BUTTON_ID);
    } else {
      ckPermMenuManagement.uninit();
    }

    let ssService = Cc["@mozilla.org/content/style-sheet-service;1"].getService(Ci.nsIStyleSheetService);
    if (this._uriCSS != null && ssService.sheetRegistered(this._uriCSS, ssService.AUTHOR_SHEET)) {
      ssService.unregisterSheet(this._uriCSS, ssService.AUTHOR_SHEET);
    }

    this._uriCSS  = null;
  } // uninit

}; // ckUi
