/* ************************************************************************
 *
 * CookieKeeper - a Mozilla add-on
 * (c) 2013-2014 Yvon TANGUY
 *
 * ***** BEGIN LICENSE BLOCK **********************************************
 *
 * This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 * ***** END LICENSE BLOCK **********************************************
 *
 * Unicode conversion.
 * Partially based on the obsolete ConvertUTF.c (no longer available on unicode.org)
 *
 * ************************************************************************
 */

const UNI_REPLACEMENT_CHAR = 0x0000FFFD|0;
const UNI_MAX_LEGAL_UTF32  = 0x0010FFFF|0;

const UNI_SUR_HIGH_START   = 0xD800|0;
const UNI_SUR_HIGH_END     = 0xDBFF|0;
const UNI_SUR_LOW_START    = 0xDC00|0;
const UNI_SUR_LOW_END      = 0xDFFF|0;

const UNI_HALF_SHIFT       = 10|0; // used for shifting by 10 bits
const UNI_HALF_BASE        = 0x10000|0;
const UNI_HALF_MASK        = 0x003FF|0;

/**
 * Once the bits are split out into bytes of UTF-8, this is a mask OR-ed
 * into the first byte, depending on how many bytes follow.  There are
 * as many entries in this table as there are UTF-8 sequence types.
 * (I.e., one byte sequence, two byte... etc.). Remember that sequencs
 * for *legal* UTF-8 will be 4 or fewer bytes total.
 */
const UNI_FIRST_BYTE_MARK = [0x00, 0x00, 0xC0, 0xE0, 0xF0, 0xF8, 0xFC];

const UNI_ERR = {
  NO_ERROR        : 0, /// No error
  SOURCE_ILLEGAL  : 1, /// Illegal source sequence
  SOURCE_EXHAUSTED: 2  /// Not enough character in source (when having surrogate pair)
}

var gUnicode = {

  /**
   * Convert a UTF-16 string to an UTF-32 array
   * @param [in string] utf16 string
   * @param [in boolean] strictConversion if true, stop at the first error,
   *                     else change invalid character with UNI_REPLACEMENT_CHAR
   * @return a object: { unicode: <array of int>, position: <int> error: <UNI_ERR error_value> }
   * The return position should be the utf16 string length if there is no error, or if strictConversion==false
   */
  utf16_to_utf32: function(utf16, strictConversion) {
    if (utf16.length == 0) {
      // Empty string, nothing to do
      return {unicode: [], error: UNI_ERR.NO_ERROR};
    }

    var utf32    = [];
    var position = 0;
    var error    = UNI_ERR.NO_ERROR;
    while (position < utf16.length) {
      var unicode  = 0;
      var advance  = 0;

      var ch = utf16.charCodeAt(position);

      if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_HIGH_END) {
        if (position+1 < utf16.length) {
          let ch2 = utf16.charCodeAt(position+1);
          if (ch2 >= UNI_SUR_LOW_START && ch2 <= UNI_SUR_LOW_END) {
            unicode = ((ch - UNI_SUR_HIGH_START) << UNI_HALF_SHIFT) + (ch2 - UNI_SUR_LOW_START) + UNI_HALF_BASE;
            advance += 2;
          } else if (strictConversion) {
            error   = UNI_ERR.SOURCE_ILLEGAL;
            unicode = UNI_REPLACEMENT_CHAR;
          } else {
            unicode = ch;
            advance++;
          }
        } else if (strictConversion) {
          unicode = UNI_REPLACEMENT_CHAR;
          error   = UNI_ERR.SOURCE_EXHAUSTED;
        } else {
          unicode = ch;
          advance++;
        }
      } else if (ch >= UNI_SUR_LOW_START && ch <= UNI_SUR_LOW_END && strictConversion) {
        // Error in sequence
        error   = UNI_ERR.SOURCE_ILLEGAL;
        unicode = UNI_REPLACEMENT_CHAR;
      } else {
        unicode = ch;
        advance++;
      }

      position += advance;
      if (error == UNI_ERR.NO_ERROR || !strictConversion) {
        utf32.push(unicode|0);
      } else {
        break;
      }
    }

    return {unicode: utf32, position: position, error: error};
  },

  /**
   * Convert an UTF-32 array to an UTF-8 string.
   * @param utf32 : an array of utf32 code point
   * @param strictConversion is true, stop at the first error
   * @return object {utf8: <string>, error: error}
   */
  utf32_to_utf8: function(utf32, strictConversion) {
    const BYTE_MASK = 0xBF;
    const BYTE_MARK = 0x80;
    var utf8array = [];
    var position  = 0;
    var error     = UNI_ERR.NO_ERROR;
    while (position < utf32.length) {
      var ch = utf32[position];
      let bytesToWrite = 0;
      if (strictConversion) {
        // UTF-16 surrogate values are illegal in UTF-32
        if (ch >= UNI_SUR_HIGH_START && ch <= UNI_SUR_LOW_END) {
          error = UNI_ERR.SOURCE_ILLEGAL;
          break;
        }
      }

      if (ch < 0x80) {
        bytesToWrite = 1;
      } else if (ch < 0x800) {
        bytesToWrite = 2;
      } else if (ch < 0x10000) {
        bytesToWrite = 3;
      } else if (ch <= UNI_MAX_LEGAL_UTF32) {
        bytesToWrite = 4;
      } else {
        bytesToWrite = 3;
        ch = UNI_REPLACEMENT_CHAR;
        result = sourceIllegal;
      }

      let utf8char = [0, 0, 0, 0];
      switch (bytesToWrite) {
      case 4:
        utf8char[3] = ((ch | BYTE_MARK) & BYTE_MASK);
        ch >>= 6;
        // fall-through
      case 3:
        utf8char[2] = ((ch | BYTE_MARK) & BYTE_MASK);
        ch >>= 6;
        // fall-through
      case 2:
        utf8char[1] = ((ch | BYTE_MARK) & BYTE_MASK);
        ch >>= 6;
        // fall-through
      case 1:
        utf8char[0] = (ch | UNI_FIRST_BYTE_MARK[bytesToWrite]);
      }
      for (let i=0; i<bytesToWrite; i++) {
        utf8array.push(String.fromCharCode(utf8char[i]));
      }
      position++;
    }

    return {utf8: utf8array.join(""), error: error};
  }

};
